/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <IPbasic/IPbasic_NLP.h>
#include <IPbasic/IPbasic_Solver.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_MA48.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_MA57.h>
#include <algorithm>

using IPbasic::int_type;
using IPbasic::MatrixWrapper;
using IPbasic::ostream_type;
using IPbasic::real_type;
using IPbasic::SparseMatrixBase;
using IPbasic::string_type;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

/*\

  Example of user defined class for sparse matrix storage

\*/

class SparseMatrix : public SparseMatrixBase<real_type>
{
    std::vector<int_type>  rows, cols;
    std::vector<real_type> vals;
    int_type               nr, nc, nnz;
    bool                   fi;

  public:
    SparseMatrix()
        : SparseMatrixBase<real_type>()
    {
        this->fi = false;
    }

    virtual ~SparseMatrix() IPBASIC_OVERRIDE;

    void init(int_type N, int_type M, int_type reserve_nnz, bool fi = false) IPBASIC_OVERRIDE
    {
        this->fi  = fi;
        this->nr  = N;
        this->nc  = M;
        this->nnz = 0;
        this->setZero();
        this->rows.reserve(reserve_nnz);
        this->cols.reserve(reserve_nnz);
        this->vals.reserve(reserve_nnz);
    }

    /*!
     *  \brief setZero the sparse matrix
     *
     */
    void clear() IPBASIC_OVERRIDE
    {
        this->nr  = 0;
        this->nc  = 0;
        this->nnz = 0;
        vals.clear();
        cols.clear();
        rows.clear();
    }

    /*!
     *  \brief setZero the sparse matrix
     *
    \*/
    void setZero() IPBASIC_OVERRIDE
    {
        this->nnz = 0;
        vals.clear();
        cols.clear();
        rows.clear();
    }

    /*!
     *  \brief Return true if index are 1-based
     */
    virtual bool FORTRAN_indexing() const IPBASIC_OVERRIDE
    {
        return false;
    }

    /*!
     *  \brief transpose the matrix
     */
    virtual void transpose() IPBASIC_OVERRIDE
    {
        for (size_t k = 0; k < size_t(this->nnz); ++k)
            std::swap(rows[k], cols[k]);
    }

    virtual void to_FORTRAN_indexing() IPBASIC_OVERRIDE
    {
        IPBASIC_ASSERT(false, "to_FORTRAN_indexing()");
    }

    virtual void to_C_indexing() IPBASIC_OVERRIDE
    {
        IPBASIC_ASSERT(false, "to_C_indexing()");
    }

    virtual int_type get_number_of_rows() const IPBASIC_OVERRIDE
    {
        return this->nr;
    }

    virtual int_type get_number_of_cols() const IPBASIC_OVERRIDE
    {
        return this->nc;
    }

    virtual int_type get_nnz() const IPBASIC_OVERRIDE
    {
        return this->nnz;
    }

    virtual void get_info(int_type & numRows, int_type & numCols, int_type & nnzOut) const IPBASIC_OVERRIDE
    {
        numRows = this->nr;
        numCols = this->nc;
        nnzOut  = this->nnz;
    }

    virtual void get_data(int_type const *& pRows, int_type const *& pCols, real_type const *& pValues) const IPBASIC_OVERRIDE
    {
        pRows   = &rows.front();
        pCols   = &cols.front();
        pValues = &vals.front();
    }

    virtual void get_matrix(MatrixWrapper<real_type> & /*M*/) const IPBASIC_OVERRIDE
    {
        IPBASIC_ASSERT(false, "get_matrix()");
    }

    virtual void get_matrix_symmetric(MatrixWrapper<real_type> & /*M*/) const IPBASIC_OVERRIDE
    {
        IPBASIC_ASSERT(false, "get_matrix_symmetric()");
    }

    virtual void get_matrix_transposed(MatrixWrapper<real_type> & /*M*/) const IPBASIC_OVERRIDE
    {
        IPBASIC_ASSERT(false, "get_matrix_transposed()");
    }

    /*!
     *  \brief insert a value into matrix using 0-based index
     *
     *  \param[in] row the row index
     *  \param[in] col the column index
     *  \param[in] val the inserted value
     *
     */
    virtual void push_value_C(int_type row, int_type col, valueType val) IPBASIC_OVERRIDE
    {
        this->rows.push_back(row);
        this->cols.push_back(col);
        this->vals.push_back(val);
        this->nnz++;
    }

    /*!
     *  \brief insert a value into matrix using 1-based index
     *
     *  \param[in] row the row index
     *  \param[in] col the column index
     *  \param[in] val the inserted value
     *
     */
    virtual void push_value_F(int_type row, int_type col, valueType val) IPBASIC_OVERRIDE
    {
        this->rows.push_back(row + 1);
        this->cols.push_back(col + 1);
        this->vals.push_back(val);
        this->nnz++;
    }

    /*!
     *  \return true if Inf or NaN are found in the sparse data
    \*/
    virtual bool foundNaN() const IPBASIC_OVERRIDE
    {
        return false;
    }

    virtual void gemv(valueType alpha, int_type /*DimX*/, valueType const x[], int_type incX, valueType /*beta*/, int_type /*DimY*/, valueType y[], int_type incY) const IPBASIC_OVERRIDE
    {
        for (size_t k = 0; k < size_t(this->nnz); ++k)
        {
            int_type  i = rows[k];
            int_type  j = cols[k];
            real_type v = alpha * vals[k];
            y[i * incY] += v * x[j * incX];
        }
    }

    virtual void gemv_Transposed(valueType alpha, int_type /*DimX*/, valueType const x[], int_type incX, valueType /*beta*/, int_type /*DimY*/, valueType y[], int_type incY) const IPBASIC_OVERRIDE
    {
        for (size_t k = 0; k < size_t(this->nnz); ++k)
        {
            int_type  i = rows[k];
            int_type  j = cols[k];
            real_type v = alpha * vals[k];
            std::swap(i, j);
            y[i * incY] += v * x[j * incX];
        }
    }

    virtual void gemv_Symmetric(valueType alpha, int_type /*DimX*/, valueType const x[], int_type incX, valueType /*beta*/, int_type /*DimY*/, valueType y[], int_type incY) const IPBASIC_OVERRIDE
    {
        for (size_t k = 0; k < size_t(this->nnz); ++k)
        {
            int_type  i = rows[k];
            int_type  j = cols[k];
            real_type v = alpha * vals[k];
            if (i == j)
            {
                y[i * incY] += v * x[j * incX];
            }
            else
            {
                y[i * incY] += v * x[j * incX];
                std::swap(i, j);
                y[i * incY] += v * x[j * incX];
            }
        }
    }
};

SparseMatrix::~SparseMatrix()
{
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class Problem : public IPbasic::NLP<real_type>
{

    SparseMatrix Constraint_Jacobian;
    SparseMatrix Lagrangian_Hessian;
    SparseMatrix MyIPMatrix;
    bool         Is_equality[2];

  public:
    Problem()
        : NLP<real_type>(&this->Constraint_Jacobian, &this->Lagrangian_Hessian, &this->MyIPMatrix, this->Is_equality)
    {
        this->Constraint_Jacobian.init(2, 2, 4);
        this->Lagrangian_Hessian.init(2, 2, 4);
        this->MyIPMatrix.init(4, 4, 16);
        this->is_equality_vec[0] = false;
        this->is_equality_vec[1] = false;
    }

    int_type state_dimension() const IPBASIC_OVERRIDE
    {
        return 2;
    }

    virtual void initial_guess(real_type x[]) const IPBASIC_OVERRIDE
    {
        x[0] = 0.6;
        x[1] = 2.0;
    }

    virtual real_type objective(real_type const x[]) IPBASIC_OVERRIDE
    {
        return x[0] * x[0] + x[1] * x[1];
    }

    virtual void objective_gradient(real_type const x[], real_type grad[]) IPBASIC_OVERRIDE
    {
        grad[0] = 2 * x[0];
        grad[1] = 2 * x[1];
    }

    virtual int_type constraint_dimension() const IPBASIC_OVERRIDE
    {
        return 2;
    }

    virtual void constraint(real_type const x[], real_type c[]) IPBASIC_OVERRIDE
    {
        c[0] = -x[1] + 1.5;
        c[1] = -x[1] + exp(x[0]);
    }

    void constraint_jacobian_update(real_type const x[]) IPBASIC_OVERRIDE
    {
        this->Constraint_Jacobian.setZero();
        this->Constraint_Jacobian.push_value_C(0, 0, 0.0);
        this->Constraint_Jacobian.push_value_C(0, 1, -1.0);
        this->Constraint_Jacobian.push_value_C(1, 0, +exp(x[0]));
        this->Constraint_Jacobian.push_value_C(1, 1, -1.0);
    }

    void lagrangian_hessian_update(real_type sigma, real_type const x[], real_type const lambda[]) IPBASIC_OVERRIDE
    {
        this->Lagrangian_Hessian.setZero();
        this->Lagrangian_Hessian.push_value_C(0, 0, 2 * sigma + lambda[1] * exp(x[0]));
        this->Lagrangian_Hessian.push_value_C(1, 1, 2 * sigma);
    }
};

int main(int, char *[])
{
    Problem                            prob;
    IPbasic::LinearSolver_MA57<double> MyHsl;
    // IPbasic::LinearSolver_MA48<double> MyHsl;
    IPbasic::Solver<double> MySolver(&prob, &MyHsl);

    if (MySolver.optimize())
    {
        std::cout << "Job done!" << std::endl;
    }
    else
    {
        std::cout << "Job not done!" << std::endl;
    }

    return 0;
}
