/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <IPbasic/IPbasic_NLP.h>
#include <IPbasic/IPbasic_Solver.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_MA48.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_MA57.h>
#include <algorithm>

using IPbasic::int_type;
using IPbasic::MatrixWrapper;
using IPbasic::ostream_type;
using IPbasic::real_type;
using IPbasic::SparseCCOOR;
using IPbasic::string_type;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class Problem : public IPbasic::NLP<real_type>
{
    SparseCCOOR<real_type> Constraint_Jacobian;
    SparseCCOOR<real_type> Lagrangian_Hessian;
    bool                   Is_equality[1];
    real_type              epsilon;

  public:
    Problem(real_type epsilon)
        : NLP<real_type>(&this->Constraint_Jacobian, &this->Lagrangian_Hessian, this->Is_equality)
        , epsilon(epsilon)
    {
        this->Constraint_Jacobian.init(1, 2, 2);
        this->Lagrangian_Hessian.init(2, 2, 4);
        this->is_equality_vec[0] = false;
    }

    int_type state_dimension() const IPBASIC_OVERRIDE
    {
        return 2;
    }

    virtual void initial_guess(real_type x[]) const IPBASIC_OVERRIDE
    {
        x[0] = 0.0;      // 0.999999;
        x[1] = 0.999999; // 0.0;
    }

    virtual real_type objective(real_type const x[]) IPBASIC_OVERRIDE
    {
        return x[0]; //+ 2.0 * x[1];
    }

    virtual void objective_gradient(real_type const[], real_type grad[]) IPBASIC_OVERRIDE
    {
        grad[0] = 1.0;
        grad[1] = 0.0; // 2.0;
    }

    virtual int_type constraint_dimension() const IPBASIC_OVERRIDE
    {
        return 1;
    }

    virtual void constraint(real_type const x[], real_type c[]) IPBASIC_OVERRIDE
    {
        c[0] = x[0] * x[0] + x[1] * x[1] - 1.0;
    }

    void constraint_jacobian_update(real_type const x[]) IPBASIC_OVERRIDE
    {
        this->Constraint_Jacobian.setZero();
        this->Constraint_Jacobian.push_value_C(0, 0, 2.0 * x[0]);
        this->Constraint_Jacobian.push_value_C(0, 1, 2.0 * x[1]);
    }

    void lagrangian_hessian_update(real_type sigma, real_type const[], real_type const lambda[]) IPBASIC_OVERRIDE
    {
        this->Lagrangian_Hessian.setZero();
        real_type val;
        val = lambda[0] * 2.0;
        this->Lagrangian_Hessian.push_value_C(0, 0, val);
        val = 0.0;
        // this->Lagrangian_Hessian.push_value_C(0, 1, val);
        this->Lagrangian_Hessian.push_value_C(1, 0, val);
        val = lambda[0] * 2.0;
        this->Lagrangian_Hessian.push_value_C(1, 1, 2 * sigma);
    }
};

int main(int, char *[])
{
    Problem                            prob(0.0);
    IPbasic::LinearSolver_MA57<double> MyHsl;
    // IPbasic::LinearSolver_MA48<double> MyHsl;
    IPbasic::IPparameter<double> Param;
    IPbasic::IPtimings<double>   Timings;
    IPbasic::Solver<double>      MySolver(&prob, &MyHsl, &Param, &Timings, IPbasic::LINESEARCH_NONE);
    Param.mu_0                = 1e-3;
    Param.mu_tol              = 1e-8;
    Param.epsilon_tol         = 1e-10;
    Param.UseMuSequ           = false;
    Param.MyLineSearchCouples = IPbasic::COUPLE_STATE_SLACK;
    Param.max_iter            = 2000;

    if (MySolver.optimize())
    {
        std::cout << "Job done!" << std::endl;
    }
    else
    {
        std::cout << "Job not done!" << std::endl;
    }

    double x[2];
    MySolver.copySolution(x);

    std::cout << "x:       " << x[0] << std::endl;
    std::cout << "y:       " << x[1] << std::endl;
    std::cout << "x^2+y^2: " << x[0] * x[0] + x[1] * x[1] << std::endl;

    return 0;
}
