/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <IPbasic/Errors/IPbasic_Errors.h>
#include <IPbasic/IPbasic.h>
#include <IPbasic/IPbasic_NLP.h>
#include <IPbasic/IPbasic_Solver.h>
#include <algorithm>

using IPbasic::int_type;
using IPbasic::ostream_type;
using IPbasic::real_type;
using IPbasic::string_type;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class Problem : public IPbasic::NLP<real_type>
{

    lapack_wrapper::SparseCCOOR<real_type> Constraint_Jacobian;
    lapack_wrapper::SparseCCOOR<real_type> Lagrangian_Hessian;
    lapack_wrapper::SparseCCOOR<real_type> MyIPMatrix;
    bool                                   Is_equality[2];

  public:
    Problem()
        : NLP<real_type>(&this->Constraint_Jacobian, &this->Lagrangian_Hessian, &this->MyIPMatrix, this->Is_equality, false, false)
    {
        this->Constraint_Jacobian.setup_as_full_column_major(2, 2);
        this->Lagrangian_Hessian.setup_as_full_column_major(2, 2);
        this->MyIPMatrix.setup_as_full_column_major(4, 4);
        // this->MyIPMatrix.setSym(true);
        this->is_equality_vec[0] = false;
        this->is_equality_vec[1] = false;
    }

    int_type state_dimension() const IPBASIC_OVERRIDE
    {
        return 2;
    }

    virtual void initial_guess(real_type x[]) const IPBASIC_OVERRIDE
    {
        x[0] = 0.6;
        x[1] = 2.0;
    }

    // ---------------------------------------------------------------------------

    virtual real_type objective(real_type const x[]) IPBASIC_OVERRIDE
    {
        return x[0] * x[0] + x[1] * x[1];
    }

    virtual void objective_gradient(real_type const x[], real_type grad[]) IPBASIC_OVERRIDE
    {
        grad[0] = 2 * x[0];
        grad[1] = 2 * x[1];
    }

    // ---------------------------------------------------------------------------

    virtual int_type constraint_dimension() const IPBASIC_OVERRIDE
    {
        return 2;
    }

    virtual void constraint(real_type const x[], real_type c[]) IPBASIC_OVERRIDE
    {
        c[0] = -x[1] + 1.5;
        c[1] = -x[1] + exp(x[0]);
    }

    void constraint_jacobian_update(real_type const x[]) IPBASIC_OVERRIDE
    {
        lapack_wrapper::MatrixWrapper<real_type> M;
        this->Constraint_Jacobian.get_full_view(M);
        M(0, 0) = 0.0;
        M(0, 1) = -1.0;
        M(1, 0) = exp(x[0]);
        M(1, 1) = -1.0;
    }

    // ---------------------------------------------------------------------------

    void lagrangian_hessian_update(real_type sigma, real_type const x[], real_type const lambda[]) IPBASIC_OVERRIDE
    {
        lapack_wrapper::MatrixWrapper<real_type> M;
        this->Lagrangian_Hessian.get_full_view(M);
        M(0, 0) = 2 * sigma + lambda[1] * exp(x[0]);
        M(0, 1) = 0.0;
        M(1, 0) = 0.0;
        M(1, 1) = 2 * sigma;
    }
};

class DenseSolver : public IPbasic::LinearSolver_interface<double>
{
    lapack_wrapper::LU<double> lu;

  public:
    /*!
     * \brief DenseSolver:
     *        Constructor of the class DenseSolver.
     */
    DenseSolver()
        : LinearSolver_interface<double>()
    {
    }

    /*!
     * \brief ~DenseSolver:
     *        Virtual destructor of the class DenseSolver.
     */
    virtual ~DenseSolver()
    {
    }

    /*\
    :|:   ___      _    _ _      __  __           _
    :|:  | _ \_  _| |__| (_)__  |  \/  |___ _ __ | |__  ___ _ _ ___
    :|:  |  _/ || | '_ \ | / _| | |\/| / -_) '  \| '_ \/ -_) '_(_-<
    :|:  |_|  \_,_|_.__/_|_\__| |_|  |_\___|_|_|_|_.__/\___|_| /__/
    \*/

    /*!
     * \brief factorize:
     *        Factorizes the matrix \f$ A\f$ for a sparse matrix.
     *
     * \param[in]     Sparse  Sparse matrix which has to be solved.
     *
     * \return True if the the matrix \f$ A\f$ could be factorized.
     *
     */
    virtual bool factorize(lapack_wrapper::SparseMatrixBase<double> const & Sparse, bool is_stored_symmetric)
    {
        int_type const * pRows;
        int_type const * pCols;
        double const *   pValues;
        int_type         numRows, numCols, nnz;
        Sparse.get_info(numRows, numCols, nnz);
        Sparse.get_data(pRows, pCols, pValues);
        lu.allocate(numRows, numCols);
        if (is_stored_symmetric)
            lu.load_sparse_sym(nnz, pValues, pRows, pCols);
        else
            lu.load_sparse(nnz, pValues, pRows, pCols);
        lu.factorize("factorize");
        return true;
    }

    /*!
     * \brief solve:
     *        Solves the linear system \f$ Ax = b\f$ for \f$ A\f$ for a sparse matrix.
     *
     * \param[in,out] RHS In:  right-hand-side of the linear system.
     *                    OUT: solution of the linear system.
     *
     * \return True if the linear system \f$ Ax = b\f$ for \f$ A\f$ could be solved.
     *
     */
    bool solve(real_type RHS[])
    {
        lu.solve(RHS);
        return true;
    }
};

int main(int, char *[])
{
    Problem     prob;
    DenseSolver MyDenseSolver;

    IPbasic::Solver<double> *      MySolver = new IPbasic::Solver<double>(&prob, &MyDenseSolver);
    IPbasic::IPparameter<double> * Param    = MySolver->getIPparameter();
    Param->mu_0                             = 1e-3;
    Param->PrintLev                         = IPbasic::PRINT_MARKED;

    if (MySolver->optimize())
    {
        std::cout << "Job done!" << std::endl;
    }
    else
    {
        std::cout << "Job not done!" << std::endl;
    }

    double x[2];
    MySolver->copySolution(x);

    std::cout << "x:       " << x[0] << std::endl;
    std::cout << "y:       " << x[1] << std::endl;
    std::cout << "x^2+y^2: " << x[0] * x[0] + x[1] * x[1] << std::endl;

    delete MySolver;
    return 0;
}
