/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file minimumenergyproblem.h
 * Header Definition der Klasse MinimumEnergyProblem.
 * Genauere Informationen sind in der Dokumentation der Klasse MinimumEnergyProblem enthalten.
 *
 * @author A. Huber
 * @since 25.09.2018
 */

#ifndef MINIMUMENERGYPROBLEM_H
#define MINIMUMENERGYPROBLEM_H

#include <OCPbasic/problem.h>

class MinimumEnergyProblem : public OCPbasic::Problem<double>
{
  public:
    /**
     * \brief MinimumEnergyProblem:
     *        Constructor of the class MinimumEnergyProblem.
     */
    explicit MinimumEnergyProblem();

    /**
     * \brief ~MinimumEnergyProblem:
     *        Virtual destructor of the class MinimumEnergyProblem.
     */
    virtual ~MinimumEnergyProblem();

    /**
     * \brief setRangeBoxBounds:
     *        Initialize the bounds of the constraints.
     */
    void setRangeBoxBounds();

    /**
     * \brief ODE:
     *        Righthand side of the ODE.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] f      Righthand side of the ODE as a vector.
     *
     */
    virtual void ODE(double t, double const y[], double const u[], double const p[], double f[]) const;

    /**
     * \brief NLCSTR:
     *        Nonlinear constaints.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] gg     Nonlinear constaints as a vector.
     *
     */
    virtual void NLCSTR(double t, double const y[], double const u[], double const p[], double gg[]) const;

    /**
     * \brief BOUND:
     *        Boundary constaints.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] psi    Boundary constaints as a vector.
     *
     */
    virtual void BOUND(double t0, double tF, double const yF[], double const y0[], double const p[], double psi[]) const;

    /**
     * \brief VARPHI:
     *        Target or objective function.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] phi    Target or objective function (vector of size 1).
     *
     */
    virtual void VARPHI(double t0, double tF, double const yF[], double const y0[], double const p[], double phi[]) const;

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Erste Ableitungen                                                                //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_ODE_X:
     *        First derivative or jacobian of the ODEs towards \f$y\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the ODEs towards \f$y\f$.
     *
     */
    virtual void D_ODE_X(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_ODE_U:
     *        First derivative or jacobian of the ODEs towards \f$u\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the ODEs towards \f$u\f$.
     *
     */
    virtual void D_ODE_U(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_ODE_P:
     *        First derivative or jacobian of the ODEs towards \f$p\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the ODEs towards \f$p\f$.
     *
     */
    virtual void D_ODE_P(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_NLCSTR_X:
     *        First derivative or jacobian of the nonlinear constraints towards \f$y\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$y\f$.
     *
     */
    virtual void D_NLCSTR_X(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_NLCSTR_U:
     *        First derivative or jacobian of the nonlinear constraints towards \f$u\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$u\f$.
     *
     */
    virtual void D_NLCSTR_U(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_NLCSTR_P:
     *        First derivative or jacobian of the nonlinear constraints towards \f$p\f$.
     *
     * \param[in]  t      Timepoint \f$t\f$.
     * \param[in]  y      States at \f$y(t)\f$.
     * \param[in]  u      Controls at \f$u(t)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the nonlinear constraints towards \f$p\f$.
     *
     */
    virtual void D_NLCSTR_P(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_BOUND_XF:
     *        First derivative or jacobian of the boundary constraints towards \f$y_f\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$y_f\f$.
     *
     */
    virtual void D_BOUND_XF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_BOUND_X0:
     *        First derivative or jacobian of the boundary constraints towards \f$y_0\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$y_0\f$.
     *
     */
    virtual void D_BOUND_X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_BOUND_P:
     *        First derivative or jacobian of the boundary constraints towards \f$p\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the boundary constraints towards \f$p\f$.
     *
     */
    virtual void D_BOUND_P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_VARPHI_XF:
     *        First derivative or jacobian of the target function towards \f$y_f\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the target function towards \f$y_f\f$.
     *
     */
    virtual void D_VARPHI_XF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_VARPHI_X0:
     *        First derivative or jacobian of the target function towards \f$y_0\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the target function towards \f$y_0\f$.
     *
     */
    virtual void D_VARPHI_X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_VARPHI_P:
     *        First derivative or jacobian of the target function towards \f$p\f$.
     *
     * \param[in]  t0     Start timepoint \f$t_0\f$.
     * \param[in]  tF     End timepoint \f$t_f\f$.
     * \param[in]  yF     States at \f$y(t_f)\f$.
     * \param[in]  y0     States at \f$y(t_0)\f$.
     * \param[in]  p      Parameters \f$p\f$.
     * \param[out] S_OUT  Jacobian of the target function towards \f$p\f$.
     *
     */
    virtual void D_VARPHI_P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Zweite Ableitungen                                                               //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_ODE_XX:
     *        Second derivative or hessian of the ODEs towards \f$y,y\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$y,y\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_XX(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_UU:
     *        Second derivative or hessian of the ODEs towards \f$u,u\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$u,u\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_UU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_PP:
     *        Second derivative or hessian of the ODEs towards \f$p,p\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$p,p\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_PP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_XU:
     *        Second derivative or hessian of the ODEs towards \f$y,u\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$y,u\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_XU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_XP:
     *        Second derivative or hessian of the ODEs towards \f$y,p\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$y,p\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_XP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_UP:
     *        Second derivative or hessian of the ODEs towards \f$u,p\f$, multiplied by the
     *        lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the ODEs towards \f$u,p\f$, multiplied by the
     *                          lagrangian multipliers.
     *
     */
    virtual void L_DD_ODE_UP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_NLCSTR_XX:
     *        Second derivative or hessian of the nonlinear constraints towards \f$y,y\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,y\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_XX(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_UU:
     *        Second derivative or hessian of the nonlinear constraints towards \f$u,u\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$u,u\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_UU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_PP:
     *        Second derivative or hessian of the nonlinear constraints towards \f$p,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$p,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_PP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_XU:
     *        Second derivative or hessian of the nonlinear constraints towards \f$y,u\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,u\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_XU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_XP:
     *        Second derivative or hessian of the nonlinear constraints towards \f$y,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$y,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_XP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_UP:
     *        Second derivative or hessian of the nonlinear constraints towards \f$u,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t            Timepoint \f$t\f$.
     * \param[in]  y            States at \f$y(t)\f$.
     * \param[in]  u            Controls at \f$u(t)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the nonlinear constraints towards \f$u,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_NLCSTR_UP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_BOUND_XFXF:
     *        Second derivative or hessian of the boundary constraints towards \f$y_f,y_f\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,y_f\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_XFXF(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_X0X0:
     *        Second derivative or hessian of the boundary constraints towards \f$y_0,y_0\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_0,y_0\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_X0X0(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_PP:
     *        Second derivative or hessian of the boundary constraints towards \f$p,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$p,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_PP(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_XFX0:
     *        Second derivative or hessian of the boundary constraints towards \f$y_f,y_0\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,y_0\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_XFX0(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_XFP:
     *        Second derivative or hessian of the boundary constraints towards \f$y_f,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_f,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_XFP(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_X0P:
     *        Second derivative or hessian of the boundary constraints towards \f$y_0,p\f$,
     *        multiplied by the lagrangian multipliers.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[in]  lambda       Lagrangian multipliers \f$\lambda(t)\f$.
     * \param[out] S_OUT        Hessian of the boundary constraints towards \f$y_0,p\f$,
     *                          multiplied by the lagrangian multipliers.
     *
     */
    virtual void L_DD_BOUND_X0P(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_VARPHI_XFXF:
     *        Second derivative or hessian of the target fuction towards \f$y_f,y_f\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,y_f\f$.
     *
     */
    virtual void L_DD_VARPHI_XFXF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_X0X0:
     *        Second derivative or hessian of the target fuction towards \f$y_0,y_0\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$y_0,y_0\f$.
     *
     */
    virtual void L_DD_VARPHI_X0X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_PP:
     *        Second derivative or hessian of the target fuction towards \f$p,p\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$p,p\f$.
     *
     */
    virtual void L_DD_VARPHI_PP(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_XFX0:
     *        Second derivative or hessian of the target fuction towards \f$y_f,y_0\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,y_0\f$.
     *
     */
    virtual void L_DD_VARPHI_XFX0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_XFP:
     *        Second derivative or hessian of the target fuction towards \f$y_f,p\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$y_f,p\f$.
     *
     */
    virtual void L_DD_VARPHI_XFP(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_X0P:
     *        Second derivative or hessian of the target fuction towards \f$y_0,p\f$.
     *
     * \param[in]  t0           Start timepoint \f$t_0\f$.
     * \param[in]  tF           End timepoint \f$t_f\f$.
     * \param[in]  yF           States at \f$y(t_f)\f$.
     * \param[in]  y0           States at \f$y(t_0)\f$.
     * \param[in]  p            Parameters \f$p\f$.
     * \param[out] S_OUT        Hessian of the target fuction towards \f$y_0,p\f$.
     *
     */
    virtual void L_DD_VARPHI_X0P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);
};

#endif // MINIMUMENERGYPROBLEM_H
