/* (C) 2019 Andreas Huber, Uni BW München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file kinmodelocp.cpp
 * Implementierung der Klasse KinModelOCP.
 * Genauere Informationen sind in der Dokumentation der Klasse KinModelOCP enthalten.
 * 
 * @author A. Huber
 * @since 24.05.2019
 */

#include <iostream>
#include <math.h>
#include "kinmodelocp.h"


KinModelOCP::KinModelOCP(double l, double r, double c_s)
    : OCPbasic::Problem<double>(4, 2, 1, 1, 7)
    , l(l)
    , r(r)
    , scale_steering_obj(c_s)
{
    this->setRangeBoxBounds();
}

KinModelOCP::~KinModelOCP()
{
    
}

void KinModelOCP::setRangeBoxBounds()
{
    // Auf Dimensionen achten!

    // Nichtlineare Nebenbedingung:
    this->g_max[0] = 1.0000000000000000199E30;
    this->g_min[0] = 0.0;

    // Randbedingungen:
    this->psi_max[0] = 0.0;
    this->psi_min[0] = 0.0;
    this->psi_max[1] = 0.0;
    this->psi_min[1] = 0.0;
    this->psi_max[2] = 0.0;
    this->psi_min[2] = 0.0;
    this->psi_max[3] = 0.0;
    this->psi_min[3] = 0.0;
    this->psi_max[4] = 0.0;
    this->psi_min[4] = 0.0;
    this->psi_max[5] = 0.0;
    this->psi_min[5] = 0.0;
    this->psi_max[6] = 0.0;
    this->psi_min[6] = 0.0;

    // Zustände:
    this->y_max[0] = 1.0000000000000000199E30;
    this->y_min[0] = -1.0000000000000000199E30;
    this->y_max[1] = 1.0000000000000000199E30;
    this->y_min[1] = -0.10000000000000000555;
    this->y_max[2] = 1.0000000000000000199E30;
    this->y_min[2] = -1.0000000000000000199E30;
    this->y_max[3] = 1.0000000000000000199E30;
    this->y_min[3] = -1.0000000000000000199E30;

    // Steuerungen:
    this->u_max[0] = 0.5;
    this->u_min[0] = -0.5;
    this->u_max[1] = 3.0;
    this->u_min[1] = 0.0;

    // Parameter:
    this->p_max[0] = 5.0;
    this->p_min[0] = 0.050000000000000002776;
}

void KinModelOCP::ODE(double , double const y[], double const u[], double const p[], double f[]) const
{
    f[0] = u[1]*p[0]*cos(y[2]);
    f[1] = u[1]*p[0]*sin(y[2]);
    f[2] = tan(u[0])*u[1]/this->l*p[0];
    f[3] = (u[0]*u[0])*p[0];
}

void KinModelOCP::NLCSTR(double , double const y[], double const [], double const [], double gg[]) const
{
    gg[0] =  (y[1]*y[1])+(y[0]*y[0])-(this->r*this->r);
}

void KinModelOCP::BOUND(double , double , double const yF[], double const y0[], double const [], double psi[]) const
{
    psi[0] =  y0[0]+1.0000000000000000e+00;
    psi[1] =  yF[0]-1.0000000000000000e+00;
    psi[2] = y0[1];
    psi[3] = yF[1];
    psi[4] = y0[2];
    psi[5] = y0[3];
    psi[6] = yF[2];
}

void KinModelOCP::VARPHI(double , double , double const yF[], double const [], double const p[], double phi[]) const
{
    phi[0] =  yF[3]*this->scale_steering_obj+p[0];
}


//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                    Erste Ableitungen                                                                //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                             Rechte Seite der DGL                                                    //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::D_ODE_X(double , double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol88 = -u[1]*p[0]*sin(y[2]);
    S_OUT(0, 2) = symbol88;
    double symbol89 = u[1]*p[0]*cos(y[2]);
    S_OUT(1, 2) = symbol89;
}

void KinModelOCP::D_ODE_U(double , double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol90 = p[0]*cos(y[2]);
    S_OUT(0, 1) = symbol90;
    double symbol91 = p[0]*sin(y[2]);
    S_OUT(1, 1) = symbol91;
    double symbol92 = u[1]*( pow(tan(u[0]),2.0)+1.0)/this->l*p[0];
    S_OUT(2, 0) = symbol92;
    double symbol93 = tan(u[0])/this->l*p[0];
    S_OUT(2, 1) = symbol93;
    double symbol94 = 2.0*u[0]*p[0];
    S_OUT(3, 0) = symbol94;
}

void KinModelOCP::D_ODE_P(double , double const y[], double const u[], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol95 = u[1]*cos(y[2]);
    S_OUT(0, 0) = symbol95;
    double symbol96 = u[1]*sin(y[2]);
    S_OUT(1, 0) = symbol96;
    double symbol97 = tan(u[0])*u[1]/this->l;
    S_OUT(2, 0) = symbol97;
    double symbol98 = (u[0]*u[0]);
    S_OUT(3, 0) = symbol98;
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                        Nichtlineare Nebenbedingungen                                                //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::D_NLCSTR_X(double , double const y[], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol99 = 2.0*y[0];
    S_OUT(0, 0) = symbol99;
    double symbol100 = 2.0*y[1];
    S_OUT(0, 1) = symbol100;
}

void KinModelOCP::D_NLCSTR_U(double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::D_NLCSTR_P(double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                               Randbedingungen                                                       //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::D_BOUND_XF(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol101 = 1.0;
    S_OUT(1, 0) = symbol101;
    double symbol102 = 1.0;
    S_OUT(3, 1) = symbol102;
    double symbol103 = 1.0;
    S_OUT(6, 2) = symbol103;
}

void KinModelOCP::D_BOUND_X0(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol104 = 1.0;
    S_OUT(0, 0) = symbol104;
    double symbol105 = 1.0;
    S_OUT(2, 1) = symbol105;
    double symbol106 = 1.0;
    S_OUT(4, 2) = symbol106;
    double symbol107 = 1.0;
    S_OUT(5, 3) = symbol107;
}

void KinModelOCP::D_BOUND_P(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                                 Zielfunktion                                                        //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::D_VARPHI_XF(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol108 = this->scale_steering_obj;
    S_OUT(0, 3) = symbol108;
}

void KinModelOCP::D_VARPHI_X0(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::D_VARPHI_P(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol109 = 1.0;
    S_OUT(0, 0) = symbol109;
}


//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                    Zweite Ableitungen                                                               //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                             Rechte Seite der DGL                                                    //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::L_DD_ODE_XX(double , double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol110 = -u[1]*p[0]*sin(y[2])*lambda[1]-u[1]*lambda[0]*p[0]*cos(y[2]);
    S_OUT(2, 2) = symbol110;
}

void KinModelOCP::L_DD_ODE_UU(double , double const [], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol111 =  2.0*lambda[3]*p[0]+2.0*lambda[2]*tan(u[0])*u[1]*( pow(tan(u[0]),2.0)+1.0)/this->l*p[0];
    S_OUT(0, 0) = symbol111;
    double symbol112 = lambda[2]*( pow(tan(u[0]),2.0)+1.0)/this->l*p[0];
    S_OUT(0, 1) = symbol112;
    double symbol113 = lambda[2]*( pow(tan(u[0]),2.0)+1.0)/this->l*p[0];
    S_OUT(1, 0) = symbol113;
}

void KinModelOCP::L_DD_ODE_PP(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_ODE_XU(double , double const y[], double const [], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol114 = -lambda[0]*p[0]*sin(y[2])+p[0]*lambda[1]*cos(y[2]);
    S_OUT(2, 1) = symbol114;
}

void KinModelOCP::L_DD_ODE_XP(double , double const y[], double const u[], double const [], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol115 =  u[1]*lambda[1]*cos(y[2])-u[1]*lambda[0]*sin(y[2]);
    S_OUT(2, 0) = symbol115;
}

void KinModelOCP::L_DD_ODE_UP(double , double const y[], double const u[], double const [], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol116 =  lambda[2]*u[1]*( pow(tan(u[0]),2.0)+1.0)/this->l+2.0*u[0]*lambda[3];
    S_OUT(0, 0) = symbol116;
    double symbol117 =  lambda[0]*cos(y[2])+sin(y[2])*lambda[1]+lambda[2]*tan(u[0])/this->l;
    S_OUT(1, 0) = symbol117;
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                        Nichtlineare Nebenbedingungen                                                //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::L_DD_NLCSTR_XX(double , double const [], double const [], double const [], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
    double symbol118 = 2.0*lambda[0];
    S_OUT(0, 0) = symbol118;
    double symbol119 = 2.0*lambda[0];
    S_OUT(1, 1) = symbol119;
}

void KinModelOCP::L_DD_NLCSTR_UU(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_NLCSTR_PP(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_NLCSTR_XU(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_NLCSTR_XP(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_NLCSTR_UP(double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                               Randbedingungen                                                       //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::L_DD_BOUND_XFXF(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_BOUND_X0X0(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_BOUND_PP(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_BOUND_XFX0(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_BOUND_XFP(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_BOUND_X0P(double , double , double const [], double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

//---------------------------------------------------------------------------------------------------------------------//
//                                                                                                                     //
//                                                 Zielfunktion                                                        //
//                                                                                                                     //
//---------------------------------------------------------------------------------------------------------------------//

void KinModelOCP::L_DD_VARPHI_XFXF(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_VARPHI_X0X0(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_VARPHI_PP(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_VARPHI_XFX0(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_VARPHI_XFP(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

void KinModelOCP::L_DD_VARPHI_X0P(double , double , double const [], double const [], double const [], OCPbasic::MatrixWrapper<double> & S_OUT)
{
    S_OUT.zero_fill();
}

