/* (C) 2019 Andreas Huber, Uni BW München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file kinmodelocp.h
 * Header Definition der Klasse KinModelOCP.
 * Genauere Informationen sind in der Dokumentation der Klasse KinModelOCP enthalten.
 * 
 * @author A. Huber
 * @since 24.05.2019
 */

#ifndef KINMODELOCP_H
#define KINMODELOCP_H
#include <OCPbasic/problem.h>

class KinModelOCP : public OCPbasic::Problem<double>
{
    double l;

    double r;

    double scale_steering_obj;
public:

    /**
     * Konstruktor KinModelOCP.
     */
    explicit KinModelOCP(double l, double r, double c_s);

    /**
     * Destruktor KinModelOCP.
     */
    virtual ~KinModelOCP();

    /**
     * \brief setRangeBoxBounds:
     *        Initialisiert und belegt die Schranken der verschiedenen Ungleichungen.
     */
    void setRangeBoxBounds();

    /**
     * \brief ODE:
     *        Rechte Seite der DGL als VektorFunktion.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[out] f      Zurückgegebenes Array mit Werten der VektorFunktion.
     *
     */
    void ODE(double t, double const y[], double const u[], double const p[], double f[]) const;

    /**
     * \brief NLCSTR:
     *        Nichtlineare Nebenbedingungen.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[out] gg     Zurückgegebenes Array mit Werten der nichtlinearen Nebenbedingungen.
     *
     */
    void NLCSTR(double t, double const y[], double const u[], double const p[], double gg[]) const;

    /**
     * \brief BOUND:
     *        Randbedingungen.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[out] psi    Zurückgegebenes Array mit Werten der Randbedingungen.
     *
     */
    void BOUND(double t0, double tF, double const yF[], double const y0[], double const p[], double psi[]) const;

    /**
     * \brief VARPHI:
     *        Zielfunktion.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[out] phi    Zurückgegebenes Array mit Werten der Zielfunktion.
     *
     */
    void VARPHI(double t0, double tF, double const yF[], double const y0[], double const p[], double phi[]) const;

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Erste Ableitungen                                                                //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_ODE_X:
     *        Erste Ableitung (Jacobimatrix) der rechten Seite der DGL nach y.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von ODE nach den Zuständen y.
     *
     */
    void D_ODE_X(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_ODE_U:
     *        Erste Ableitung (Jacobimatrix) der rechten Seite der DGL nach u.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von ODE nach den Steuerungen u.
     *
     */
    void D_ODE_U(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_ODE_P:
     *        Erste Ableitung (Jacobimatrix) der rechten Seite der DGL nach p.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von ODE nach den Parametern p.
     *
     */
    void D_ODE_P(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_NLCSTR_X:
     *        Erste Ableitung (Jacobimatrix) der nichtlineare Nebenbedingungen nach y.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von NLCSTR nach den Zuständen y.
     *
     */
    void D_NLCSTR_X(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_NLCSTR_U:
     *        Erste Ableitung (Jacobimatrix) der nichtlineare Nebenbedingungen nach u.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von NLCSTR nach den Steuerungen u.
     *
     */
    void D_NLCSTR_U(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_NLCSTR_P:
     *        Erste Ableitung (Jacobimatrix) der nichtlineare Nebenbedingungen der DGL nach p.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von NLCSTR nach den Parametern p.
     *
     */
    void D_NLCSTR_P(double t, double const y[], double const u[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_BOUND_XF:
     *        Erste Ableitung (Jacobimatrix) der Randbedingungen nach yf.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von BOUND nach den Zuständen yf.
     *
     */
    void D_BOUND_XF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_BOUND_X0:
     *        Erste Ableitung (Jacobimatrix) der Randbedingungen nach y0.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von BOUND nach den Zuständen y0.
     *
     */
    void D_BOUND_X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_BOUND_P:
     *        Erste Ableitung (Jacobimatrix) der Randbedingungen nach p.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von BOUND nach den Parametern p.
     *
     */
    void D_BOUND_P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief D_VARPHI_XF:
     *        Erste Ableitung (Jacobimatrix) der Zielfunktion nach yf.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von VARPHI nach den Zuständen yf.
     *
     */
    void D_VARPHI_XF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_VARPHI_X0:
     *        Erste Ableitung (Jacobimatrix) der Zielfunktion nach y0.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von VARPHI nach den Zuständen y0.
     *
     */
    void D_VARPHI_X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief D_VARPHI_P:
     *        Erste Ableitung (Jacobimatrix) der Zielfunktion nach p.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Jacobimatrix von VARPHI nach den Parametern p.
     *
     */
    void D_VARPHI_P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);


    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Zweite Ableitungen                                                               //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_ODE_UU:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach x und x, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach x und x, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_XX(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_UU:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach u und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach u und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_UU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_PP:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_PP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_XU:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach y und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach y und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_XU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_XP:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach y und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach y und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_XP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_ODE_UP:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach u und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von ODE nach u und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_ODE_UP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_NLCSTR_XX:
     *        Zweite Ableitung (Hessematrix 3D) der nichtlineare Nebenbedingungen nach x und x, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach x und x, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_XX(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_UU:
     *        Zweite Ableitung (Hessematrix 3D) der nichtlineare Nebenbedingungen nach u und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach u und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_UU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_PP:
     *        Zweite Ableitung (Hessematrix 3D) der nichtlineare Nebenbedingungen nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_PP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_XU:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach x und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach x und u, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_XU(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_XP:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach x und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach x und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_XP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_NLCSTR_UP:
     *        Zweite Ableitung (Hessematrix 3D) der rechten Seite der DGL nach u und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t      Zeitpunkt t.
     * \param[in]  y      Vektor-Array der Stützstelle.
     * \param[in]  u      Steuerungs-Array der Stützstelle.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von NLCSTR nach u und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_NLCSTR_UP(double t, double const y[], double const u[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_BOUND_XFXF:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach yF und yF, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach yF und yF, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_XFXF(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_X0X0:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach y0 und y0, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach y0 und y0, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_X0X0(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_PP:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach p und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_PP(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_XFX0:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach yF und y0, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach yF und y0, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_XFX0(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_XFP:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach yF und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach yF und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_XFP(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_BOUND_X0P:
     *        Zweite Ableitung (Hessematrix 3D) der Randbedingungen nach y0 und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     * \param[in]  lambda Lagrange-Multiplikatoren.
     *
     * @return Hessematrix-3D von BOUND nach y0 und p, mulitpliziert mit den Lagrange-Multiplikatoren lambda (=> SparseMatrix).
     *
     */
    void L_DD_BOUND_X0P(double t0, double tF, double const yF[], double const y0[], double const p[], double const lambda[], OCPbasic::MatrixWrapper<double> & S_OUT);

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    /**
     * \brief L_DD_VARPHI_XFXF:
     *        Zweite Ableitung der Zielfunktion nach yF und yF.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach yF und yF.
     *
     */
    void L_DD_VARPHI_XFXF(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_X0X0:
     *        Zweite Ableitung der Zielfunktion nach y0 und y0.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach y0 und y0.
     *
     */
    void L_DD_VARPHI_X0X0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_PP:
     *        Zweite Ableitung der Zielfunktion nach p und p.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach p und p.
     *
     */
    void L_DD_VARPHI_PP(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_XFX0:
     *        Zweite Ableitung der Zielfunktion nach yF und y0.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach yF und y0.
     *
     */
    void L_DD_VARPHI_XFX0(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_XFP:
     *        Zweite Ableitung der Zielfunktion nach yF und p.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach yF und p.
     *
     */
    void L_DD_VARPHI_XFP(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);

    /**
     * \brief L_DD_VARPHI_X0P:
     *        Zweite Ableitung der Zielfunktion nach y0 und p.
     *
     * \param[in]  t0     Anfangszeitpunkt.
     * \param[in]  tF     Endzeitpunkt.
     * \param[in]  yF     Vektor-Array der Stützstelle zum Endzeitpunkt.
     * \param[in]  y0     Vektor-Array der Stützstelle zum Anfangszeitpunkt.
     * \param[in]  p      Parameter-Array der Stützstelle.
     *
     * @return Hessematrix von VARPHI nach y0 und p.
     *
     */
    void L_DD_VARPHI_X0P(double t0, double tF, double const yF[], double const y0[], double const p[], OCPbasic::MatrixWrapper<double> & S_OUT);
};

#endif // KINMODELOCP_H

