/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*!
 * \file IPbasic_Data.cpp
 * Implementation of the class IPdata.
 *
 * \author A. Huber
 * \since 05.02.2019
 */

#include "IPbasic_Data.h"
#include "IPbasic.h"

namespace IPbasic
{

    template <typename real>
    void IterationDataPoint<real>::printHeader(ostream_type & out)
    {
        if (this->print > PRINT_NONE)
        {
            this->printLine(out);
            out << "|"
                << "iter" << '\t' << "mu         " << '\t' << "obj        " << '\t' << "d_kkt      " << '\t' << "d_sc_kkt   " << '\t' << "d_constr   " << '\t' << "alpha      " << '\t' << "alpha_s    "
                << '\t' << "alpha_l "
                << "    "
                << "|" << std::endl;
            this->printLine(out);
        }
    }

    template <typename real>
    void IterationDataPoint<real>::printLine(ostream_type & out)
    {
        if (this->print > PRINT_NONE)
        {
            out << "|-------------------------------------------"
                << "--------------------------------------------"
                << "--------------------------------------------|" << std::endl;
        }
    }

    template <typename real>
    ostream_type & operator<<(ostream_type & out, IterationDataPoint<real> const & a)
    {
        out << std::fixed << "|" << a.counter;
        if (a.print >= PRINT_MARKED)
        {
            real one = real(1.0);
            TermColor::set3color(a.mu, one, a.tolMu);
            out << std::scientific << '\t' << a.mu;
            TermColor::set_COLOR_DEFAULT();
            out << '\t' << a.obj << '\t';
            TermColor::set3color(a.Delta_KKT, one, a.tolKKT);
            out << a.Delta_KKT;
            TermColor::set_COLOR_DEFAULT();
            out << '\t';
            TermColor::set3color(a.scaled_Delta_KKT, one, a.tolKKT);
            out << a.scaled_Delta_KKT;
            TermColor::set_COLOR_DEFAULT();
            out << '\t';
            TermColor::set3color(a.Delta_CSTR, one, a.tolKKT);
            out << a.Delta_CSTR;
            TermColor::set_COLOR_DEFAULT();
            out << '\t';
            if (a.counter > 0)
            {
                TermColor::set3colorR(a.alpha, one, a.tolAlpha);
                out << a.alpha;
                TermColor::set_COLOR_DEFAULT();
            }
            else
            {
                out << "-" << '\t';
            }
            out << '\t';
            if (a.counter > 0)
            {
                TermColor::set3colorR(a.alpha_s, one, a.tolAlpha);
                out << a.alpha_s;
                TermColor::set_COLOR_DEFAULT();
            }
            else
            {
                out << "-" << '\t';
            }
            out << '\t';
            if (a.counter > 0)
            {
                TermColor::set3colorR(a.alpha_l, one, a.tolAlpha);
                out << a.alpha_l;
                TermColor::set_COLOR_DEFAULT();
            }
            else
            {
                out << "-"
                    << "           ";
            }
            out << "|";
        }
        else if (a.print == PRINT_MORE)
        {
            // TermColor::set_COLOR_DEFAULT();
            out << std::scientific << '\t' << a.mu << '\t' << a.obj << '\t' << a.Delta_KKT << '\t' << a.scaled_Delta_KKT << '\t' << a.Delta_CSTR << '\t';
            if (a.counter > 0)
                out << a.alpha;
            else
                out << "-" << '\t';
            out << '\t';
            if (a.counter > 0)
                out << a.alpha_s;
            else
                out << "-" << '\t';
            out << '\t';
            if (a.counter > 0)
                out << a.alpha_l;
            else
                out << "-"
                    << "           ";
            out << "|";
        }
        out << std::fixed;
        return out;
    }

    template <typename real>
    IPdata<real>::IPdata(NLP<real> * _Problem, IPparameter<real> * IPParams, IPtimings<real> * Timings)
        : mem_real("IPdata")
        , Problem(_Problem)
    {
        // Set default flags:
        this->has_initial_guess_slack  = false;
        this->has_initial_guess_lambda = false;
        // Set data:
        this->sigma = real(1);
        // Set IP Parameter:
        if (IPParams == nullptr)
        {
            this->Parameter         = new IPparameter<real>();
            this->allocedParameters = true;
        }
        else
        {
            this->Parameter         = IPParams;
            this->allocedParameters = false;
        }
        // Set IP Timing:
        if (Timings == nullptr)
        {
            this->Timings        = new IPtimings<real>();
            this->allocedTimings = true;
        }
        else
        {
            this->Timings        = Timings;
            this->allocedTimings = false;
        }

        // Get Dimensions:
        this->DimStates = this->Problem->state_dimension();
        this->DimLambda = this->Problem->constraint_dimension();

        mem_real.allocate(size_t(3 * this->DimStates + 6 * this->DimLambda));

        // Allocate Data:
        this->states            = mem_real(size_t(this->DimStates));
        this->lambda            = mem_real(size_t(this->DimLambda));
        this->slack             = mem_real(size_t(this->DimLambda));
        this->ConstraintsVector = mem_real(size_t(this->DimLambda));
        this->grad_objective    = mem_real(size_t(this->DimStates));
        this->is_equality_vec.resize(size_t(this->DimLambda));

        // Get Pointer to SparseData:
        this->constraint_jacobian = this->Problem->get_constraint_jacobian();
        this->lagrangian_hessian  = this->Problem->get_lagrangian_hessian();

        // Init IPdata:
        this->initDataFromProblem();

        // Allocate vectors of the searchdirection:
        this->IP_d_states = mem_real(size_t(this->DimStates));
        this->IP_d_slack  = mem_real(size_t(this->DimLambda));
        this->IP_d_lambda = mem_real(size_t(this->DimLambda));
        this->mu_sequ     = mem_real(size_t(this->DimLambda));
        this->mu_max      = this->Parameter->mu_0;

        // Set searchdirections to initial value:
        for (int_type i = 0; i < this->DimStates; ++i)
            this->IP_d_states[i] = 0;
        for (int_type i = 0; i < this->DimLambda; ++i)
            this->IP_d_slack[i] = this->IP_d_lambda[i] = 0;
    }

    template <typename real>
    IPdata<real>::~IPdata()
    {
        mem_real.free();
        this->is_equality_vec.clear();
        if (this->allocedParameters)
            delete this->Parameter;
        if (this->allocedTimings)
            delete this->Timings;
    }

    /*\
    :|:             _    _ _                      _
    :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _
    :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_|
    :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_|
    :|:  |_|
    \*/

    template <typename real>
    void IPdata<real>::initDataFromProblem()
    {
        // Copy Initial guess and data:
        this->Problem->initial_guess(this->states);
        this->Problem->constraint(this->states, this->ConstraintsVector);

        // Set equality_vec:
        for (size_t i = 0; i < size_t(this->DimLambda); ++i)
            this->is_equality_vec[i] = this->Problem->is_equality(int_type(i));

        // Set initial guess slack:
        if (!this->Problem->initial_guess_slack(this->slack))
        {
            this->has_initial_guess_slack = false;
            for (size_t i = 0; i < size_t(this->DimLambda); ++i)
            {
                if (this->ConstraintsVector[i] < 0)
                    this->slack[i] = -this->ConstraintsVector[i];
                else
                    this->slack[i] = real(1.0);
            }
        }
        else
        {
            this->has_initial_guess_slack = true;
        }

        // Set initial guess lambda:
        if (!this->Problem->initial_guess_lambda(this->lambda))
        {
            this->has_initial_guess_lambda = false;
            for (size_t i = 0; i < size_t(this->DimLambda); ++i)
            {
                // This is later set by other method:
                this->lambda[i] = this->Parameter->FailSaveLagrangeInitialValue;
            }
        }
        else
        {
            this->has_initial_guess_lambda = true;
        }

        // Set iteration data:
        this->IterData     = IterationDataPoint<real>(this->Parameter->mu_tol, this->Parameter->epsilon_tol, this->Parameter->tol_alpha, this->Parameter->PrintLev);
        this->IterData.mu  = this->Parameter->mu_0;
        this->IterData.obj = this->Problem->objective(this->states);
    }

    template <typename real>
    NLP<real> * IPdata<real>::getProblem() const
    {
        return this->Problem;
    }

    template <typename real>
    IPtimings<real> * IPdata<real>::getTimings() const
    {
        return this->Timings;
    }

    template <typename real>
    IPparameter<real> * IPdata<real>::getIPparameter() const
    {
        return this->Parameter;
    }

    template <typename real>
    bool IPdata<real>::checkForValidMembers() const
    {
        if (nullptr == this->Problem)
            return false;
        if (!this->Problem->checkValidMembers())
            return false;
        if (nullptr == this->Parameter)
            return false;
        if (nullptr == this->constraint_jacobian)
            return false;
        if (nullptr == this->lagrangian_hessian)
            return false;
        return true;
    }

    template <typename real>
    void IPdata<real>::updateData()
    {
        // Evaluate jacobian:
        this->LocalDataTimer.tic();
        this->Problem->constraint_jacobian_update(this->states);
        this->LocalDataTimer.toc();
        this->Timings->TimeJacobianUpdate += this->LocalDataTimer.elapsed_s();
        // Evaluate hessian:
        this->LocalDataTimer.tic();
        this->Problem->lagrangian_hessian_update(this->sigma, this->states, this->lambda);
        this->LocalDataTimer.toc();
        this->Timings->TimeHessianUpdate += this->LocalDataTimer.elapsed_s();
        // Evaluate constraints:
        this->Problem->constraint(this->states, this->ConstraintsVector);
        // Evaluate target function:
        this->IterData.obj = this->Problem->objective(this->states);
        // Evaluate gradient of the target function:
        this->Problem->objective_gradient(this->states, this->grad_objective);
    }

    template <typename real>
    real IPdata<real>::get_Delta_d() const
    {
        real delMax = this->Parameter->delta_max;
        real hlp    = 0;
        for (size_t i = 0; i < size_t(this->DimLambda); ++i)
            hlp += std::abs(this->lambda[i]);
        hlp /= this->DimLambda;
        real delta_d = std::max(delMax, hlp) / delMax;
        return delta_d;
    }

    template <typename real>
    real IPdata<real>::get_Delta_c() const
    {
        real     delMax = this->Parameter->delta_max;
        real     hlp    = 0;
        int_type count  = 0;
        for (size_t i = 0; i < size_t(this->DimLambda); ++i)
        {
            if (!this->is_equality_vec[i])
            {
                ++count;
                hlp += std::abs(this->lambda[i]);
            }
        }
        hlp /= count;
        real delta_c = std::max(delMax, hlp) / delMax;
        return delta_c;
    }

    template <typename real>
    bool IPdata<real>::check_INF_NAN() const
    {
        // Check iterdata:
        if (this->IterData.check_INF_NAN())
            return true;
        // Check vectors:
        for (size_t i = 0; i < size_t(this->DimStates); ++i)
        {
            if (!isRegular(this->states[i]))
                return true;
            if (!isRegular(this->IP_d_states[i]))
                return true;
        }
        for (size_t i = 0; i < size_t(this->DimLambda); ++i)
        {
            if (!isRegular(this->lambda[i]))
                return true;
            if (!isRegular(this->slack[i]))
                return true;
            if (!isRegular(this->IP_d_lambda[i]))
                return true;
            if (!isRegular(this->IP_d_slack[i]))
                return true;
        }

        // Check matrices:
        if (this->Parameter->CheckMatricesFor_INF_NAN)
        {
            // Check jacobian:
            if (this->constraint_jacobian->foundNaN())
                return true;
            // Check hessian:
            if (this->lagrangian_hessian->foundNaN())
                return true;
        }
        // If checks passed function returns true:
        return false;
    }

    template <typename real>
    real IPdata<real>::getTau(int_type index) const
    {
        if (this->Parameter->UseMuSequ)
        {
            real tau_min = this->Parameter->tau_min;
            real tau     = real(1.0) - this->mu_sequ[index];
            return std::max(tau_min, tau);
        }
        else
        {
            real tau_min = this->Parameter->tau_min;
            real mu      = this->IterData.mu;
            real tau     = real(1.0) - mu;
            return std::max(tau_min, tau);
        }
    }

    template <typename real>
    void IPdata<real>::updateMuSequence() const
    {
        // Get current mu:
        real mu = this->IterData.mu;
        // Default fill with mu:
        // std::fill(this->mu_sequ, this->mu_sequ + this->DimLambda, mu);
        // Manipulate mu:
        for (size_t i = 0; i < size_t(this->DimLambda); ++i)
        {
            if (this->is_equality_vec[i])
            {
                this->mu_sequ[i] = 0.0;
            }
            else
            {
                if (this->slack[i] < this->mu_sequ[i] * 100)
                {
                    if ((-this->IP_d_slack[i]) > this->mu_sequ[i] * 100)
                    {
                        this->mu_sequ[i] = this->mu_sequ[i] * 100.0;
                        // std::cout << this->mu_sequ[i] << std::endl;
                    }
                    else
                    {
                        this->mu_sequ[i] = mu;
                    }
                }
                else
                {
                    this->mu_sequ[i] = mu;
                }
            }
        }
    }

    template class IterationDataPoint<float>;
    template class IterationDataPoint<double>;

    template class IPdata<float>;
    template class IPdata<double>;

    template ostream_type & operator<<(ostream_type & out, IterationDataPoint<float> const & a);

    template ostream_type & operator<<(ostream_type & out, IterationDataPoint<double> const & a);

} // namespace IPbasic
