/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*!
 * \file IPbasic_KKTSystem.cpp
 * Implementation of the class KKTSystem.
 *
 * \author A. Huber
 * \since 05.02.2019
 */

#include "IPbasic_KKTSystem.h"
#include "Errors/IPbasic_Errors.h"
#include "IPbasic.h"

namespace IPbasic
{

    template <typename real>
    KKTSystem<real>::KKTSystem(IPdata<real> const * IPD, LinearSolver_interface<real> * _LinSolver)
        : mem_real("KKTSystem")
        , LinSolver(_LinSolver)
    {
        // Get dimensions:
        int_type DimStates = IPD->DimStates;
        int_type DimLambda = IPD->DimLambda;
        int_type DimRHS    = DimStates + DimLambda;
        int_type DimKKT    = DimStates + 2 * DimLambda;
        // Allocate IPMatrix:
        this->IPmatrix = IPD->getProblem()->get_IPMatrix();
        // Allocate vector of the right-hand-side:
        mem_real.allocate(size_t(DimRHS + DimKKT));
        this->RHS = mem_real(size_t(DimRHS));
        this->KKT = mem_real(size_t(DimKKT));
    }

    template <typename real>
    KKTSystem<real>::~KKTSystem()
    {
        mem_real.free();
    }

    /*\
    :|:             _    _ _                      _
    :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    bool KKTSystem<real>::updateSearchDirection(IPdata<real> const * IPD)
    {
        // Get dimensions:
        int_type          DimStates                  = IPD->DimStates;
        int_type          DimLambda                  = IPD->DimLambda;
        int_type          DimRHS                     = DimStates + DimLambda;
        IPtimings<real> * MyIPtimings                = IPD->getTimings();
        bool              Flag_IPmatrix_is_symmetric = IPD->getProblem()->is_IPmatrix_symmetric();

        this->getRHSVector(IPD, this->RHS);
        this->getIPMatrix(IPD, this->IPmatrix);

        for (size_t i = 0; i < size_t(DimRHS); ++i)
            this->RHS[i] *= real(-1);
        /* For testing linear system:
        if (IPD->IterData.counter == 0)
            this->LinSolver->printSystemToFiles(*this->IPmatrix, RHS, "IPMatrix.csv", "Vectors.csv");
        */
        this->LocalTimerKKT.tic();
        if (!this->LinSolver->factorize(*this->IPmatrix, Flag_IPmatrix_is_symmetric))
        {
            this->LocalTimerKKT.toc();
            MyIPtimings->TimeFactorize += this->LocalTimerKKT.elapsed_s();
            return false;
        }
        this->LocalTimerKKT.toc();
        MyIPtimings->TimeFactorize += this->LocalTimerKKT.elapsed_s();
        this->LocalTimerKKT.tic();
        if (!this->LinSolver->solve(this->RHS))
        {
            this->LocalTimerKKT.toc();
            MyIPtimings->TimeSolve += this->LocalTimerKKT.elapsed_s();
            return false;
        }
        this->LocalTimerKKT.toc();
        MyIPtimings->TimeSolve += this->LocalTimerKKT.elapsed_s();

        // Copy search direction (states and lambdas):
        std::copy_n(&this->RHS[0], DimStates, IPD->IP_d_states);
        std::copy_n(&this->RHS[DimStates], DimLambda, IPD->IP_d_lambda);

        // Get search direction for the slacks:
        real * d_slack = IPD->IP_d_slack;
        // Set to zero:
        for (size_t i = 0; i < size_t(DimLambda); ++i)
            d_slack[i] = real(0.0);

        vec_bool const &               is_equ = IPD->is_equality_vec;
        SparseMatrixBase<real> const * Jacobi = IPD->constraint_jacobian;
        // Get -\Nabla g * d_x:
        Jacobi->gemv(-1.0, DimStates, IPD->IP_d_states, 1, 1.0, DimLambda, d_slack, 1);
        real * constr = IPD->ConstraintsVector;
        real * slack  = IPD->slack;

        // Get -\Nabla g * d_x - (g + s):
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            if (is_equ[i])
                d_slack[i] = real(0.0);
            else
                d_slack[i] -= (constr[i] + slack[i]);
        }

        return true;
    }

    template <typename real>
    void KKTSystem<real>::getKKTErrors(IPdata<real> const * IPD, real & DelKKT, real & ScaledDelKKT, real & DelConstr)
    {
        this->getKKTVector(IPD, this->KKT);
        real del_d = IPD->get_Delta_d();
        real del_c = IPD->get_Delta_c();

        real Del_Nab_x_Lagr = absmax(IPD->DimStates, &this->KKT[0], 1);

        real Del_mu_s_lambda = absmax(IPD->DimLambda, &this->KKT[IPD->DimStates], 1);

        DelConstr = absmax(IPD->DimLambda, &this->KKT[IPD->DimStates + IPD->DimLambda], 1);

        real hlp     = std::max(Del_Nab_x_Lagr, Del_mu_s_lambda);
        DelKKT       = std::max(hlp, DelConstr);
        hlp          = std::max(Del_Nab_x_Lagr / del_d, Del_mu_s_lambda / del_c);
        ScaledDelKKT = std::max(hlp, DelConstr);
    }

    template <typename real>
    bool KKTSystem<real>::CreateInitialGuessLambda(IPdata<real> const * IPD)
    {
        bool Flag_IPmatrix_is_symmetric = IPD->getProblem()->is_IPmatrix_symmetric();
        // Get dimensions and lagrange multipliers:
        int_type          DimStates   = IPD->DimStates;
        int_type          DimLambda   = IPD->DimLambda;
        real *            lambda      = IPD->lambda;
        vec_bool const &  is_equ      = IPD->is_equality_vec;
        IPtimings<real> * MyIPtimings = IPD->getTimings();
        // Setup RHS:
        this->getInitialGuessLambdaRHS(IPD, this->RHS);
        // Setup IGLambda matrix:
        this->getInitialGuessLambdaMatrix(IPD, this->IPmatrix);

        // this->LinSolver->printSystemToFiles(*this->IPmatrix, this->RHS, "IGLambda.txt", "Vec.txt");
        this->LocalTimerKKT.tic();
        bool ret = this->LinSolver->factorize(*this->IPmatrix, Flag_IPmatrix_is_symmetric);
        this->LocalTimerKKT.toc();
        MyIPtimings->TimeFactorize += this->LocalTimerKKT.elapsed_s();
        if (ret)
        {
            this->LocalTimerKKT.tic();
            ret = this->LinSolver->solve(this->RHS);
            this->LocalTimerKKT.toc();
            MyIPtimings->TimeSolve += this->LocalTimerKKT.elapsed_s();
        }

        // Save Initial guess to lambda:
        if (ret)
        {
            for (size_t i = 0; i < size_t(DimLambda); ++i)
            {
                lambda[i] = this->RHS[size_t(DimStates) + i];
                if (!is_equ.at(i) && lambda[i] <= real(0.0))
                {
                    // Errors::set_Warning("Initial guess for lagrange multipliers < 0. Set to 1!");
                    lambda[i] = IPD->getIPparameter()->FailSaveLagrangeInitialValue;
                }
            }
        }
        else
        {
            Errors::set_Warning("Could not create initial guess for lagrange multipliers!");
            std::fill(lambda, lambda + DimLambda, IPD->getIPparameter()->FailSaveLagrangeInitialValue);
        }

        return ret;
    }

    template <typename real>
    bool KKTSystem<real>::checkForValidMembers()
    {
        if (nullptr == this->LinSolver)
            return false;
        return true;
    }

    /*\
    :|:                _          _          _                   _
    :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| |  _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/_| \___/\__\___\__|\__\___\__,_| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    void KKTSystem<real>::getKKTVector(IPdata<real> const * IPD, real KKT_vec[])
    {
        // Barrier parameter mu:
        real   mu      = IPD->IterData.mu;
        real * mu_sequ = IPD->mu_sequ;
        // Copy gradient of the target function to KKT_vec:
        std::copy_n(IPD->grad_objective, IPD->DimStates, KKT_vec);
        // Add Jacobi^\top*lambda to the gradient of the target function:
        IPD->constraint_jacobian->gemv_Transposed(1.0, IPD->DimLambda, IPD->lambda, 1, 1.0, IPD->DimStates, KKT_vec, 1);

        size_t offs  = size_t(IPD->DimStates);
        size_t offs2 = size_t(IPD->DimStates + IPD->DimLambda);
        for (size_t i = 0; i < size_t(IPD->DimLambda); ++i)
        {
            if (IPD->is_equality_vec[i])
            {
                // Add 0 = 0:
                KKT_vec[offs + i] = 0.0;
                // Add Constraints to KKT_vec:
                KKT_vec[offs2 + i] = IPD->ConstraintsVector[i];
            }
            else
            {
                if (IPD->getIPparameter()->UseMuSequ)
                {
                    // Add -mu * e + slack*lambda = 0:
                    KKT_vec[offs + i] = -mu_sequ[i] + IPD->slack[i] * IPD->lambda[i];
                }
                else
                {
                    // Add -mu * e + slack*lambda = 0:
                    KKT_vec[offs + i] = -mu + IPD->slack[i] * IPD->lambda[i];
                }
                // Add Constraints to KKT_vec:
                KKT_vec[offs2 + i] = IPD->ConstraintsVector[i] + IPD->slack[i];
            }
        }
    }

    template <typename real>
    void KKTSystem<real>::getRHSVector(IPdata<real> const * IPD, real RHS_vec[])
    {
        // Barrier parameter mu:
        real   mu      = IPD->IterData.mu;
        real * mu_sequ = IPD->mu_sequ;
        // Copy gradient of the target function to KKT:
        std::copy_n(IPD->grad_objective, IPD->DimStates, RHS_vec);
        // Add Jacobi^\top*lambda to the gradient of the target function:
        IPD->constraint_jacobian->gemv_Transposed(1.0, IPD->DimLambda, IPD->lambda, 1, 1.0, IPD->DimStates, RHS_vec, 1);

        // Add Constraints to KKT:
        size_t offs = size_t(IPD->DimStates);
        for (size_t i = 0; i < size_t(IPD->DimLambda); ++i)
        {
            if (IPD->is_equality_vec[i])
            {
                RHS_vec[offs + i] = IPD->ConstraintsVector[i];
            }
            else
            {
                if (IPD->getIPparameter()->UseMuSequ)
                {
                    RHS_vec[offs + i] = (mu_sequ[i] / (IPD->lambda[i])) + IPD->ConstraintsVector[i];
                }
                else
                {
                    RHS_vec[offs + i] = (mu / (IPD->lambda[i])) + IPD->ConstraintsVector[i];
                }
            }
        }
    }

    template <typename real>
    void KKTSystem<real>::getIPMatrix(IPdata<real> const * IPD, SparseMatrixBase<real> * IP_Sparse)
    {
        this->LocalTimerIPSystem.tic();

        bool Flag_Hessian_is_symmetric  = IPD->getProblem()->is_lagrangian_hessian_symmetric();
        bool Flag_IPmatrix_is_symmetric = IPD->getProblem()->is_IPmatrix_symmetric();

        IPtimings<real> * MyIPtimings = IPD->getTimings();
        // Clear IPMatrix:
        IP_Sparse->setZero();
        // Get lambda, slack and is_equ:
        real const *     lambda          = IPD->lambda;
        real const *     slack           = IPD->slack;
        vec_bool const & is_equality_vec = IPD->is_equality_vec;
        real             RegDiagEqu      = IPD->getIPparameter()->RegDiagMatrixEqu;

        // Get jacobian and hessian:
        SparseMatrixBase<real> * Jac  = IPD->getProblem()->get_constraint_jacobian();
        SparseMatrixBase<real> * Hess = IPD->getProblem()->get_lagrangian_hessian();

        // Set dimensions:
        // int_type numRowsIP = IPD->DimStates + IPD->DimLambda;
        // int_type numColsIP = numRowsIP;
        // int_type nnzIP     = Jac->get_nnz() + Hess->get_nnz() + IPD->DimLambda;

        // Init IP_Sparse: @@@@@@@@@@@@@@@@@@@@@
        // IP_Sparse->init(numRowsIP, numColsIP, nnzIP, true);

        IP_Sparse->push_matrix(IPD->DimStates, 0, *Jac);
        if (!Flag_IPmatrix_is_symmetric)
            IP_Sparse->push_matrix(0, IPD->DimStates, *Jac, true);

        if (Flag_IPmatrix_is_symmetric == Flag_Hessian_is_symmetric)
        {
            IP_Sparse->push_matrix(0, 0, *Hess);
        }
        else
        {
            IPbasic::Errors::set_Error("NOT YET IMPLEMENTED");
        }

        // Push diagonal matrix of lagrange multipliers and slack variables:
        int_type row_col_base = IPD->DimStates;
        int_type row_col      = IPD->DimStates;
        for (int_type index = 0; index < IPD->DimLambda; ++index)
        {
            size_t ii = size_t(index);
            row_col   = row_col_base + index;
            real tmp  = is_equality_vec[ii] ? RegDiagEqu : // equality
                           -(slack[ii] / lambda[ii]);      // inequality
            IP_Sparse->push_value_C(row_col, row_col, tmp);
        }
        this->LocalTimerIPSystem.toc();
        MyIPtimings->TimeSetupIPSystem += this->LocalTimerIPSystem.elapsed_s();
    }

    template <typename real>
    void KKTSystem<real>::getInitialGuessLambdaRHS(IPdata<real> const * IPD, real * IGLambdaRHS)
    {
        // Get dimensions:
        int_type DimStates     = IPD->DimStates;
        int_type DimLambda     = IPD->DimLambda;
        real *   GradOBJVector = IPD->grad_objective;

        // Get jacobian:
        SparseMatrixBase<real> * Jac = IPD->getProblem()->get_constraint_jacobian();

        // First DimStates + DimLambda rows:
        real val = real(0.0);
        for (size_t i = 0; i < size_t(DimStates + DimLambda); ++i)
            IGLambdaRHS[i] = val;

        // RHS add -J \nabla f:
        Jac->gemv(real(-1.0), DimStates, GradOBJVector, 1, real(1.0), DimLambda, &IGLambdaRHS[DimStates], 1);
    }

    template <typename real>
    void KKTSystem<real>::getInitialGuessLambdaMatrix(IPdata<real> const * IPD, SparseMatrixBase<real> * IGLambda)
    {
        // Get dimensions:
        int_type DimStates                  = IPD->DimStates;
        int_type DimLambda                  = IPD->DimLambda;
        bool     Flag_IPmatrix_is_symmetric = IPD->getProblem()->is_IPmatrix_symmetric();

        // Clear IGLambda:
        IGLambda->setZero();

        // Get jacobian:
        SparseMatrixBase<real> * Jac = IPD->getProblem()->get_constraint_jacobian();

        // Set dimensions:
        // int_type numRowsIG = DimStates + DimLambda;
        // int_type numColsIG = numRowsIG;
        // int_type nnzIG     = Jac->get_nnz() + DimStates + DimLambda;

        // Init IGLambda: @@@@@@@@@@@@@@@@@
        // IGLambda->init(numRowsIG, numColsIG, nnzIG, true);

        real val = real(1.0);
        // Push I into IGLambda:
        for (int_type i = 0; i < DimStates; ++i)
            IGLambda->push_value_C(i, i, val);

        // Push jacobian to IGLambda:
        IGLambda->push_matrix(DimStates, 0, *Jac);
        if (!Flag_IPmatrix_is_symmetric)
            IGLambda->push_matrix(0, DimStates, *Jac, true);

        // Push regularisation parameter:
        val = -IPD->getIPparameter()->RegDiagMatrixInitialGuessLambda;
        for (int_type i = 0; i < DimLambda; ++i)
            IGLambda->push_value_C(DimStates + i, DimStates + i, val);
    }

    template class KKTSystem<float>;
    template class KKTSystem<double>;
} // namespace IPbasic
