/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Solver.cpp
 * Implementation of the class Solver.
 *
 * @author A. Huber
 * @since 05.02.2019
 */

#include "IPbasic_Solver.h"
#include "Errors/IPbasic_Errors.h"
#include "LineSearch/IPbasic_Armijo.h"
#include "LineSearch/IPbasic_L1_Penalty.h"
#include <math.h>

namespace IPbasic
{

    template <typename real>
    Solver<real>::Solver(NLP<real> * Problem, LinearSolver_interface<real> * LinSolver, IPparameter<real> * IPParams, IPtimings<real> * Timings, LineSearchMethod UsedLineSearchMethod)
    {
        this->InteriorPointData    = new IPdata<real>(Problem, IPParams, Timings);
        this->IPKKTSystem          = new KKTSystem<real>(this->InteriorPointData, LinSolver);
        this->IPLineSearchFunction = new L1_Penalty<real>(this->InteriorPointData);
        if (UsedLineSearchMethod == LINESEARCH_ARMIJO_L1_PENALTY)
        {
            // Use Armijo linesearch with L1-penalty-function:
            this->IPLineSearch = new Armijo<real>(this->IPLineSearchFunction);
        }
        else if (UsedLineSearchMethod == LINESEARCH_NONE)
        {
            // Use no Linesearch:
            this->IPLineSearch = new LineSearch<real>();
        }
        else
        {
            Errors::set_Warning("No propper linesearch flag!");
            this->IPLineSearch = new LineSearch<real>();
        }
    }

    template <typename real>
    Solver<real>::~Solver()
    {
        delete this->InteriorPointData;
        delete this->IPKKTSystem;
        delete this->IPLineSearch;
        delete this->IPLineSearchFunction;
    }

    /*\
    :|:             _    _ _                      _
    :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    bool Solver<real>::checkMaxIter()
    {
        return this->InteriorPointData->IterData.counter >= this->getIPparameter()->max_iter;
    }

    template <typename real>
    IPparameter<real> * Solver<real>::getIPparameter()
    {
        return this->InteriorPointData->getIPparameter();
    }

    template <typename real>
    IPtimings<real> * Solver<real>::getIPtimings()
    {
        return this->InteriorPointData->getTimings();
    }

    template <typename real>
    bool Solver<real>::optimize()
    {

        this->TimerOptimize.tic();

        if (!this->checkForValidMembers())
        {
            Errors::set_Error("Unvalid members!");
            return false;
        }

        // For Savety reason:
        this->IPKKTSystem->setPointerIPMatrix(this->InteriorPointData->getProblem()->get_IPMatrix());

        // Set initial guess x0:
        this->InteriorPointData->initDataFromProblem();

        IterationDataPoint<real> * IterData    = &this->InteriorPointData->IterData;
        NLP<real> *                Prob        = this->InteriorPointData->getProblem();
        IPtimings<real> *          MyIPtimings = this->getIPtimings();

        int_type DimStates = this->InteriorPointData->DimStates;
        int_type DimLambda = this->InteriorPointData->DimLambda;
        this->InteriorPointData->updateData();

        if (!this->InteriorPointData->has_initial_guess_lambda)
            this->IPKKTSystem->CreateInitialGuessLambda(this->InteriorPointData);

        // Define Errors:
        real DelKKT, ScaledDelKKT, DelConstr;
        this->IPKKTSystem->getKKTErrors(this->InteriorPointData, DelKKT, ScaledDelKKT, DelConstr);
        IterData->Delta_CSTR       = DelConstr;
        IterData->Delta_KKT        = DelKKT;
        IterData->scaled_Delta_KKT = ScaledDelKKT;
        IPparameter<real> * Param  = this->InteriorPointData->getIPparameter();
        if (this->InteriorPointData->getIPparameter()->PrintLev != PRINT_MARKED)
        {
            Errors::ColoredErrors   = false;
            Errors::ColoredWarnings = false;
        }
        if (this->InteriorPointData->getIPparameter()->PrintLev > PRINT_LESS)
        {
            IterData->printHeader(std::cout);
            std::cout << *IterData << std::endl;
        }

        if (this->InteriorPointData->check_INF_NAN())
        {
            Errors::set_Error("Inf or nan in initial data!");
            return false;
        }
        Errors::resetErrors();
        while (!this->checkFinished())
        {
            if (Errors::get_Error_Counter() > 0)
            {
                this->TimerOptimize.toc();
                MyIPtimings->TimeOptimize += this->TimerOptimize.elapsed_s();
                return false;
            }
            if (IterData->counter >= 1)
                this->updateMu(IterData->mu, IterData->scaled_Delta_KKT);

            if (Param->UseMuSequ)
                this->InteriorPointData->updateMuSequence();

            // Update IPD Data for current states, slacks and lambdas:
            this->LocalTimerSolver.tic();
            this->InteriorPointData->updateData();
            this->LocalTimerSolver.toc();
            MyIPtimings->TimeDataUpdate += this->LocalTimerSolver.elapsed_s();

            // Get new search direction:
            if (!this->IPKKTSystem->updateSearchDirection(this->InteriorPointData))
            {
                Errors::set_Error("Could not solve linear system!");
                this->TimerOptimize.toc();
                MyIPtimings->TimeOptimize += this->TimerOptimize.elapsed_s();
                return false;
            }

            // Get new stepsize:
            this->LocalTimerSolver.tic();
            real alpha_max = this->getIPparameter()->alpha_max;
            real alpha_s;
            real alpha_l;
            real alpha;
            this->getAlphaMax(alpha_s, alpha_l);
            if (Param->MyLineSearchCouples == COUPLES_NONE)
            {
                alpha = this->IPLineSearch->getStepsize(this->InteriorPointData, alpha_max);
            }
            else if (Param->MyLineSearchCouples == COUPLE_STATE_SLACK)
            {
                alpha   = this->IPLineSearch->getStepsize(this->InteriorPointData, alpha_s);
                alpha_s = alpha;
            }
            else
            { // Param->MyLineSearchCouples == COUPLE_STATE_SLACK_LAMBDA
                alpha_max = std::min(alpha_l, alpha_s);
                alpha     = this->IPLineSearch->getStepsize(this->InteriorPointData, alpha_max);
                alpha_s   = alpha;
                alpha_l   = alpha;
            }

            // Apply new step:
            this->applyStep(alpha, alpha_s, alpha_l);
            this->LocalTimerSolver.toc();
            MyIPtimings->TimeLineSearch += this->LocalTimerSolver.elapsed_s();

            // Update errors:
            this->IPKKTSystem->getKKTErrors(this->InteriorPointData, DelKKT, ScaledDelKKT, DelConstr);

            // Update iteration data:
            real   obj_old             = IterData->obj;
            real * d_states            = this->InteriorPointData->IP_d_states;
            real * d_slack             = this->InteriorPointData->IP_d_slack;
            real * d_lambda            = this->InteriorPointData->IP_d_lambda;
            real   del_states          = absmax(DimStates, d_states, 1);
            real   del_slack           = absmax(DimLambda, d_slack, 1);
            real   del_lambda          = absmax(DimLambda, d_lambda, 1);
            IterData->Delta_CSTR       = DelConstr;
            IterData->Delta_KKT        = DelKKT;
            IterData->scaled_Delta_KKT = ScaledDelKKT;
            IterData->alpha            = alpha;
            IterData->alpha_s          = alpha_s;
            IterData->alpha_l          = alpha_l;
            IterData->obj              = Prob->objective(this->InteriorPointData->states);
            IterData->Delta_obj        = std::abs(obj_old - IterData->obj);
            IterData->Delta_d          = std::max(std::max(del_states, del_slack), del_lambda);
            IterData->counter++;
            if (this->InteriorPointData->getIPparameter()->PrintLev > PRINT_LESS)
                std::cout << *IterData << std::endl;
            if (this->InteriorPointData->check_INF_NAN())
            {
                Errors::set_Error("Inf or nan in data!");
                this->TimerOptimize.toc();
                MyIPtimings->TimeOptimize += this->TimerOptimize.elapsed_s();
                return false;
            }
            if (Param->BreakpointMaxTime)
            {
                this->TimerOptimize.toc();
                real time = this->TimerOptimize.elapsed_s();
                if (time >= Param->maximalTime)
                {
                    MyIPtimings->TimeOptimize += this->TimerOptimize.elapsed_s();
                    Errors::set_Error("Breakpoint due to maximal time!");
                    return false;
                }
            }
        }

        this->TimerOptimize.toc();
        MyIPtimings->TimeOptimize += this->TimerOptimize.elapsed_s();
        return true;
    }

    template <typename real>
    void Solver<real>::copySolution(real x[]) const
    {
        // Get data:
        int_type     DimStates = this->InteriorPointData->DimStates;
        real const * x_Solver  = this->InteriorPointData->states;
        // Copy data:
        std::copy(x_Solver, x_Solver + DimStates, x);
    }

    template <typename real>
    void Solver<real>::copySolution(real x[], real slack[], real lambda[]) const
    {
        // Get dimensions:
        int_type DimStates = this->InteriorPointData->DimStates;
        int_type DimLambda = this->InteriorPointData->DimLambda;
        // Get data:
        real const * x_Solver      = this->InteriorPointData->states;
        real const * slack_Solver  = this->InteriorPointData->slack;
        real const * lambda_Solver = this->InteriorPointData->lambda;
        // Copy data:
        std::copy(x_Solver, x_Solver + DimStates, x);
        std::copy(slack_Solver, slack_Solver + DimLambda, slack);
        std::copy(lambda_Solver, lambda_Solver + DimLambda, lambda);
    }

    template <typename real>
    IterationDataPoint<real> Solver<real>::getLastIterationDataPoint() const
    {
        return this->InteriorPointData->IterData;
    }

    /*\
    :|:                _          _          _                   _
    :|:   _ __ _ _ ___| |_ ___ __| |_ ___ __| |  _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ '_/ _ \  _/ -_) _|  _/ -_) _` | | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/_| \___/\__\___\__|\__\___\__,_| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    bool Solver<real>::checkForValidMembers()
    {
        if (nullptr == this->InteriorPointData)
            return false;
        if (nullptr == this->IPKKTSystem)
            return false;
        if (nullptr == this->IPLineSearch)
            return false;
        if (!this->InteriorPointData->checkForValidMembers())
            return false;
        if (!this->IPKKTSystem->checkForValidMembers())
            return false;
        return true;
    }

    template <typename real>
    bool Solver<real>::checkFinished()
    {
        IterationDataPoint<real> * IterData = &this->InteriorPointData->IterData;
        IPparameter<real> *        Param    = this->InteriorPointData->getIPparameter();
        real                       mu       = IterData->mu;
        real                       KKT_sc   = IterData->scaled_Delta_KKT;
        real                       Constr   = IterData->Delta_CSTR;
        real                       tolKKT   = Param->epsilon_tol;
        real                       tolmu    = Param->mu_tol;
        int_type                   counter  = IterData->counter;
        if (counter >= Param->max_iter)
        {
            IterData->printLine(std::cout);
            Errors::set_Warning("Maximum number of iterations!");
            return true;
        }

        // if ((Constr <= tolKKT) && (KKT <= tolKKT) && (mu <= tolmu))
        if ((Constr <= tolKKT) && (KKT_sc <= tolKKT) && (mu <= tolmu))
        {
            IterData->printLine(std::cout);
            return true;
        }
        else
        {
            return false;
        }
    }

    template <typename real>
    void Solver<real>::getAlphaMax(real & alphaMax_s, real & alphaMax_lambda)
    {
        IPdata<real> *      IPD       = this->InteriorPointData;
        IPparameter<real> * Param     = IPD->getIPparameter();
        int_type            DimLambda = IPD->DimLambda;
        vec_bool const &    is_equ    = IPD->is_equality_vec;
        real *              slack     = IPD->slack;
        real *              lambda    = IPD->lambda;
        real *              d_slack   = IPD->IP_d_slack;
        real *              d_lambda  = IPD->IP_d_lambda;

        real beta       = Param->beta;
        real tol_alpha  = Param->tol_alpha;
        real la_tol     = Param->lambda_tol;
        alphaMax_s      = Param->alpha_max;
        alphaMax_lambda = Param->alpha_max;

        // Maximal stepsize of slack variables:
        bool Break = false;
        real value = 0.0;
        while (alphaMax_s > tol_alpha && Break == false)
        {
            Break = true;
            for (int_type i = 0; i < DimLambda; ++i)
            {
                if (!is_equ[size_t(i)])
                {
                    real tau = IPD->getTau(i);
                    value    = slack[size_t(i)] + alphaMax_s * d_slack[size_t(i)];
                    if (value < (1 - tau) * slack[size_t(i)])
                        Break = false;
                }
            }
            if (Break == false)
                alphaMax_s *= beta;
        }
        if (Break == false)
        {
            Errors::set_Warning("Minimal stepsize alpha_s for the slack variables!");
            alphaMax_s = tol_alpha;
        }

        // Maximal stepsize of lagrange multipliers:
        Break = false;
        while (alphaMax_lambda > tol_alpha && Break == false)
        {
            Break = true;
            for (int_type i = 0; i < DimLambda; ++i)
            {
                if (!is_equ[size_t(i)])
                {
                    real tau = IPD->getTau(i);
                    value    = lambda[size_t(i)] + alphaMax_lambda * d_lambda[size_t(i)];
                    if (value < (1 - tau) * lambda[i] + la_tol)
                        Break = false;
                }
            }
            if (Break == false)
                alphaMax_lambda *= beta;
        }
        if (Break == false)
        {
            Errors::set_Warning("Minimal stepsize alpha_l for the lagrange multipliers!");
            alphaMax_lambda = tol_alpha;
        }
    }

    template <typename real>
    void Solver<real>::applyStep(real alpha, real alpha_s, real alpha_l)
    {
        IPdata<real> *   IPD       = this->InteriorPointData;
        int_type         DimStates = IPD->DimStates;
        int_type         DimLambda = IPD->DimLambda;
        real *           states    = IPD->states;
        real *           slack     = IPD->slack;
        real *           lambda    = IPD->lambda;
        real *           d_states  = IPD->IP_d_states;
        real *           d_slack   = IPD->IP_d_slack;
        real *           d_lambda  = IPD->IP_d_lambda;
        vec_bool const & is_equ    = IPD->is_equality_vec;

        // Apply step for states:
        for (size_t i = 0; i < size_t(DimStates); ++i)
            states[i] += alpha * d_states[i];

        // Apply step for slack variables and lagrange multipliers:
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            if (is_equ[i])
                slack[i] = 0.0;
            else
                slack[i] += alpha_s * d_slack[i];
            lambda[i] += alpha_l * d_lambda[i];
        }
    }

    template <typename real>
    void Solver<real>::updateMu(real & mu, real Error)
    {
        IPparameter<real> * Param = this->InteriorPointData->getIPparameter();
        if (Error <= Param->kappa_epsilon * mu)
        {
            real a  = Param->mu_tol / real(10);
            real b1 = Param->kappa_mu * mu;
            real b2 = pow(mu, Param->Theta_mu);
            real b  = std::min(b1, b2);
            mu      = std::max(a, b);
        }
        if (mu <= Param->mu_tol)
            mu = Param->mu_tol;
    }

    template class Solver<float>;
    template class Solver<double>;

} // namespace IPbasic
