/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_Armijo.cpp
 * Implementation of the class Armijo.
 *
 * @author A. Huber
 * @since 08.02.2019
 */

#include "IPbasic_Armijo.h"
#include "../Errors/IPbasic_Errors.h"

namespace IPbasic
{

    template <typename real>
    Armijo<real>::Armijo(LineSearchFunction<real> * _MyLineSearchFunction)
        : LineSearch<real>()
        , MyLineSearchFunction(_MyLineSearchFunction)
    {
    }

    template <typename real>
    Armijo<real>::~Armijo()
    {
    }

    /*\
    :|:             _    _ _                      _
    :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    real Armijo<real>::getStepsize(IPdata<real> * IPD, real alpha_max)
    {
        real sigma     = IPD->getIPparameter()->sigma;
        real beta      = IPD->getIPparameter()->beta;
        real alpha     = alpha_max / beta;
        real tol       = IPD->getIPparameter()->tol_alpha;
        real F_x       = this->MyLineSearchFunction->getF(IPD, 0.0);
        real F_d_alpha = this->MyLineSearchFunction->getF(IPD, alpha);
        real dF_x      = this->MyLineSearchFunction->getDF(IPD);
        if (dF_x >= real(0.0))
        {
            Errors::set_Warning("Directional derivative >= 0 in armijo linesearch!");
            return alpha_max;
        }
        while (true)
        {
            alpha *= beta;
            if (alpha <= tol)
            {
                Errors::set_Warning("Minimal stepsize in armijo reached!");
                return tol;
            }
            F_d_alpha = this->MyLineSearchFunction->getF(IPD, alpha);
            if (!IPbasic::isRegular(F_d_alpha))
            {
                continue;
            }
            if (F_d_alpha <= F_x + sigma * alpha * dF_x)
            {
                break;
            }
        }
        return alpha;
    }

    template class Armijo<float>;
    template class Armijo<double>;

} // namespace IPbasic
