/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file IPbasic_L1_Penalty.cpp
 * Implementation of the class L1_Penalty.
 *
 * @author A. Huber
 * @since 08.02.2019
 */

#include "IPbasic_L1_Penalty.h"
#include <vector>

namespace IPbasic
{

    template <typename real>
    L1_Penalty<real>::L1_Penalty(IPdata<real> * IPD)
        : LineSearchFunction<real>()
        , mem_real("L1_Penalty")
    {
        int_type DimStates = IPD->DimStates;
        int_type DimLambda = IPD->DimLambda;
        mem_real.allocate(size_t(DimStates + 3 * DimLambda));
        this->local_states      = mem_real(size_t(DimStates));
        this->local_slack       = mem_real(size_t(DimLambda));
        this->local_constraints = mem_real(size_t(DimLambda));
        this->eta               = mem_real(size_t(DimLambda));

        std::copy_n(IPD->states, DimStates, this->local_states);
        std::copy_n(IPD->slack, DimLambda, this->local_slack);

        for (int_type i = 0; i < DimLambda; ++i)
        {
            this->eta[i]               = real(1.0);
            this->local_constraints[i] = real(0.0);
        }
    }

    template <typename real>
    L1_Penalty<real>::~L1_Penalty()
    {
        mem_real.free();
    }

    /*\
    :|:             _    _ _                      _
    :|:   _ __ _  _| |__| (_)__   _ __  ___ _ __ | |__  ___ _ _ ___
    :|:  | '_ \ || | '_ \ | / _| | '  \/ -_) '  \| '_ \/ -_) '_(_-<
    :|:  | .__/\_,_|_.__/_|_\__| |_|_|_\___|_|_|_|_.__/\___|_| /__/
    :|:  |_|
    \*/

    template <typename real>
    real L1_Penalty<real>::getF(IPdata<real> * IPD, real alpha)
    {
        // Get Data:
        this->updatePenaltyParams(IPD, alpha);
        NLP<real> *      Problem = IPD->getProblem();
        real             mu      = IPD->IterData.mu;
        vec_bool const & is_equ  = IPD->is_equality_vec;
        // Get dimensions:
        int_type DimStates = IPD->DimStates;
        int_type DimLambda = IPD->DimLambda;
        // Copy states and slack variables:
        std::copy_n(IPD->states, DimStates, this->local_states);
        std::copy_n(IPD->slack, DimLambda, this->local_slack);

        // Set local states and slack:
        for (size_t i = 0; i < size_t(DimStates); ++i)
            this->local_states[i] += alpha * IPD->IP_d_states[i];
        for (size_t i = 0; i < size_t(DimLambda); ++i)
            this->local_slack[i] += alpha * IPD->IP_d_slack[i];

        // Evaluate Problem:
        real obj = Problem->objective(this->local_states);
        Problem->constraint(this->local_states, this->local_constraints);

        real L1_Pen = obj;
        real constr = real(0.0);

        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            // Only for inequalities:
            if (!is_equ[i])
            {
                L1_Pen -= mu * log(this->local_slack[i]);
                constr = this->local_constraints[i] + this->local_slack[i];
                L1_Pen += this->eta[i] * std::abs(constr);
            }
            else
            {
                constr = this->local_constraints[i];
                L1_Pen += this->eta[i] * std::abs(constr);
            }
        }
        return L1_Pen;
    }

    template <typename real>
    real L1_Penalty<real>::getDF(IPdata<real> * IPD)
    {
        // Get Data:
        this->updatePenaltyParams(IPD, real(0.0));
        real             mu     = IPD->IterData.mu;
        vec_bool const & is_equ = IPD->is_equality_vec;
        // Get dimensions:
        int_type DimStates = IPD->DimStates;
        int_type DimLambda = IPD->DimLambda;

        real dL1_Pen = real(0.0);

        // Nabla f(x)^\top d_x:
        for (int_type i = 0; i < DimStates; ++i)
            dL1_Pen += IPD->grad_objective[i] * IPD->IP_d_states[i];

        real constr = real(0.0);
        this->LocalConstraintsToNablaD(IPD);

        // Rest:
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            // Only for inequalities:
            if (!is_equ[i])
            {
                // - mu \sum d_s/s :
                dL1_Pen -= mu * ((IPD->IP_d_slack[i]) / (IPD->slack[i]));
            }
            constr = IPD->ConstraintsVector[i] + IPD->slack[i];
            if (constr > real(0.0))
            {
                // \sum eta Nabla(c) d :
                dL1_Pen += this->eta[i] * this->local_constraints[i];
            }
            else if (constr < real(0.0))
            {
                // -\sum eta Nabla(c) d :
                dL1_Pen -= this->eta[i] * this->local_constraints[i];
            }
            else
            {
                // \sum eta \abs(Nabla(c) d) :
                dL1_Pen += this->eta[i] * std::abs(this->local_constraints[i]);
            }
        }
        return dL1_Pen;
    }

    /*\
    :|:   ___     _          _         __  __           _
    :|:  | _ \_ _(_)_ ____ _| |_ ___  |  \/  |___ _ __ | |__  ___ _ _ ___
    :|:  |  _/ '_| \ V / _` |  _/ -_) | |\/| / -_) '  \| '_ \/ -_) '_(_-<
    :|:  |_| |_| |_|\_/\__,_|\__\___| |_|  |_\___|_|_|_|_.__/\___|_| /__/
    \*/

    template <typename real>
    void L1_Penalty<real>::LocalConstraintsToNablaD(IPdata<real> * IPD)
    {
        int_type                       DimStates = IPD->DimStates;
        int_type                       DimLambda = IPD->DimLambda;
        vec_bool const &               is_equ    = IPD->is_equality_vec;
        SparseMatrixBase<real> const * Jacobi    = IPD->constraint_jacobian;
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            if (!is_equ[i])
            {
                this->local_constraints[i] = IPD->IP_d_slack[i];
            }
            else
            {
                this->local_constraints[i] = real(0.0);
            }
        }

        Jacobi->gemv(1.0, DimStates, IPD->IP_d_states, 1, real(1.0), DimLambda, this->local_constraints, 1);
    }

    template <typename real>
    void L1_Penalty<real>::updatePenaltyParams(IPdata<real> * IPD, real alpha)
    {
        int_type     DimLambda = IPD->DimLambda;
        real const * lambda    = IPD->lambda;
        real const * d_lambda  = IPD->IP_d_lambda;
        real         labs      = real(0.0);
        real         cmp       = real(0.0);
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            labs         = std::abs(lambda[i] + alpha * d_lambda[i]);
            cmp          = (this->eta[i] + labs) / real(2.0);
            this->eta[i] = std::max(labs, cmp);
        }
        /*
        int_type     DimLambda = IPD->DimLambda;
        real const * lambda    = IPD->lambda;
        real const * d_lambda  = IPD->IP_d_lambda;
        real         labs      = real(0.0);
        real         cmp       = real(0.0);
        for (size_t i = 0; i < size_t(DimLambda); ++i)
        {
            labs         = std::abs(lambda[i] + d_lambda[i]);
            cmp          = lambda[i] + d_lambda[i];
            this->eta[i] = std::max(labs, cmp);
        }
        */
    }

    template class L1_Penalty<float>;
    template class L1_Penalty<double>;

} // namespace IPbasic
