/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  Copyright (C) 2019                                                      |
 |                                                                          |
 |      Enrico Bertolazzi (1) & Andreas Huber (2)                           |
 |      (1) Dipartimento di Ingegneria Industriale                          |
 |          Universita` degli Studi di Trento                               |
 |          email: enrico.bertolazzi@unitn.it                               |
 |      (2) Fakultät für Luft- und Raumfahrttechnik                         |
 |          Universität der Bundeswehr München                              |
 |          email: info@ocpbasic.com                                        |
\*--------------------------------------------------------------------------*/

/**
 * @file hsl.h
 * Header Definition der Fortran-Routinen von MA57 und MA48.
 *
 * @author A. Huber and E.Bertolazzi
 * @since 13.11.2018
\*/

#include "../../IPbasic_Config.h"

#ifdef IPBASIC_USE_HSL

#ifndef HSL_H
#define HSL_H

#ifdef __cplusplus
extern "C"
{
#endif

#ifndef F77NAME
#define F77NAME(A) A##_
#endif

    /**
     * \brief ma57id_:
     *        Get the control parameters of MA57.
     *
     * \param[out] cntl   Controlparameter für double.
     * \param[out] icntl  Controlparameter für int.
     *
    \*/
    extern void F77NAME(ma57id)(double cntl[5], int icntl[20]);

    /**
     * \brief ma57ad_:
     *        Symbolic decomposition and choice of the pivot elements
     *
     * \param[in]  n        Ordnung der Matrix.
     * \param[in]  nz       Anzahl der Sparse Elemente.
     * \param[in]  irn      Zeilenarray der Sparsematrix.
     * \param[in]  jcn      Spaltenarray der Sparsematrix.
     * \param[in]  lkeep    Dimension des keep-arrays (\f$lkeep \geq 5 n+nz+max(n,nz)+42\f$).
     * \param[out] keep     Array von Integern für Pivot Strategie.
     * \param[out] iw       Workspace von Integern (\f$5n\f$).
     * \param[in]  icntl    Controlparameter für int.
     * \param[out] info     Info-Array für int.
     * \param[out] rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57ad)(int const * n, int const * nz, int const irn[], int const jcn[], int const * lkeep, int keep[], int iw[], int const icntl[20], int info[40], double rinfo[20]);

    /**
     * \brief ma57bd_:
     *        Faktorisiert die Sparsematrix.
     *
     * \param[in]  n        Ordnung der Matrix.
     * \param[in]  nz       Anzahl der Sparse Elemente.
     * \param[in]  a        Wertearray der Sparsematrix.
     * \param[out] fact     Faktoren der Sparsematrix für double.
     * \param[in]  lfact    Dimension des Faktorenarrays für double.
     * \param[out] ifact    Faktoren der Sparsematrix für int.
     * \param[in]  lifact   Dimension des Faktorenarrays für int.
     * \param[in]  lkeep    Dimension des keep-arrays (\f$lkeep \geq 5 n+nz+max(n,nz)+42\f$).
     * \param[in]  keep     Array von Integern für Pivot Strategie.
     * \param[out] iw       Workspace von Integern (\f$5n\f$).
     * \param[in]  icntl    Controlparameter für int.
     * \param[in]  cntl     Controlparameter für double.
     * \param[out] info     Info-Array für int.
     * \param[out] rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57bd)(int const *  n,
                                int const *  nz,
                                double const a[],
                                double       fact[],
                                int const *  lfact,
                                int          ifact[],
                                int const *  lifact,
                                int const *  lkeep,
                                int const    keep[],
                                int          iw[],
                                int const    icntl[20],
                                double const cntl[5],
                                int          info[40],
                                double       rinfo[20]);

    /**
     * \brief ma57cd_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite.
     *
     * \param[in]     job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]     n        Ordnung der Matrix.
     * \param[in]     fact     Faktoren der Sparsematrix für double.
     * \param[in]     lfact    Dimension des Faktorenarrays für double.
     * \param[in]     ifact    Faktoren der Sparsematrix für int.
     * \param[in]     lifact   Dimension des Faktorenarrays für int.
     * \param[in]     nrhs     Anzahl der rechten Seiten.
     * \param[in,out] rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[in]     lrhs     Zeilendimension der rechten Seite.
     * \param[out]    w        Workspace für double.
     * \param[in]     lw       Dimension des Workspaces für double.
     * \param[out]    iw       Workspace für int.
     * \param[in]     icntl    Controlparameter für int.
     * \param[out]    info     Info-Array für int.
     *
    \*/
    extern void F77NAME(ma57cd)(int const *  job,
                                int const *  n,
                                double const fact[],
                                int const *  lfact,
                                int const    ifact[],
                                int const *  lifact,
                                int const *  nrhs,
                                double       rhs[],
                                int const *  lrhs,
                                double       w[],
                                int const *  lw,
                                int          iw[],
                                int const    icntl[20],
                                int          info[40]);

    /**
     * \brief ma57dd_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite
     *        mit iterativer Verfeinerung.
     *
     * \param[in]     job      Art der linearen Gleichungssystem Lösung.
     * \param[in]     n        Ordnung der Matrix.
     * \param[in]     ne       Anzahl der nicht Null Elemente in the Ursprungsmatrix.
     * \param[in]     a        Wertearray der Ursprungsmatrix.
     * \param[in]     irn      Indexarray der Zeilen der Ursprungsmatrix.
     * \param[in]     jcn      Indexarray der Spalten der Ursprungsmatrix.
     * \param[in]     fact     Faktoren der Sparsematrix für double.
     * \param[in]     lfact    Dimension des Faktorenarrays für double.
     * \param[in]     ifact    Faktoren der Sparsematrix für int.
     * \param[in]     lifact   Dimension des Faktorenarrays für int.
     * \param[in]     rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[out]    x        Lösung linearen Gleichungssystems.
     * \param[out]    resid    Residuum der Lösung des linearen Gleichungssystems.
     * \param[out]    w        Workspace für double.
     * \param[out]    iw       Workspace für int.
     * \param[in]     icntl    Controlparameter für int.
     * \param[in]     cntl     Controlparameter für double.
     * \param[out]    info     Info-Array für int.
     * \param[out]    rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57dd)(int const *  job,
                                int const *  n,
                                int const *  ne,
                                double const a[],
                                int const    irn[],
                                int const    jcn[],
                                double const fact[],
                                int const *  lfact,
                                int const    ifact[],
                                int const *  lifact,
                                double const rhs[],
                                double       x[],
                                double       resid[],
                                double       w[],
                                int          iw[],
                                int const    icntl[20],
                                double const cntl[5],
                                int          info[40],
                                double       rinfo[20]);

    /**
     * \brief ma57i_:
     *        Setzt die Controlparameter von MA57.
     *
     * \param[out] cntl     Controlparameter für double.
     * \param[out] icntl    Controlparameter für int.
     *
    \*/
    extern void F77NAME(ma57i)(float cntl[5], int icntl[20]);

    /**
     * \brief ma57a_:
     *        Symbolische Zerlegung und Wahl der Pivotelemente.
     *
     * \param[in]  n        Ordnung der Matrix.
     * \param[in]  nz       Anzahl der Sparse Elemente.
     * \param[in]  irn      Zeilenarray der Sparsematrix.
     * \param[in]  jcn      Spaltenarray der Sparsematrix.
     * \param[in]  lkeep    Dimension des keep-arrays (\f$lkeep \geq 5 n+nz+max(n,nz)+42\f$).
     * \param[out] keep     Array von Integern für Pivot Strategie.
     * \param[out] iw       Workspace von Integern (\f$5n\f$).
     * \param[in]  icntl    Controlparameter für int.
     * \param[out] info     Info-Array für int.
     * \param[out] rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57a)(int const * n, int const * nz, int const irn[], int const jcn[], int const * lkeep, int keep[], int iw[], int const icntl[20], int info[40], float rinfo[20]);

    /**
     * \brief ma57b_:
     *        Faktorisiert die Sparsematrix.
     *
     * \param[in]  n        Ordnung der Matrix.
     * \param[in]  nz       Anzahl der Sparse Elemente.
     * \param[in]  a        Wertearray der Sparsematrix.
     * \param[out] fact     Faktoren der Sparsematrix für double.
     * \param[in]  lfact    Dimension des Faktorenarrays für double.
     * \param[out] ifact    Faktoren der Sparsematrix für int.
     * \param[in]  lifact   Dimension des Faktorenarrays für int.
     * \param[in]  lkeep    Dimension des keep-arrays (\f$lkeep \geq 5 n+nz+max(n,nz)+42\f$).
     * \param[in]  keep     Array von Integern für Pivot Strategie.
     * \param[out] iw       Workspace von Integern (\f$5n\f$).
     * \param[in]  icntl    Controlparameter für int.
     * \param[in]  cntl     Controlparameter für double.
     * \param[out] info     Info-Array für int.
     * \param[out] rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57b)(int const * n,
                               int const * nz,
                               float const a[],
                               float       fact[],
                               int const * lfact,
                               int const   ifact[],
                               int const * lifact,
                               int const * lkeep,
                               int const   keep[],
                               int         iw[],
                               int const   icntl[20],
                               float const cntl[5],
                               int         info[40],
                               float       rinfo[20]);

    /**
     * \brief ma57c_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite.
     *
     * \param[in]     job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]     n        Ordnung der Matrix.
     * \param[in]     fact     Faktoren der Sparsematrix für double.
     * \param[in]     lfact    Dimension des Faktorenarrays für double.
     * \param[in]     ifact    Faktoren der Sparsematrix für int.
     * \param[in]     lifact   Dimension des Faktorenarrays für int.
     * \param[in]     nrhs     Anzahl der rechten Seiten.
     * \param[in,out] rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[in]     ldRhs    Zeilendimension der rechten Seite.
     * \param[out]    w        Workspace für double.
     * \param[in]     lw       Dimension des Workspaces für double.
     * \param[out]    iw       Workspace für int.
     * \param[in]     icntl    Controlparameter für int.
     * \param[out]    info     Info-Array für int.
     *
    \*/
    extern void F77NAME(ma57c)(int const * job,
                               int const * n,
                               float const fact[],
                               int const * lfact,
                               int const   ifact[],
                               int const * lifact,
                               int const * nrhs,
                               float       rhs[],
                               int const * ldRhs,
                               float       w[],
                               int const * lw,
                               int         iw[],
                               int const   icntl[20],
                               int         info[40]);

    /**
     * \brief ma57d_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite
     *        mit iterativer Verfeinerung.
     *
     * \param[in]     job      Art der linearen Gleichungssystem Lösung.
     * \param[in]     n        Ordnung der Matrix.
     * \param[in]     ne       Anzahl der nicht Null Elemente in the Ursprungsmatrix.
     * \param[in]     a        Wertearray der Ursprungsmatrix.
     * \param[in]     irn      Indexarray der Zeilen der Ursprungsmatrix.
     * \param[in]     jcn      Indexarray der Spalten der Ursprungsmatrix.
     * \param[in]     fact     Faktoren der Sparsematrix für double.
     * \param[in]     lfact    Dimension des Faktorenarrays für double.
     * \param[in]     ifact    Faktoren der Sparsematrix für int.
     * \param[in]     lifact   Dimension des Faktorenarrays für int.
     * \param[in]     rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[out]    x        Lösung linearen Gleichungssystems.
     * \param[out]    resid    Residuum der Lösung des linearen Gleichungssystems.
     * \param[out]    w        Workspace für double.
     * \param[out]    iw       Workspace für int.
     * \param[in]     icntl    Controlparameter für int.
     * \param[in]     cntl     Controlparameter für double.
     * \param[out]    info     Info-Array für int.
     * \param[out]    rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma57d)(int const * job,
                               int const * n,
                               int const * ne,
                               float const a[],
                               int const   irn[],
                               int const   jcn[],
                               float const fact[],
                               int const * lfact,
                               int const   ifact[],
                               int const * lifact,
                               float const rhs[],
                               float       x[],
                               float *     resid,
                               float       w[],
                               int         iw[],
                               int const   icntl[20],
                               float const cntl[5],
                               int         info[40],
                               float       rinfo[20]);

    /**
     * \brief ma48id_:
     *        Setzt die Controlparameter von MA48.
     *
     * \param[out] cntl      Controlparameter für double.
     * \param[out] icntl     Controlparameter für int.
     *
    \*/
    extern void F77NAME(ma48id)(double cntl[5], int icntl[20]);

    /**
     * \brief ma48ad_:
     *        Symbolische Zerlegung und Wahl der Pivotelemente.
     *
     * \param[in]      nrow     Anzahl der Zeilen.
     * \param[in]      ncol     Anzahl der Spalten.
     * \param[in]      nz       Anzahl der Elemente der Sparsematrix.
     * \param[in]      job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]      la       Hilfsdomension \f$la\f$.
     * \param[in,out]  a        Wertearray der Sparsematrix.
     * \param[in,out]  irn      Zeilenarray der Sparsematrix.
     * \param[in,out]  jcn      Spaltenarray der Sparsematrix.
     * \param[out]     keep     Array von Integern für Pivot Strategie.
     * \param[in]      cntl     Controlparameter für double.
     * \param[in]      icntl    Controlparameter für int.
     * \param[out]     iw       Workspace für int.
     * \param[out]     info     Info-Array für int.
     * \param[out]     rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma48ad)(int const *  nrow,
                                int const *  ncol,
                                int const *  nz,
                                int const *  job,
                                int const *  la,
                                double const a[],
                                int          irn[],
                                int          jcn[],
                                int          keep[],
                                double const cntl[10],
                                int const    icntl[20],
                                int          iw[],
                                int          info[20],
                                double       rinfo[10]);

    /**
     * \brief ma48bd_:
     *        Faktorisiert die Sparsematrix.
     *
     * \param[in]      nrow     Anzahl der Zeilen.
     * \param[in]      ncol     Anzahl der Spalten.
     * \param[in]      nz       Anzahl der Elemente der Sparsematrix.
     * \param[in]      job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]      la       Hilfsdomension \f$la\f$.
     * \param[in,out]  a        Wertearray der Sparsematrix.
     * \param[in,out]  irn      Zeilenarray der Sparsematrix.
     * \param[in,out]  jcn      Spaltenarray der Sparsematrix.
     * \param[in]      keep     Array von Integern für Pivot Strategie.
     * \param[in]      cntl     Controlparameter für double.
     * \param[in]      icntl    Controlparameter für int.
     * \param[out]     w        Workspace für double.
     * \param[out]     iw       Workspace für int.
     * \param[out]     info     Info-Array für int.
     * \param[out]     rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma48bd)(int const *  nrow,
                                int const *  ncol,
                                int const *  nz,
                                int const *  job,
                                int const *  la,
                                double       a[],
                                int          irn[],
                                int          jcn[],
                                int const    keep[],
                                double const cntl[10],
                                int const    icntl[20],
                                double       w[],
                                int          iw[],
                                int          info[20],
                                double       rinfo[10]);

    /**
     * \brief ma48cd_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite.
     *
     * \param[in]  nrow     Anzahl der Zeilen.
     * \param[in]  ncol     Anzahl der Spalten.
     * \param[in]  itrans   Gibt an, ob die Sparsematrix transponiert ist.
     * \param[in]  job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]  la       Hilfsdomension \f$la\f$.
     * \param[in]  a        Wertearray der Sparsematrix.
     * \param[in]  irn      Zeilenarray der Sparsematrix.
     * \param[in]  keep     Array von Integern für Pivot Strategie.
     * \param[in]  cntl     Controlparameter für double.
     * \param[in]  icntl    Controlparameter für int.
     * \param[in]  rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[out] x        Lösung des linearen Gleichungssystems.
     * \param[out] errors   Errorarray des Lösers.
     * \param[in]  w        Workspace für double.
     * \param[in]  iw       Workspace für int.
     * \param[out] info     Info-Array für int.
     *
    \*/
    extern void F77NAME(ma48cd)(int const *  nrow,
                                int const *  ncol,
                                int const *  itrans,
                                int const *  job,
                                int const *  la,
                                double const a[],
                                int const    irn[],
                                int const    keep[],
                                double const cntl[10],
                                int const    icntl[20],
                                double const rhs[],
                                double       x[],
                                double       errors[3],
                                double const w[],
                                int const    iw[],
                                int          info[20]);

    /**
     * \brief ma48i_:
     *        Setzt die Controlparameter von MA48.
     *
     * \param[out] cntl      Controlparameter für double.
     * \param[out] icntl     Controlparameter für int.
     *
    \*/
    extern void F77NAME(ma48i)(float cntl[5], int icntl[20]);

    /**
     * \brief ma48a_:
     *        Symbolische Zerlegung und Wahl der Pivotelemente.
     *
     * \param[in]      nrow     Anzahl der Zeilen.
     * \param[in]      ncol     Anzahl der Spalten.
     * \param[in]      nz       Anzahl der Elemente der Sparsematrix.
     * \param[in]      job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]      la       Hilfsdomension \f$la\f$.
     * \param[in,out]  a        Wertearray der Sparsematrix.
     * \param[in,out]  irn      Zeilenarray der Sparsematrix.
     * \param[in,out]  jcn      Spaltenarray der Sparsematrix.
     * \param[out]     keep     Array von Integern für Pivot Strategie.
     * \param[in]      cntl     Controlparameter für double.
     * \param[in]      icntl    Controlparameter für int.
     * \param[out]     iw       Workspace für int.
     * \param[out]     info     Info-Array für int.
     * \param[out]     rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma48a)(int const * nrow,
                               int const * ncol,
                               int const * nz,
                               int const * job,
                               int const * la,
                               float       a[],
                               int         irn[],
                               int         jcn[],
                               int         keep[],
                               float const cntl[10],
                               int const   icntl[20],
                               int         iw[],
                               int         info[20],
                               float       rinfo[10]);

    /**
     * \brief ma48b_:
     *        Faktorisiert die Sparsematrix.
     *
     * \param[in]      nrow     Anzahl der Zeilen.
     * \param[in]      ncol     Anzahl der Spalten.
     * \param[in]      nz       Anzahl der Elemente der Sparsematrix.
     * \param[in]      job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]      la       Hilfsdomension \f$la\f$.
     * \param[in,out]  a        Wertearray der Sparsematrix.
     * \param[in,out]  irn      Zeilenarray der Sparsematrix.
     * \param[in,out]  jcn      Spaltenarray der Sparsematrix.
     * \param[in]      keep     Array von Integern für Pivot Strategie.
     * \param[in]      cntl     Controlparameter für double.
     * \param[in]      icntl    Controlparameter für int.
     * \param[out]     w        Workspace für double.
     * \param[out]     iw       Workspace für int.
     * \param[out]     info     Info-Array für int.
     * \param[out]     rinfo    Info-Array für double.
     *
    \*/
    extern void F77NAME(ma48b)(int const * nrow,
                               int const * ncol,
                               int const * nz,
                               int const * job,
                               int const * la,
                               float       a[],
                               int         irn[],
                               int         jcn[],
                               int const   keep[],
                               float const cntl[10],
                               int const   icntl[20],
                               float       w[],
                               int         iw[],
                               int         info[20],
                               float       rinfo[10]);

    /**
     * \brief ma48c_:
     *        Löst das lineare Gleichungssystem bezüglich einer rechten Seite.
     *
     * \param[in]  nrow     Anzahl der Zeilen.
     * \param[in]  ncol     Anzahl der Spalten.
     * \param[in]  itrans   Gibt an, ob die Sparsematrix transponiert ist.
     * \param[in]  job      Art der linearen Gleichungssystem Lösung (Hier \f$job = 1\f$).
     * \param[in]  la       Hilfsdomension \f$la\f$.
     * \param[in]  a        Wertearray der Sparsematrix.
     * \param[in]  irn      Zeilenarray der Sparsematrix.
     * \param[in]  keep     Array von Integern für Pivot Strategie.
     * \param[in]  cntl     Controlparameter für double.
     * \param[in]  icntl    Controlparameter für int.
     * \param[in]  rhs      Rechte Seite des linearen Gleichungssystems.
     * \param[out] x        Lösung des linearen Gleichungssystems.
     * \param[out] errors   Errorarray des Lösers.
     * \param[in]  w        Workspace für double.
     * \param[in]  iw       Workspace für int.
     * \param[out] info     Info-Array für int.
     *
    \*/
    extern void F77NAME(ma48c)(int const * nrow,
                               int const * ncol,
                               int const * itrans,
                               int const * job,
                               int const * la,
                               float const a[],
                               int const   irn[],
                               int const   keep[],
                               float const cntl[10],
                               int const   icntl[20],
                               float const rhs[],
                               float       x[],
                               float       errors[3],
                               float const w[],
                               int const   iw[],
                               int         info[20]);

#ifdef __cplusplus
}
#endif

#endif // HSL_H

#endif // IPBASIC_USE_HSL
