/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file DiscreteDimension.cpp
 * Implementation of the class DiscreteDimension.
 *
 * @author A. Huber
 * @since 27.09.2018
 */

#include "DiscreteDimension.h"
#include "../linalg/OCPbasic.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#include <math.h>
#include <memory.h>

namespace OCPbasic
{

#define NEW_MATW(NR, NC) new MatrixWrapper<real>(mem_real(size_t((NR) * (NC))), NR, NC, NR)
#define SETUP_MATW(NR, NC) setup(mem_real(size_t((NR) * (NC))), NR, NC, NR)

    using IPbasic::Errors;

    template <typename real>
    DiscreteDimension<real>::DiscreteDimension(Problem<real> * P, int_type N, real INFTY, real TOL)
        : mem_real("DiscreteDimension")
        , INF(INFTY)
        , TOL(TOL)
    {
        // Allokieren des Speichers:
        this->BoundTypeArrayNLCSTR.resize(P->N_NLCSTR);
        this->BoundTypeArrayBOUND.resize(P->N_BOUND);
        this->BoundTypeArraySTATE.resize(P->Ny);
        this->BoundTypeArrayCONTROL.resize(P->Nu);
        this->BoundTypeArrayPARAM.resize(P->Np);
        // Identifikation der Schranken und Einzeldimensionen:
        this->BoundDimNLCSTR.N  = P->N_NLCSTR;
        this->BoundDimBOUND.N   = P->N_BOUND;
        this->BoundDimSTATE.N   = P->Ny;
        this->BoundDimCONTROL.N = P->Nu;
        this->BoundDimPARAM.N   = P->Np;
        this->setIndexArrays(P);

        int_type nalloc = this->BoundDimSTATE.N_Dim_ROW * this->BoundDimSTATE.N + this->BoundDimCONTROL.N_Dim_ROW * this->BoundDimCONTROL.N + this->BoundDimPARAM.N_Dim_ROW * this->BoundDimPARAM.N;

        this->mem_real.allocate(size_t(nalloc));

        // Belege Matrizen für Zustands-, Steuerungs- und Parameterbeschränkungen:
        this->C_Constr_Y.SETUP_MATW(this->BoundDimSTATE.N_Dim_ROW, this->BoundDimSTATE.N);
        this->C_Constr_Y.zero_fill();

        this->C_Constr_U.SETUP_MATW(this->BoundDimCONTROL.N_Dim_ROW, this->BoundDimCONTROL.N);
        this->C_Constr_U.zero_fill();

        this->C_Constr_P.SETUP_MATW(this->BoundDimPARAM.N_Dim_ROW, this->BoundDimPARAM.N);
        this->C_Constr_P.zero_fill();

        this->set_internal_Constr_YUP();

        // Belege Index Arrays:
        this->Index_NLCSTR.resize(P->N_NLCSTR);
        this->set_internal_Index_NLCSTR();
        this->Index_BOUND.resize(P->N_BOUND);
        this->set_internal_Index_BOUND();

        this->N           = N;
        int_type DimBlock = this->BoundDimNLCSTR.N_Dim_ROW + this->BoundDimSTATE.N_Dim_ROW + this->BoundDimCONTROL.N_Dim_ROW;
        this->Dim_Schlupf = (this->N + 1) * (DimBlock) + this->BoundDimBOUND.N_Dim_ROW + this->BoundDimPARAM.N_Dim_ROW;
        this->DimLambdaIP = this->N * P->Ny + this->Dim_Schlupf;
        this->Case_lambda_k.resize(this->DimLambdaIP);
        std::fill(this->Case_lambda_k.begin(), this->Case_lambda_k.end(), 0x02);
        this->set_CaseLambda();
    }

    template <typename real>
    DiscreteDimension<real>::~DiscreteDimension()
    {
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   public Memberfunktionen                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void DiscreteDimension<real>::set_lambda_hesse(real const lambda_IP[], real lambda_hesse[])
    {
        int_type Ny       = this->BoundDimSTATE.N;
        int_type N_BOUND  = this->BoundDimBOUND.N;
        int_type N_NLCSTR = this->BoundDimNLCSTR.N;
        int_type Index    = 0;
        int_type Index_k  = 0;
        // Belegen der ODE-Beschränkungen:
        memcpy(lambda_hesse, lambda_IP, (this->N) * Ny * sizeof(real));
        // Belegen der BOUND-Beschränkungen:
        Index   = (this->N) * Ny;
        Index_k = (this->N) * Ny;
        for (int_type j = 0; j < N_BOUND; ++j)
        {

            int_type Ij  = Index + j;
            int_type IBj = Index_k + this->Index_BOUND[j];

            switch (this->BoundTypeArrayBOUND[j])
            {
            case DiscreteDefs::INEQUALITY:
                lambda_hesse[Ij] = lambda_IP[IBj] + lambda_IP[IBj + 1];
                break;
            case DiscreteDefs::EQUATION:
                lambda_hesse[Ij] = lambda_IP[IBj];
                break;
            case DiscreteDefs::LOWER_BARRIER:
                lambda_hesse[Ij] = -lambda_IP[IBj];
                break;
            case DiscreteDefs::UPPER_BARRIER:
                lambda_hesse[Ij] = lambda_IP[IBj];
                break;
            case DiscreteDefs::UNCONSTRAINED:
                lambda_hesse[Ij] = 0;
                break;
            }
        }

        // Belegen der NLCSTR-Beschränkungen:
        for (int_type i = 0; i < this->N + 1; ++i)
        {
            Index   = (this->N) * Ny + N_BOUND + i * N_NLCSTR;
            Index_k = (this->N) * Ny + (this->BoundDimBOUND.N_Dim_ROW) + i * (this->BoundDimNLCSTR.N_Dim_ROW + this->BoundDimSTATE.N_Dim_ROW + this->BoundDimCONTROL.N_Dim_ROW);
            for (int_type j = 0; j < N_NLCSTR; ++j)
            {

                int_type Ij  = Index + j;
                int_type INj = Index_k + this->Index_NLCSTR[j];

                switch (this->BoundTypeArrayNLCSTR[j])
                {
                case DiscreteDefs::INEQUALITY:
                    lambda_hesse[Ij] = lambda_IP[INj] + lambda_IP[INj + 1];
                    break;
                case DiscreteDefs::EQUATION:
                    lambda_hesse[Ij] = lambda_IP[INj];
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    lambda_hesse[Ij] = -lambda_IP[INj];
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    lambda_hesse[Ij] = lambda_IP[INj];
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    lambda_hesse[Ij] = 0;
                    break;
                }
            }
        }
    }

    template <typename real>
    bool DiscreteDimension<real>::get_ParameterConstraintsMatrix(MatrixWrapper<real> & Q)
    {
        if (this->C_Constr_P.numRows() == Q.numRows() && this->C_Constr_P.numCols() == Q.numCols())
        {
            Q.load(this->C_Constr_P);
            return true;
        }
        else
        {
            return false;
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_CaseLambda()
    {
        int_type Ny       = this->BoundDimSTATE.N;
        int_type Nu       = this->BoundDimCONTROL.N;
        int_type Np       = this->BoundDimPARAM.N;
        int_type N_BOUND  = this->BoundDimBOUND.N;
        int_type N_NLCSTR = this->BoundDimNLCSTR.N;
        int_type Index_k  = 0;
        // Belegen der ODE-Beschränkungen:
        for (int_type i = 0; i < (this->N) * Ny; ++i)
            this->Case_lambda_k[i] = 0;

        // Belegen der BOUND-Beschränkungen:
        Index_k = (this->N) * Ny;
        for (int_type j = 0; j < N_BOUND; ++j)
        {
            int_type IBj = Index_k + this->Index_BOUND[j];
            switch (this->getBoundTypeArrayBOUND(j))
            {
            case DiscreteDefs::INEQUALITY:
                this->Case_lambda_k[IBj]     = 1;
                this->Case_lambda_k[IBj + 1] = 1;
                break;
            case DiscreteDefs::EQUATION:
                this->Case_lambda_k[IBj] = 0;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                this->Case_lambda_k[IBj] = 1;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                this->Case_lambda_k[IBj] = 1;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                break;
            }
        }
        // Belegen der NLCSTR-Beschränkungen:
        int_type Index_STATE   = 0;
        int_type Index_CONTROL = 0;
        for (int_type i = 0; i < this->N + 1; ++i)
        {
            Index_k = (this->N) * Ny + (this->BoundDimBOUND.N_Dim_ROW) + i * (this->BoundDimNLCSTR.N_Dim_ROW + this->BoundDimSTATE.N_Dim_ROW + this->BoundDimCONTROL.N_Dim_ROW);
            for (int_type j = 0; j < N_NLCSTR; ++j)
            {

                int_type INj = Index_k + this->Index_NLCSTR[j];
                switch (this->BoundTypeArrayNLCSTR[j])
                {
                case DiscreteDefs::INEQUALITY:
                    this->Case_lambda_k[INj]     = 1;
                    this->Case_lambda_k[INj + 1] = 1;
                    break;
                case DiscreteDefs::EQUATION:
                    this->Case_lambda_k[INj] = 0;
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    this->Case_lambda_k[INj] = 1;
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    this->Case_lambda_k[INj] = 1;
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    break;
                }
            }
            // Belegen der STATE-Beschränkungen:
            Index_STATE = 0;
            for (int_type j = 0; j < Ny; ++j)
            {
                int_type II = Index_k + Index_STATE;
                switch (this->BoundTypeArraySTATE[j])
                {
                case DiscreteDefs::INEQUALITY:
                    this->Case_lambda_k[II]     = 1;
                    this->Case_lambda_k[II + 1] = 1;
                    Index_STATE += 2;
                    break;
                case DiscreteDefs::EQUATION:
                    this->Case_lambda_k[II] = 0;
                    ++Index_STATE;
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    this->Case_lambda_k[II] = 1;
                    ++Index_STATE;
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    this->Case_lambda_k[II] = 1;
                    ++Index_STATE;
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    break;
                }
            }
            // Belegen der CONTROL-Beschränkungen:
            Index_CONTROL = 0;
            for (int_type j = 0; j < Nu; ++j)
            {

                int_type III = Index_k + Index_STATE + Index_CONTROL;

                switch (this->BoundTypeArrayCONTROL[j])
                {
                case DiscreteDefs::INEQUALITY:
                    this->Case_lambda_k[III]     = 1;
                    this->Case_lambda_k[III + 1] = 1;
                    Index_CONTROL += 2;
                    break;
                case DiscreteDefs::EQUATION:
                    this->Case_lambda_k[III] = 0;
                    ++Index_CONTROL;
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    this->Case_lambda_k[III] = 1;
                    ++Index_CONTROL;
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    this->Case_lambda_k[III] = 1;
                    ++Index_CONTROL;
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    break;
                }
            }
        }
        // Belegen der Parameter-Beschränkungen:
        Index_k              = (this->N) * Ny + (this->BoundDimBOUND.N_Dim_ROW) + (this->N + 1) * (this->BoundDimNLCSTR.N_Dim_ROW + this->BoundDimSTATE.N_Dim_ROW + this->BoundDimCONTROL.N_Dim_ROW);
        int_type Index_PARAM = 0;
        for (int_type j = 0; j < Np; ++j)
        {
            int_type II = Index_k + Index_PARAM;
            switch (this->BoundTypeArrayPARAM[j])
            {
            case DiscreteDefs::INEQUALITY:
                this->Case_lambda_k[II]     = 1;
                this->Case_lambda_k[II + 1] = 1;
                Index_PARAM += 2;
                break;
            case DiscreteDefs::EQUATION:
                this->Case_lambda_k[II] = 0;
                ++Index_PARAM;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                this->Case_lambda_k[II] = 1;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                this->Case_lambda_k[II] = 1;
                ++Index_PARAM;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                break;
            }
        }
    }

    template <typename real>
    int_type DiscreteDimension<real>::getDimLambdaIP()
    {
        return this->DimLambdaIP;
    }

    template <typename real>
    int_type DiscreteDimension<real>::getDimSIP()
    {
        return this->Dim_Schlupf;
    }

    template <typename real>
    void DiscreteDimension<real>::set_C_Constr(MatrixWrapper<real> & G, MatrixWrapper<real> & C_)
    {
        C_.zero_fill();
        int_type DimSpalenC = G.numCols();
        if (DimSpalenC != C_.numCols())
        {
            Errors::set_Error("Column Dimension der Nebenbedingungen stimmt nicht überein!");
            return;
        }

        int_type N_NLCSTR = this->BoundDimNLCSTR.N;
        if (N_NLCSTR > 0)
        {
            for (int_type row = 0; row < N_NLCSTR; ++row)
            {
                int_type INr = this->Index_NLCSTR[row];
                switch (this->getBoundTypeArrayNLCSTR(row))
                {
                case DiscreteDefs::INEQUALITY:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(G, row, C_, INr);
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(G, row, C_, INr + 1);
                    break;
                case DiscreteDefs::EQUATION:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(G, row, C_, INr);
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(G, row, C_, INr);
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(G, row, C_, INr);
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    // Nichts zu tun!
                    break;
                }
            }
        }
        // Sets box bounds of states:
        C_.load_block(this->C_Constr_Y, this->BoundDimNLCSTR.N_Dim_ROW, 0);
        // Sets box bounds of controls:
        C_.load_block(this->C_Constr_U, this->BoundDimNLCSTR.N_Dim_ROW + this->BoundDimSTATE.N_Dim_ROW, this->BoundDimSTATE.N);
    }

    template <typename real>
    void DiscreteDimension<real>::set_Q_Constr(MatrixWrapper<real> & q, MatrixWrapper<real> & Q)
    {
        Q.zero_fill();
        int_type Np       = this->BoundDimPARAM.N;
        int_type N_NLCSTR = this->BoundDimNLCSTR.N;
        if (Np > 0 && N_NLCSTR > 0)
        {
            // Belege Teilmatrix der nichtlinearen Nebenbedingungen für Parameter:
            for (int_type row = 0; row < N_NLCSTR; ++row)
            {

                int_type INr = this->Index_NLCSTR[row];

                switch (this->BoundTypeArrayNLCSTR[row])
                {
                case DiscreteDefs::INEQUALITY:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(q, row, Q, INr);
                    // Q.CopyScMultRowFromMatrix(INr, row, q, 1.0);
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(q, row, Q, INr + 1);
                    // Q.CopyScMultRowFromMatrix(INr + 1, row, q, -1.0);
                    break;
                case DiscreteDefs::EQUATION:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(q, row, Q, INr);
                    // Q.CopyScMultRowFromMatrix(INr, row, q, 1.0);
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(q, row, Q, INr);
                    // Q.CopyScMultRowFromMatrix(INr, row, q, -1.0);
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(q, row, Q, INr);
                    // Q.CopyScMultRowFromMatrix(INr, row, q, 1.0);
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    // Nichts zu tun!
                    break;
                }
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_PSI(MatrixWrapper<real> & psi, MatrixWrapper<real> & PSI)
    {
        PSI.zero_fill();
        int_type N_BOUND = this->BoundDimBOUND.N;
        if (N_BOUND > 0)
        {
            // Belege Teilmatrix der nichtlinearen Nebenbedingungen für Parameter:
            for (int_type row = 0; row < N_BOUND; ++row)
            {

                int_type IBr = this->Index_BOUND[row];

                switch (this->BoundTypeArrayBOUND[row])
                {
                case DiscreteDefs::INEQUALITY:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(psi, row, PSI, IBr);
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(psi, row, PSI, IBr + 1);
                    break;
                case DiscreteDefs::EQUATION:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(psi, row, PSI, IBr);
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    // Kopiere negative Row:
                    OCPbasic::CopyNegRow<real>(psi, row, PSI, IBr);
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    // Kopiere Row:
                    OCPbasic::CopyRow<real>(psi, row, PSI, IBr);
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    // Nichts zu tun!
                    break;
                }
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_Vector_C_from_G(real const g[], real const y_k[], real const u_k[], Problem<real> const * P, real c[])
    {
        // Anteil der nichtlinearen Nebenbedingungen:
        int_type N_NLCSTR = this->BoundDimNLCSTR.N;
        if (N_NLCSTR > 0)
        {

            for (int_type row = 0; row < N_NLCSTR; ++row)
            {

                int_type INr = this->Index_NLCSTR[row];

                switch (this->getBoundTypeArrayNLCSTR(row))
                {
                case DiscreteDefs::INEQUALITY:
                    // Kopiere Row:
                    c[INr] = g[row] - P->g_max[row];
                    // Kopiere negative Row:
                    c[INr] = -g[row] + P->g_min[row];
                    break;
                case DiscreteDefs::EQUATION:
                    // Kopiere Row:
                    c[INr] = g[row] - P->g_max[row];
                    break;
                case DiscreteDefs::LOWER_BARRIER:
                    // Kopiere negative Row:
                    c[INr] = -g[row] + P->g_min[row];
                    break;
                case DiscreteDefs::UPPER_BARRIER:
                    // Kopiere Row:
                    c[INr] = g[row] - P->g_max[row];
                    break;
                case DiscreteDefs::UNCONSTRAINED:
                    // Nichts zu tun!
                    break;
                }
            }
        }

        // Belege Boxschranken für Zustände:
        int_type NState = this->getBoundDimSTATE().N;
        int_type iState = this->getBoundDimNLCSTR().N_Dim_ROW;
        for (int_type row = 0; row < NState; ++row)
        {
            switch (this->getBoundTypeArraySTATE(row))
            {
            case DiscreteDefs::INEQUALITY:
                // Kopiere Row:
                c[iState] = y_k[row] - P->y_max[row];
                ++iState;
                // Kopiere negative Row:
                c[iState] = -y_k[row] + P->y_min[row];
                ++iState;
                break;
            case DiscreteDefs::EQUATION:
                // Kopiere Row:
                c[iState] = y_k[row] - P->y_max[row];
                ++iState;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                // Kopiere negative Row:
                c[iState] = -y_k[row] + P->y_min[row];
                ++iState;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                // Kopiere Row:
                c[iState] = y_k[row] - P->y_max[row];
                ++iState;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                // Nichts zu tun!
                break;
            }
        }
        // Belege Boxschranken für Steuerungen:
        int_type NControl = this->getBoundDimCONTROL().N;
        for (int_type row = 0; row < NControl; ++row)
        {
            switch (this->getBoundTypeArrayCONTROL(row))
            {
            case DiscreteDefs::INEQUALITY:
                // Kopiere Row:
                c[iState] = u_k[row] - P->u_max[row];
                ++iState;
                // Kopiere negative Row:
                c[iState] = -u_k[row] + P->u_min[row];
                ++iState;
                break;
            case DiscreteDefs::EQUATION:
                // Kopiere Row:
                c[iState] = u_k[row] - P->u_max[row];
                ++iState;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                // Kopiere negative Row:
                c[iState] = -u_k[row] + P->u_min[row];
                ++iState;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                // Kopiere Row:
                c[iState] = u_k[row] - P->u_max[row];
                ++iState;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                // Nichts zu tun!
                break;
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_Vector_IPbound_from_bound(real const bound[], Problem<real> const * P, real OUTBound[])
    {
        int_type NBound = this->getBoundDimBOUND().N;
        for (int_type row = 0; row < NBound; ++row)
        {

            int_type IBr = this->Index_BOUND[row];

            switch (this->getBoundTypeArrayBOUND(row))
            {
            case DiscreteDefs::INEQUALITY:
                // Kopiere Row:
                OUTBound[IBr] = bound[row] - P->psi_max[row];
                // Kopiere negative Row:
                OUTBound[IBr] = -bound[row] + P->psi_min[row];
                break;
            case DiscreteDefs::EQUATION:
                // Kopiere Row:
                OUTBound[IBr] = bound[row] - P->psi_max[row];
                break;
            case DiscreteDefs::LOWER_BARRIER:
                // Kopiere negative Row:
                OUTBound[IBr] = -bound[row] + P->psi_min[row];
                break;
            case DiscreteDefs::UPPER_BARRIER:
                // Kopiere Row:
                OUTBound[IBr] = bound[row] - P->psi_max[row];
                break;
            case DiscreteDefs::UNCONSTRAINED:
                // Nichts zu tun!
                break;
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::get_isEquVec(bool isEquVec[])
    {
        for (int_type i = 0; i < this->DimLambdaIP; ++i)
            isEquVec[i] = this->Case_lambda_k[i] == 0;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                  private Memberfunktionen                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void DiscreteDimension<real>::setIndexArrays(Problem<real> * P)
    {
        bool bmax = false;
        bool bmin = false;
        // Nichtlineare Nebenbedingungen:
        for (int_type i = 0; i < P->N_NLCSTR; ++i)
        {
            // Überprüfung für Gleichungsbedingungen:
            if (std::abs(P->g_max[i] - P->g_min[i]) <= this->TOL)
            {
                this->BoundTypeArrayNLCSTR[i] = DiscreteDefs::EQUATION;
                this->BoundDimNLCSTR.N_Dim_ROW++;
                this->BoundDimNLCSTR.N_GL++;
            }
            else
            { // Überprüfe INF und -INF
                // Identifikation der Grenzen:
                if (P->g_max[i] > this->INF)
                    bmax = true;
                if (P->g_min[i] < -this->INF)
                    bmin = true;
                // Fallunterscheidung:
                if (bmax && bmin)
                {
                    this->BoundTypeArrayNLCSTR[i] = DiscreteDefs::UNCONSTRAINED;
                    this->BoundDimNLCSTR.N_ZL++;
                }
                else if (bmax)
                {
                    this->BoundTypeArrayNLCSTR[i] = DiscreteDefs::LOWER_BARRIER;
                    this->BoundDimNLCSTR.N_Dim_ROW++;
                    this->BoundDimNLCSTR.N_US++;
                }
                else if (bmin)
                {
                    this->BoundTypeArrayNLCSTR[i] = DiscreteDefs::UPPER_BARRIER;
                    this->BoundDimNLCSTR.N_Dim_ROW++;
                    this->BoundDimNLCSTR.N_OS++;
                }
                else
                {
                    this->BoundTypeArrayNLCSTR[i] = DiscreteDefs::INEQUALITY;
                    this->BoundDimNLCSTR.N_Dim_ROW += 2;
                    this->BoundDimNLCSTR.N_UGL++;
                }
                // Zurücksetzen der Variablen:
                bmax = false;
                bmin = false;
            }
        }

        // Randbedingungen:
        for (int_type i = 0; i < P->N_BOUND; ++i)
        {
            // Überprüfung für Gleichungsbedingungen:
            if (std::abs(P->psi_min[i] - P->psi_max[i]) <= this->TOL)
            {
                this->BoundTypeArrayBOUND[i] = DiscreteDefs::EQUATION;
                this->BoundDimBOUND.N_Dim_ROW++;
                this->BoundDimBOUND.N_GL++;
            }
            else
            { // Überprüfe INF und -INF
                // Identifikation der Grenzen:
                if (P->psi_max[i] > this->INF)
                    bmax = true;
                if (P->psi_min[i] < -this->INF)
                    bmin = true;
                // Fallunterscheidung:
                if (bmax && bmin)
                {
                    this->BoundTypeArrayBOUND[i] = DiscreteDefs::UNCONSTRAINED;
                    this->BoundDimBOUND.N_ZL++;
                }
                else if (bmax)
                {
                    this->BoundTypeArrayBOUND[i] = DiscreteDefs::LOWER_BARRIER;
                    this->BoundDimBOUND.N_Dim_ROW++;
                    this->BoundDimBOUND.N_US++;
                }
                else if (bmin)
                {
                    this->BoundTypeArrayBOUND[i] = DiscreteDefs::UPPER_BARRIER;
                    this->BoundDimBOUND.N_Dim_ROW++;
                    this->BoundDimBOUND.N_OS++;
                }
                else
                {
                    this->BoundTypeArrayBOUND[i] = DiscreteDefs::INEQUALITY;
                    this->BoundDimBOUND.N_Dim_ROW += 2;
                    this->BoundDimBOUND.N_UGL++;
                }
                // Zurücksetzen der Variablen:
                bmax = false;
                bmin = false;
            }
        }

        // Zustandsbeschränkungen:
        for (int_type i = 0; i < P->Ny; ++i)
        {
            // Überprüfung für Gleichungsbedingungen:
            if (std::abs(P->y_min[i] - P->y_max[i]) <= this->TOL)
            {
                this->BoundTypeArraySTATE[i] = DiscreteDefs::EQUATION;
                this->BoundDimSTATE.N_Dim_ROW++;
                this->BoundDimSTATE.N_GL++;
            }
            else
            { // Überprüfe INF und -INF
                // Identifikation der Grenzen:
                if (P->y_max[i] > this->INF)
                    bmax = true;
                if (P->y_min[i] < -this->INF)
                    bmin = true;
                // Fallunterscheidung:
                if (bmax && bmin)
                {
                    this->BoundTypeArraySTATE[i] = DiscreteDefs::UNCONSTRAINED;
                    this->BoundDimSTATE.N_ZL++;
                }
                else if (bmax)
                {
                    this->BoundTypeArraySTATE[i] = DiscreteDefs::LOWER_BARRIER;
                    this->BoundDimSTATE.N_Dim_ROW++;
                    this->BoundDimSTATE.N_US++;
                }
                else if (bmin)
                {
                    this->BoundTypeArraySTATE[i] = DiscreteDefs::UPPER_BARRIER;
                    this->BoundDimSTATE.N_Dim_ROW++;
                    this->BoundDimSTATE.N_OS++;
                }
                else
                {
                    this->BoundTypeArraySTATE[i] = DiscreteDefs::INEQUALITY;
                    this->BoundDimSTATE.N_Dim_ROW += 2;
                    this->BoundDimSTATE.N_UGL++;
                }
                // Zurücksetzen der Variablen:
                bmax = false;
                bmin = false;
            }
        }

        // Steuerungsbeschränkungen:
        for (int_type i = 0; i < P->Nu; ++i)
        {
            // Überprüfung für Gleichungsbedingungen:
            if (std::abs(P->u_min[i] - P->u_max[i]) <= this->TOL)
            {
                this->BoundTypeArrayCONTROL[i] = DiscreteDefs::EQUATION;
                this->BoundDimCONTROL.N_Dim_ROW++;
                this->BoundDimCONTROL.N_GL++;
            }
            else
            { // Überprüfe INF und -INF
                // Identifikation der Grenzen:
                if (P->u_max[i] > this->INF)
                    bmax = true;
                if (P->u_min[i] < -this->INF)
                    bmin = true;
                // Fallunterscheidung:
                if (bmax && bmin)
                {
                    this->BoundTypeArrayCONTROL[i] = DiscreteDefs::UNCONSTRAINED;
                    this->BoundDimCONTROL.N_ZL++;
                }
                else if (bmin)
                {
                    this->BoundTypeArrayCONTROL[i] = DiscreteDefs::LOWER_BARRIER;
                    this->BoundDimCONTROL.N_Dim_ROW++;
                    this->BoundDimCONTROL.N_US++;
                }
                else if (bmax)
                {
                    this->BoundTypeArrayCONTROL[i] = DiscreteDefs::UPPER_BARRIER;
                    this->BoundDimCONTROL.N_Dim_ROW++;
                    this->BoundDimCONTROL.N_OS++;
                }
                else
                {
                    this->BoundTypeArrayCONTROL[i] = DiscreteDefs::INEQUALITY;
                    this->BoundDimCONTROL.N_Dim_ROW += 2;
                    this->BoundDimCONTROL.N_UGL++;
                }
                // Zurücksetzen der Variablen:
                bmax = false;
                bmin = false;
            }
        }

        // Parameterbeschränkungen:
        for (int_type i = 0; i < P->Np; ++i)
        {
            // Überprüfung für Gleichungsbedingungen:
            if (std::abs(P->p_min[i] - P->p_max[i]) <= this->TOL)
            {
                this->BoundTypeArrayPARAM[i] = DiscreteDefs::EQUATION;
                this->BoundDimPARAM.N_Dim_ROW++;
                this->BoundDimPARAM.N_GL++;
            }
            else
            { // Überprüfe INF und -INF
                // Identifikation der Grenzen:
                if (P->p_max[i] > this->INF)
                    bmax = true;
                if (P->p_min[i] < -this->INF)
                    bmin = true;
                // Fallunterscheidung:
                if (bmax && bmin)
                {
                    this->BoundTypeArrayPARAM[i] = DiscreteDefs::UNCONSTRAINED;
                    this->BoundDimPARAM.N_ZL++;
                }
                else if (bmin)
                {
                    this->BoundTypeArrayPARAM[i] = DiscreteDefs::LOWER_BARRIER;
                    this->BoundDimPARAM.N_Dim_ROW++;
                    this->BoundDimPARAM.N_US++;
                }
                else if (bmax)
                {
                    this->BoundTypeArrayPARAM[i] = DiscreteDefs::UPPER_BARRIER;
                    this->BoundDimPARAM.N_Dim_ROW++;
                    this->BoundDimPARAM.N_OS++;
                }
                else
                {
                    this->BoundTypeArrayPARAM[i] = DiscreteDefs::INEQUALITY;
                    this->BoundDimPARAM.N_Dim_ROW += 2;
                    this->BoundDimPARAM.N_UGL++;
                }
                // Zurücksetzen der Variablen:
                bmax = false;
                bmin = false;
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_internal_Constr_YUP()
    {
        // Belege Teilmatrix für die Zustände unter Berücksichtigung der Schranken:
        this->set_internal_linear_Constr(&this->BoundDimSTATE, &this->BoundTypeArraySTATE.front(), this->C_Constr_Y);
        // Belege Teilmatrix für die Steuerungen unter Berücksichtigung der Schranken:
        this->set_internal_linear_Constr(&this->BoundDimCONTROL, &this->BoundTypeArrayCONTROL.front(), this->C_Constr_U);

        if (this->BoundDimPARAM.N > 0)
        {
            // Belege Teilmatrix für die Parameter unter Berücksichtigung der Schranken:
            this->set_internal_linear_Constr(&this->BoundDimPARAM, &this->BoundTypeArrayPARAM.front(), this->C_Constr_P);
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_internal_linear_Constr(DiscreteDefs::BoundDimension * BoundDim, DiscreteDefs::BoundType const BOUNDTYPE[], MatrixWrapper<real> & M)
    {
        int_type index = 0;
        // Belege Teilmatrix unter Berücksichtigung der Schranken:
        for (int_type i = 0; i < BoundDim->N; ++i)
        {
            switch (BOUNDTYPE[i])
            {
            case DiscreteDefs::INEQUALITY:
                M(index, i) = 1.0;
                ++index;
                M(index, i) = -1.0;
                ++index;
                break;
            case DiscreteDefs::EQUATION:
                M(index, i) = 1.0;
                ++index;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                M(index, i) = -1.0;
                ++index;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                M(index, i) = 1.0;
                ++index;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                // Nichts zu tun!
                break;
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_internal_Index_NLCSTR()
    {
        int_type index = 0;
        // Belege Indexarray unter Berücksichtigung der Schranken:
        for (int_type i = 0; i < this->BoundDimNLCSTR.N; ++i)
        {
            switch (this->BoundTypeArrayNLCSTR[i])
            {
            case DiscreteDefs::INEQUALITY:
                this->Index_NLCSTR[i] = index;
                index += 2;
                break;
            case DiscreteDefs::EQUATION:
                this->Index_NLCSTR[i] = index;
                index++;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                this->Index_NLCSTR[i] = index;
                index++;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                this->Index_NLCSTR[i] = index;
                index++;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                this->Index_NLCSTR[i] = index;
                break;
            }
        }
    }

    template <typename real>
    void DiscreteDimension<real>::set_internal_Index_BOUND()
    {
        int_type index = 0;
        // Belege Indexarray unter Berücksichtigung der Schranken:
        for (int_type i = 0; i < this->BoundDimBOUND.N; ++i)
        {
            switch (this->BoundTypeArrayBOUND[i])
            {
            case DiscreteDefs::INEQUALITY:
                this->Index_BOUND[i] = index;
                index += 2;
                break;
            case DiscreteDefs::EQUATION:
                this->Index_BOUND[i] = index;
                index++;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                this->Index_BOUND[i] = index;
                index++;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                this->Index_BOUND[i] = index;
                index++;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                this->Index_BOUND[i] = index;
                break;
            }
        }
    }

    template class DiscreteDimension<double>;
    template class DiscreteDimension<float>;

} // namespace OCPbasic
