/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file trapez.cpp
 * Implementation of the class Trapez.
 *
 * @author A. Huber
 * @since 27.09.2018
 */

#include "trapez.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#if defined(IPBASIC_USE_OPENMP)
#include <omp.h>
#endif

#if defined(IPBASIC_USE_LAPACK_BLAS)
#include <IPbasic/IPbasic.h>
#endif

namespace OCPbasic
{

    using IPbasic::Errors;

    template <typename real>
    Trapez<real>::Trapez(int_type N, real t0, real tF, Problem<real> * P)
        : Discretization<real>(N, t0, tF, P)
        , mem_real("Trapez")
    {
        int_type & Ny       = this->P->Ny;
        int_type & Np       = this->P->Np;
        int_type & Nu       = this->P->Nu;
        int_type   Nyu      = this->P->Ny + this->P->Nu;
        int_type   Nyup     = this->P->Ny + this->P->Nu + this->P->Np;
        int_type & N_NLCSTR = this->P->N_NLCSTR;
        int_type & N_BOUND  = this->P->N_BOUND;

        int_type rowDimC = this->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->MyDims->getBoundDimSTATE().N_Dim_ROW + this->MyDims->getBoundDimCONTROL().N_Dim_ROW;

        int_type NRowQ = this->MyDims->getBoundDimPARAM().N_Dim_ROW;
        int_type NColQ = this->MyDims->getBoundDimPARAM().N;

        this->h                  = (this->tF - this->t0) / this->N;
        this->N_Statevector      = (this->N + 1) * (Ny + Nu) + Np;
        this->N_Constraintvector = Ny * (this->N) + N_NLCSTR * (this->N + 1) + N_BOUND;

        int_type nnz = this->N_Statevector + this->N_Constraintvector + Nyu * Nyu + Np * Np + (this->N + 1) * Nyu * (Nyup) + this->N * Ny * (2 * Nyu + Np) + 2 * N_BOUND * Nyu +
                       2 * this->MyDims->getBoundDimBOUND().N_Dim_ROW * Nyu + N_BOUND * Np + (this->N + 1) * ((N_NLCSTR + rowDimC) * (Nyup)) + NRowQ * NColQ +
                       this->MyDims->getBoundDimBOUND().N_Dim_ROW * Np + Ny * Ny * 3 + Ny * Nu * 5 + Np * Np + Ny * Np * 3 + Np * Nu * 3 + Nu * Nu * 3 + N_NLCSTR * (Ny + Nu) + 4 * Ny + Np +
                       P->N_NLCSTR + N_BOUND + this->P->Ny + this->P->Nu;

        // Allokiere Hilfsmatrizen:
        mem_real.allocate(nnz);

        this->x      = mem_real(size_t(this->N_Statevector));
        this->lambda = mem_real(size_t(this->N_Constraintvector));

        // init with 0:
        std::fill_n(this->x, this->N_Statevector, 0);
        std::fill_n(this->lambda, this->N_Constraintvector, 0);

        // Calculate Array Dimensions:
        {
            std::vector<int_type> iDimsHesse(this->N + 2);
            std::vector<int_type> iDimsJacobi(2 * (this->N + 1) + 1);
            this->get_iDimHesse(&iDimsHesse.front());
            this->get_iDimJacobi(&iDimsJacobi.front());
            this->HesseMatrix  = new CollectorMatrix<real>(this->N + 2, this->N + 2, &iDimsHesse.front(), &iDimsHesse.front(), SYM);
            this->JacobiMatrix = new CollectorMatrix<real>(2 * (this->N + 1) + 1, this->N + 2, &iDimsJacobi.front(), &iDimsHesse.front(), NOT_SYM);
        }

        // Belege Speicher der Matrizen:
        // Hessematrix:
        this->Array_W.resize(this->N + 1);
        this->Array_P.resize(this->N + 1);

#define NEW_MATW(NR, NC) new MatrixWrapper<real>(mem_real(size_t((NR) * (NC))), NR, NC, NR)
#define SETUP_MATW(NR, NC) setup(mem_real(size_t((NR) * (NC))), NR, NC, NR)

        this->PHI.SETUP_MATW(Nyu, Nyu);
        this->PHI.zero_fill();

        this->MP.SETUP_MATW(Np, Np);
        this->MP.zero_fill();

        this->HesseMatrix->PushMatrix(this->N, 0, &this->PHI);
        this->HesseMatrix->PushMatrix(this->N + 1, this->N + 1, &this->MP);

        for (int_type i = 0; i <= this->N; ++i)
        {
            this->Array_W[i].SETUP_MATW(Nyu, Nyu);
            this->Array_P[i].SETUP_MATW(Nyu, Np);
            this->Array_W[i].zero_fill();
            this->Array_P[i].zero_fill();
            this->HesseMatrix->PushMatrix(i, i, &this->Array_W[i]);
            this->HesseMatrix->PushMatrixTransposed(this->N + 1, i, &this->Array_P[i]);
        }

        // Jacobimatrix:
        // Matrizen der ODE Constraints:
        this->Array_M.resize(this->N);
        this->Array_S.resize(this->N);
        this->Array_R.resize(this->N);
        for (int_type i = 0; i < this->N; i++)
        {
            this->Array_M[i].SETUP_MATW(Ny, Nyu);
            this->Array_S[i].SETUP_MATW(Ny, Nyu);
            this->Array_R[i].SETUP_MATW(Ny, Np);
            this->Array_M[i].zero_fill();
            this->Array_S[i].zero_fill();
            this->Array_R[i].zero_fill();
            this->JacobiMatrix->PushMatrix(i, i, &this->Array_M[i]);
            this->JacobiMatrix->PushMatrix(i, i + 1, &this->Array_S[i]);
            this->JacobiMatrix->PushMatrix(i, this->N + 1, &this->Array_R[i]);
        }

        // Matrizen der Boundary Constraints:
        this->b_Lower.SETUP_MATW(N_BOUND, Nyu);
        this->b_Upper.SETUP_MATW(N_BOUND, Nyu);
        this->PSI_Lower.SETUP_MATW(this->MyDims->getBoundDimBOUND().N_Dim_ROW, Nyu);
        this->PSI_Upper.SETUP_MATW(this->MyDims->getBoundDimBOUND().N_Dim_ROW, Nyu);
        this->R.SETUP_MATW(N_BOUND, Np);
        this->b_Lower.zero_fill();
        this->b_Upper.zero_fill();
        this->PSI_Lower.zero_fill();
        this->PSI_Upper.zero_fill();
        this->R.zero_fill();
        this->JacobiMatrix->PushMatrix(this->N, 0, &this->PSI_Lower);
        this->JacobiMatrix->PushMatrix(this->N, this->N, &this->PSI_Upper);

        // Matrizen der nonlinear Constraints:
        this->Array_G.resize(this->N + 1);
        this->Array_q.resize(this->N + 1);
        this->Array_C_CONSTR.resize(this->N + 1);
        this->Array_Q.resize(this->N + 1);
        // int_type rowDimC = this->MyDims->getBoundDimNLCSTR().N_Dim_ROW +
        //                   this->MyDims->getBoundDimSTATE().N_Dim_ROW +
        //                   this->MyDims->getBoundDimCONTROL().N_Dim_ROW;
        for (int_type i = 0; i <= this->N; ++i)
        {
            this->Array_G[i].SETUP_MATW(N_NLCSTR, Nyu);
            this->Array_q[i].SETUP_MATW(N_NLCSTR, Np);
            this->Array_C_CONSTR[i].SETUP_MATW(rowDimC, Nyu);
            this->Array_Q[i].SETUP_MATW(rowDimC, Np);
            this->Array_G[i].zero_fill();
            this->Array_q[i].zero_fill();
            this->Array_C_CONSTR[i].zero_fill();
            this->Array_Q[i].zero_fill();
            this->JacobiMatrix->PushMatrix(this->N + 1 + i, i, &this->Array_C_CONSTR[i]);
            this->JacobiMatrix->PushMatrix(this->N + 1 + i, this->N + 1, &this->Array_Q[i]);
        }

        // Parameter Beschränkungen:
        // int_type NRowQ = this->MyDims->getBoundDimPARAM().N_Dim_ROW;
        // int_type NColQ = this->MyDims->getBoundDimPARAM().N;
        this->Q.SETUP_MATW(NRowQ, NColQ);
        this->R_Hat.SETUP_MATW(this->MyDims->getBoundDimBOUND().N_Dim_ROW, Np);
        this->R_Hat.zero_fill();
        if (!this->MyDims->get_ParameterConstraintsMatrix(this->Q))
        {
            IPbasic::Errors::set_Error("Dimension der Parametermatrix Q ist falsch!");
        }
        this->JacobiMatrix->PushMatrix(this->N, this->N + 1, &this->R_Hat);

        // Allokiere Hilfsmatrizen:
        this->EvalHelpNyNy1.SETUP_MATW(Ny, Ny);
        this->EvalHelpNyNy2.SETUP_MATW(Ny, Ny);
        this->EvalHelpNyNy3.SETUP_MATW(Ny, Ny);
        this->EvalHelpNyNu1.SETUP_MATW(Ny, Nu);
        this->EvalHelpNyNu2.SETUP_MATW(Ny, Nu);
        this->EvalHelpNyNu3.SETUP_MATW(Ny, Nu);
        this->EvalHelpNyNp1.SETUP_MATW(Ny, Np);
        this->EvalHelpNyNp2.SETUP_MATW(Ny, Np);
        this->EvalHelpNpNp1.SETUP_MATW(Np, Np);
        this->EvalHelpNpNy1.SETUP_MATW(Np, Ny);
        this->EvalHelpNpNu1.SETUP_MATW(Np, Nu);
        this->EvalHelpNuNu1.SETUP_MATW(Nu, Nu);
        this->EvalHelpNuNu2.SETUP_MATW(Nu, Nu);
        this->EvalHelpNuNu3.SETUP_MATW(Nu, Nu);
        this->EvalHelpNuNp1.SETUP_MATW(Nu, Np);
        this->EvalHelpNuNp2.SETUP_MATW(Nu, Np);
        this->EvalHelpNuNy1.SETUP_MATW(Nu, Ny);
        this->EvalHelpNuNy2.SETUP_MATW(Nu, Ny);
        this->EvalHelp_N_NLCTR_Ny.SETUP_MATW(N_NLCSTR, Ny);
        this->EvalHelp_N_NLCTR_Nu.SETUP_MATW(N_NLCSTR, Nu);
        this->EvalHelp1Ny1.SETUP_MATW(1, Ny);
        this->EvalHelp1Ny2.SETUP_MATW(1, Ny);
        this->EvalHelp1Np1.SETUP_MATW(1, Np);
        this->EvalHelpNyNy1.zero_fill();
        this->EvalHelpNyNy2.zero_fill();
        this->EvalHelpNyNy3.zero_fill();
        this->EvalHelpNyNu1.zero_fill();
        this->EvalHelpNyNu2.zero_fill();
        this->EvalHelpNyNu3.zero_fill();
        this->EvalHelpNyNp1.zero_fill();
        this->EvalHelpNyNp2.zero_fill();
        this->EvalHelpNpNp1.zero_fill();
        this->EvalHelpNpNy1.zero_fill();
        this->EvalHelpNpNu1.zero_fill();
        this->EvalHelpNuNu1.zero_fill();
        this->EvalHelpNuNu2.zero_fill();
        this->EvalHelpNuNu3.zero_fill();
        this->EvalHelpNuNp1.zero_fill();
        this->EvalHelpNuNp2.zero_fill();
        this->EvalHelpNuNy1.zero_fill();
        this->EvalHelpNuNy2.zero_fill();
        this->EvalHelp_N_NLCTR_Ny.zero_fill();
        this->EvalHelp_N_NLCTR_Nu.zero_fill();
        this->EvalHelp1Ny1.zero_fill();
        this->EvalHelp1Ny2.zero_fill();
        this->EvalHelp1Np1.zero_fill();

        // Allokiere Hilfsvekoren:
        this->EvalHelp_f_k    = this->mem_real(size_t(Ny));
        this->EvalHelp_f_kk   = this->mem_real(size_t(Ny));
        this->EvalHelp_g      = this->mem_real(size_t(P->N_NLCSTR));
        this->EvalHelp_psi    = this->mem_real(size_t(P->N_BOUND));
        this->EvalHelp_eigens = this->mem_real(size_t(this->P->Ny + this->P->Nu));
    }

    template <typename real>
    Trapez<real>::~Trapez()
    {
        // Loesche Matrizen:
        delete this->HesseMatrix;
        delete this->JacobiMatrix;

        // Loesche Daten der Hessematrix:
        this->Array_W.clear();
        this->Array_P.clear();
        this->Array_M.clear();
        this->Array_S.clear();
        this->Array_R.clear();
        this->Array_G.clear();
        this->Array_q.clear();
        this->Array_C_CONSTR.clear();
        this->Array_Q.clear();

        this->mem_real.free();
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                     public Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Trapez<real>::regularizeHessian()
    {
        this->LocalTimerDiscretization.tic();
        if (this->MyOCP_Params->MyRegularization == OCPbasic::REGULARIZE_NONE)
        {
            return;
        }
        else if (this->MyOCP_Params->MyRegularization == OCPbasic::REGULARIZE_GERSCHGORIN)
        {
            // Done Later:
            return;
        }
        else if (this->MyOCP_Params->MyRegularization == OCPbasic::REGULARIZE_PARTIAL_EIGENS)
        {
            int_type Nyu         = this->P->Ny + this->P->Nu;
            real     lowestEigen = real(0.0);

#if defined(IPBASIC_USE_OPENMP) && defined(IPBASIC_USE_LAPACK_BLAS)
            if (omp_get_proc_bind() == omp_proc_bind_false)
            {
                real * eigens = this->EvalHelp_eigens;

                for (int_type i = 0; i <= this->N; i++)
                {
                    int_type info = SymEigen.get_Symmetric_Eigenvalues(this->Array_W[i], eigens);
                    if (info == 0)
                    {
                        for (int_type k = 0; k < Nyu; k++)
                        {
                            if (lowestEigen > eigens[k])
                            {
                                lowestEigen = eigens[k];
                            }
                        }
                    }
                    else
                    {
                        Errors::set_Warning("Could not solve local eigenvalue problem!");
                    }
                }
            }
            else
            {
#pragma omp parallel shared(Nyu, lowestEigen) default(none)
                {
                    std::vector<real> vecEigens;
                    vecEigens.resize(Nyu);
                    std::vector<real> vecHlpMatrix;
                    vecHlpMatrix.resize(Nyu * Nyu);
                    std::vector<real> vecWork;
                    int_type          lwork = 3 * Nyu;
                    vecWork.resize(lwork);
                    real * eigens = vecEigens.data();
                    real * matr   = vecHlpMatrix.data();
                    real * work   = vecWork.data();
                    char   jobz   = 'N';
                    char   uplo   = 'L';
#pragma omp for
                    for (int_type i = 0; i <= this->N; i++)
                    {
                        std::copy_n(this->Array_W[i].get_data(), Nyu * Nyu, matr);
                        int_type info = 0;
                        IPbasic::syev(&jobz, &uplo, &Nyu, matr, &Nyu, eigens, work, &lwork, &info);
// IPbasic::
#pragma omp critical
                        {
                            if (info == 0)
                            {
                                for (int_type k = 0; k < Nyu; k++)
                                {
                                    if (lowestEigen > eigens[k])
                                    {
                                        lowestEigen = eigens[k];
                                    }
                                }
                            }
                            else
                            {
                                Errors::set_Warning("Could not solve local eigenvalue problem!");
                            }
                        }
                    }
                }
            }
#else
            real * eigens = this->EvalHelp_eigens;

            for (int_type i = 0; i <= this->N; i++)
            {
                int_type info = SymEigen.get_Symmetric_Eigenvalues(this->Array_W[i], eigens);
                if (info == 0)
                {
                    for (int_type k = 0; k < Nyu; k++)
                    {
                        if (lowestEigen > eigens[k])
                        {
                            lowestEigen = eigens[k];
                        }
                    }
                }
                else
                {
                    Errors::set_Warning("Could not solve local eigenvalue problem!");
                }
            }
#endif
            if (lowestEigen <= real(-0.1))
            {
                for (int_type i = 0; i <= this->N; i++)
                {
                    for (int_type k = 0; k < Nyu; k++)
                    {
                        this->Array_W[i](k, k) += -lowestEigen;
                    }
                }
            }
        }
        else
        {
            Errors::set_Warning("Unknown regularization of hessian!");
        }
        this->LocalTimerDiscretization.toc();
        this->MyOCP_Timings->TimeRegularizeHessian += this->LocalTimerDiscretization.elapsed_s();
    }

    template <typename real>
    bool Trapez<real>::writeToSolution(Solution<real> & Sol)
    {
        if (this->N + 1 != Sol.grid_X && this->N + 1 != Sol.grid_U)
            return false;
        int_type Nx = this->MyDims->getBoundDimSTATE().N;
        int_type Nu = this->MyDims->getBoundDimCONTROL().N;
        int_type Np = this->MyDims->getBoundDimPARAM().N;
        for (int_type i = 0; i <= this->N; ++i)
        {
            real * yk = this->getYk(i);
            real * uk = this->getUk(i);
            memcpy(Sol.X[i], yk, Nx * sizeof(real));
            memcpy(Sol.U[i], uk, Nu * sizeof(real));
        }
        // Write parameter:
        if (Np > 0)
        {
            if (Np != Sol.dim_P)
                return false;
            real * p = this->getP();
            memcpy(Sol.P, p, Np * sizeof(real));
        }
        return true;
    }

    template <typename real>
    bool Trapez<real>::readFromSolution(Solution<real> & Sol)
    {
        if (this->N + 1 != Sol.grid_X && this->N + 1 != Sol.grid_U)
            return false;
        int_type Nx = this->MyDims->getBoundDimSTATE().N;
        int_type Nu = this->MyDims->getBoundDimCONTROL().N;
        int_type Np = this->MyDims->getBoundDimPARAM().N;
        for (int_type i = 0; i <= this->N; ++i)
        {
            real * yk = this->getYk(i);
            real * uk = this->getUk(i);
            memcpy(yk, Sol.X[i], Nx * sizeof(real));
            memcpy(uk, Sol.U[i], Nu * sizeof(real));
        }
        // Read parameter:
        if (Np > 0)
        {
            if (Np != Sol.dim_P)
                return false;
            real * p = this->getP();
            memcpy(p, Sol.P, Np * sizeof(real));
        }
        return true;
    }

    template <typename real>
    void Trapez<real>::update_IPData(const real * x, const real * lambda_IP)
    {
        // Update der Hessematrix:
        this->update_Hessian(x, lambda_IP);
        // Update der Jacobimatrix:
        this->update_Jacobian(x);
    }

    template <typename real>
    real * Trapez<real>::getP()
    {
        if (this->P->Np > 0)
            return &(this->x[(this->N + 1) * (this->P->Ny + this->P->Nu)]);
        else
            return &(this->x[0]);
    }

    template <typename real>
    real * Trapez<real>::getYk(int_type Index)
    {
        return &(this->x[Index * (this->P->Ny + this->P->Nu)]);
    }

    template <typename real>
    real * Trapez<real>::getUk(int_type Index)
    {
        return &(this->x[(Index + 1) * this->P->Ny + Index * this->P->Nu]);
    }

    template <typename real>
    real * Trapez<real>::getLAMBDAk(int_type Index)
    {
        return &(this->lambda[Index * this->P->Ny]);
    }

    template <typename real>
    real * Trapez<real>::getSIGMA()
    {
        return &(this->lambda[(this->N) * this->P->Ny]);
    }

    template <typename real>
    real * Trapez<real>::getMUk(int_type Index)
    {
        return &(this->lambda[(this->N) * this->P->Ny + this->P->N_BOUND + Index * this->P->N_NLCSTR]);
    }

    template <typename real>
    real Trapez<real>::getVARPHI(const real * states)
    {
        int_type     Index = 0;
        const real * y0    = &states[Index * (this->P->Ny + this->P->Nu)];
        Index              = this->N;
        const real * yF    = &states[Index * (this->P->Ny + this->P->Nu)];
        const real * p;
        if (this->P->Np > 0)
            p = &(states[(this->N + 1) * (this->P->Ny + this->P->Nu)]);
        else
            p = &(states[0]);
        real VARPHI;
        this->P->VARPHI(this->t0, this->tF, yF, y0, p, &VARPHI);
        return VARPHI;
    }

    template <typename real>
    void Trapez<real>::Grad_und_VARPHI(const real * states, real * VARPHI, real * Grad)
    {
        int_type     N     = this->getAnzVar();
        int_type     Index = 0;
        const real * y0    = &states[Index * (this->P->Ny + this->P->Nu)];
        Index              = this->N;
        const real * yF    = &states[Index * (this->P->Ny + this->P->Nu)];
        const real * p;
        if (this->P->Np > 0)
            p = &(states[(this->N + 1) * (this->P->Ny + this->P->Nu)]);
        else
            p = &(states[0]);

        this->P->VARPHI(this->t0, this->tF, yF, y0, p, VARPHI);

        for (int_type i = 0; i < N; ++i)
            Grad[i] = 0.0;

        this->EvalHelp1Ny1.zero_fill();
        this->EvalHelp1Ny2.zero_fill();
        this->EvalHelp1Np1.zero_fill();
        this->P->D_VARPHI_X0(this->t0, this->tF, yF, y0, p, this->EvalHelp1Ny1);
        this->P->D_VARPHI_XF(this->t0, this->tF, yF, y0, p, this->EvalHelp1Ny2);
        this->P->D_VARPHI_P(this->t0, this->tF, yF, y0, p, this->EvalHelp1Np1);
        // Kopiere Vektor X0:
        int_type NX0 = 0;
        memcpy(&Grad[NX0], this->EvalHelp1Ny1.get_data(), this->P->Ny * sizeof(real));
        // Kopiere Vektor XF:
        int_type NXF = (this->N) * this->P->Ny + (this->N) * this->P->Nu;
        memcpy(&Grad[NXF], this->EvalHelp1Ny2.get_data(), this->P->Ny * sizeof(real));
        // Kopiere Vektor P:
        int_type NP = (this->N + 1) * (this->P->Ny + this->P->Nu);
        memcpy(&Grad[NP], this->EvalHelp1Np1.get_data(), this->P->Np * sizeof(real));
    }

    template <typename real>
    void Trapez<real>::CON_Diskr(const real * states, real * GG)
    {
        real         t    = this->t0;
        const real * y0   = 0;
        const real * yF   = 0;
        const real * y_k  = 0;
        const real * y_kk = 0;
        const real * u_k  = 0;
        const real * u_kk = 0;
        const real * p;
        if (this->P->Np > 0)
            p = &(states[(this->N + 1) * (this->P->Ny + this->P->Nu)]);
        else
            p = &(states[0]);

        int_type NBOUND = this->MyDims->getBoundDimBOUND().N_Dim_ROW;
        int_type NC     = this->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->MyDims->getBoundDimSTATE().N_Dim_ROW + this->MyDims->getBoundDimCONTROL().N_Dim_ROW;

        for (int_type Index = 0; Index < this->N; ++Index)
        {
            // Rechte Seite:
            // Determine Zeitpunkt:
            t = this->t0 + this->h * Index;
            // Determine Row Index:
            int_type i = this->P->Ny * Index;
            // Determine y_k, y_{k+1} und u_k:
            // y_k:
            y_k = &states[Index * (this->P->Ny + this->P->Nu)];
            ;
            // y_{k+1}:
            y_kk = &states[(Index + 1) * (this->P->Ny + this->P->Nu)];
            ;
            // u_k:
            u_k = &states[(Index + 1) * this->P->Ny + Index * this->P->Nu];
            // u_{k+1}:
            u_kk = &states[(Index + 2) * this->P->Ny + (Index + 1) * this->P->Nu];
            ;

            // Auswerten der ODEs:
            this->P->ODE(t, y_k, u_k, p, this->EvalHelp_f_k);
            this->P->ODE(t, y_kk, u_kk, p, this->EvalHelp_f_kk);

            // Trapezregel:
            for (int_type l = 0; l < this->P->Ny; ++l)
                GG[i + l] = y_kk[l] - y_k[l] - (this->h / 2.0) * (this->EvalHelp_f_k[l] + this->EvalHelp_f_kk[l]);

            // Nichtlineare Nebenbedingungen:
            // Determine Row Index:
            i = this->P->Ny * (this->N) + NBOUND + NC * Index;
            this->P->NLCSTR(t, y_k, u_k, p, this->EvalHelp_g);
            this->MyDims->set_Vector_C_from_G(this->EvalHelp_g, y_k, u_k, this->P, &GG[i]);
        }
        // N:
        // Determine Row Index:
        int_type i = this->P->Ny * (this->N) + NBOUND + NC * (this->N);
        this->P->NLCSTR(t, y_kk, u_kk, p, this->EvalHelp_g); // Änderung y_k+1 und u_k+1
        this->MyDims->set_Vector_C_from_G(this->EvalHelp_g, y_kk, u_kk, this->P, &GG[i]);

        // Randbedingungen:
        y0 = getYk(0);
        yF = getYk(this->N);
        this->P->BOUND(this->t0, this->tF, yF, y0, p, this->EvalHelp_psi);
        i = this->P->Ny * (this->N);
        this->MyDims->set_Vector_IPbound_from_bound(this->EvalHelp_psi, this->P, &GG[i]);

        // Parameter:
        i                 = this->P->Ny * (this->N) + NBOUND + NC * (this->N + 1);
        int_type counterp = 0;
        for (int_type kp = 0; kp < this->P->Np; kp++)
        {
            switch (this->MyDims->getBoundTypeArrayPARAM(kp))
            {
            case DiscreteDefs::INEQUALITY:
                GG[i + counterp] = p[kp] - this->P->p_max[kp];
                counterp++;
                GG[i + counterp] = -p[kp] + this->P->p_min[kp];
                counterp++;
                break;
            case DiscreteDefs::EQUATION:
                GG[i + counterp] = p[kp] - this->P->p_max[kp];
                counterp++;
                break;
            case DiscreteDefs::LOWER_BARRIER:
                GG[i + counterp] = -p[kp] + this->P->p_min[kp];
                counterp++;
                break;
            case DiscreteDefs::UPPER_BARRIER:
                GG[i + counterp] = p[kp] - this->P->p_max[kp];
                counterp++;
                break;
            case DiscreteDefs::UNCONSTRAINED:
                // Nothing to do!
                break;
            }
        }
    }

    template <typename real>
    void Trapez<real>::update_Jacobian(const real * x)
    {
        // Kopiere Zustandsvektor:
        memcpy(this->x, x, this->N_Statevector * sizeof(real));
        this->MyTimer.tic();
        this->ClearJacobi();
        this->JacobiSetODEConstraints();
        this->JacobiSetBoundConstraints();
        this->JacobiSetNLConstraints();
        this->MyTimer.toc();
        this->TimeJacobi += this->MyTimer.elapsed_s();

        this->MyTimer.tic();
        // Update der Teilmatrizen PSI_Lower und PSI_Upper:
        this->MyDims->set_PSI(this->b_Lower, this->PSI_Lower);
        this->MyDims->set_PSI(this->b_Upper, this->PSI_Upper);
        this->MyDims->set_PSI(this->R, this->R_Hat);

        // Update der Teilmatrizen C und Q:
        for (int_type i = 0; i < this->N + 1; ++i)
        {
            this->MyDims->set_C_Constr(this->Array_G[i], this->Array_C_CONSTR[i]);
            this->MyDims->set_Q_Constr(this->Array_q[i], this->Array_Q[i]);
        }
        this->MyTimer.toc();
        this->TimeIP += this->MyTimer.elapsed_s();
    }

    template <typename real>
    void Trapez<real>::update_Hessian(real const x[], real const lambda_IP[])
    {
        // Kopiere Zustandsvektor:
        memcpy(this->x, x, this->N_Statevector * sizeof(real));
        // Kopiere Vektor der Lagrangemultiplikatoren:
        this->MyDims->set_lambda_hesse(lambda_IP, this->lambda);
        // Setup Hesse:
        this->MyTimer.tic();
        this->ClearHesse();
        this->HesseSetInitialFinalState();
        this->HesseSetGridPointSubmatrix();
        this->HesseSetSumParameterMatrix();
        this->MyTimer.toc();
        this->TimeHesse += this->MyTimer.elapsed_s();
    }

    template <typename real>
    int_type Trapez<real>::getDimStateIP()
    {
        return (this->N + 1) * (this->P->Ny + this->P->Nu) + this->P->Np;
    }

    template <typename real>
    int_type Trapez<real>::getDimLambdaIP()
    {
        return this->MyDims->getDimLambdaIP();
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    private Memberfunktionen                                                         //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Trapez<real>::get_iDimHesse(int_type iDimHesse[])
    {
        // States and constrolls:
        for (int_type i = 0; i < (this->N + 1); ++i)
            iDimHesse[i] = this->P->Ny + this->P->Nu;
        // Parameter:
        iDimHesse[this->N + 1] = this->P->Np;
    }

    template <typename real>
    void Trapez<real>::get_iDimJacobi(int_type iDimJacobi[])
    {
        int_type rowDimC = this->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->MyDims->getBoundDimSTATE().N_Dim_ROW + this->MyDims->getBoundDimCONTROL().N_Dim_ROW;
        // ODE Constraints:
        for (int_type i = 0; i < (this->N); ++i)
            iDimJacobi[i] = this->P->Ny;
        // Boundary Constraints:
        iDimJacobi[this->N] = this->MyDims->getBoundDimBOUND().N_Dim_ROW;
        // Nonlinear Constraints:
        for (int_type i = 0; i < this->N + 1; ++i)
            iDimJacobi[this->N + 1 + i] = rowDimC;
        // Parameter Constraints:
        iDimJacobi[2 * this->N + 2] = this->MyDims->getBoundDimPARAM().N_Dim_ROW;
    }

    template <typename real>
    void Trapez<real>::get_iDimIP(int_type iDimIP[])
    {
        // Bandmatrix:
        int_type Index;
        int_type rowDimC = this->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->MyDims->getBoundDimSTATE().N_Dim_ROW + this->MyDims->getBoundDimCONTROL().N_Dim_ROW;
        for (int_type i = 0; i < this->N; ++i)
        {
            Index             = i * 3;
            iDimIP[Index]     = this->P->Ny + this->P->Nu;
            iDimIP[Index + 1] = rowDimC;
            iDimIP[Index + 2] = this->P->Ny;
        }
        // Letzter Block:
        Index             = this->N * 3;
        iDimIP[Index]     = this->P->Ny + this->P->Nu;
        iDimIP[Index + 1] = rowDimC;
        iDimIP[Index + 2] = this->MyDims->getBoundDimPARAM().N_Dim_ROW;
        // Restmatrix:
        Index             = 3 * (this->N + 1);
        iDimIP[Index]     = this->MyDims->getBoundDimBOUND().N_Dim_ROW;
        iDimIP[Index + 1] = this->MyDims->getBoundDimPARAM().N;
    }

    template <typename real>
    void Trapez<real>::ClearJacobi()
    {
        // Jacobimatrix:
        // Matrizen der ODE Constraints:
        for (int_type i = 0; i < this->N; ++i)
        {
            this->Array_M[i].zero_fill();
            this->Array_S[i].zero_fill();
            this->Array_R[i].zero_fill();
        }

        // Matrizen der Boundary Constraints:
        this->b_Lower.zero_fill();
        this->b_Upper.zero_fill();
        this->R.zero_fill();

        // Matrizen der nonlinear Constraints:
        for (int_type i = 0; i < this->N + 1; ++i)
        {
            this->Array_G[i].zero_fill();
            this->Array_q[i].zero_fill();
        }
    }

    template <typename real>
    void Trapez<real>::ClearHesse()
    {
        // Hessematrix:
        this->MP.zero_fill();
        this->PHI.zero_fill();
        for (int_type i = 0; i <= this->N; ++i)
        {
            this->Array_W[i].zero_fill();
            this->Array_P[i].zero_fill();
        }
    }

    template <typename real>
    void Trapez<real>::JacobiSetODEConstraints()
    {

        this->EvalHelpNyNy1.zero_fill();
        this->EvalHelpNyNu1.zero_fill();
        this->EvalHelpNyNp1.zero_fill();

        real   t  = this->t0;       // t
        real * p  = this->getP();   // p
        real * yK = this->getYk(0); // y_{k}
        real * yKp;                 // y_{k+1}
        real * uK = this->getUk(0); // u_{k}
        real * uKp;                 // u_{k+1}

        // Auswertung fuer t_0:
        this->P->D_ODE_X(t, yK, uK, p, this->EvalHelpNyNy1);
        this->P->D_ODE_U(t, yK, uK, p, this->EvalHelpNyNu1);
        this->P->D_ODE_P(t, yK, uK, p, this->EvalHelpNyNp1);

        for (int_type index = 0; index < this->N; ++index)
        {
            this->EvalHelpNyNy2.zero_fill();
            this->EvalHelpNyNu2.zero_fill();
            t   = this->t0 + this->h * index;
            yK  = this->getYk(index);
            yKp = this->getYk(index + 1);
            uK  = this->getUk(index);
            uKp = this->getUk(index + 1);
            // Auswertung fuer t_{k+1}:
            this->P->D_ODE_X(t + this->h, yKp, uKp, p, this->EvalHelpNyNy2);
            this->P->D_ODE_U(t + this->h, yKp, uKp, p, this->EvalHelpNyNu2);

            // Array M:
            // *this->EvalHelpNyNy1 = (-h/2) * D_ODE_X - I:
            this->EvalHelpNyNy1.scale_by(-this->h / 2);
            for (int_type i = 0; i < this->P->Ny; ++i)
            {
                real hlp                  = this->EvalHelpNyNy1(i, i);
                this->EvalHelpNyNy1(i, i) = hlp - 1.0;
            }
            // this->Array_M[index]->PushMatrix(0, 0, this->EvalHelpNyNy1);
            this->Array_M[index].load_block(this->EvalHelpNyNy1, 0, 0);

            this->EvalHelpNyNy1.load(this->EvalHelpNyNy2);

            // *this->EvalHelpNyNy1 = (-h/2) * D_ODE_U:
            this->EvalHelpNyNu1.scale_by(-this->h / 2);
            // this->Array_M[index]->PushMatrix(0, this->P->Ny, this->EvalHelpNyNu1);
            this->Array_M[index].load_block(this->EvalHelpNyNu1, 0, this->P->Ny);
            // Auswertung fuer t_k in naechsten Schritt:
            this->EvalHelpNyNu1.load(this->EvalHelpNyNu2);

            // Array S:
            // *this->EvalHelpNyNy2 = (-h/2) * D_ODE_X + I:
            this->EvalHelpNyNy2.scale_by(-this->h / 2);
            for (int_type i = 0; i < this->P->Ny; ++i)
            {
                real hlp                  = this->EvalHelpNyNy2(i, i);
                this->EvalHelpNyNy2(i, i) = hlp + 1.0;
            }
            // this->Array_S[index]->PushMatrix(0, 0, this->EvalHelpNyNy2);
            this->Array_S[index].load_block(this->EvalHelpNyNy2, 0, 0);

            // *this->EvalHelpNyNy1 = (-h/2) * D_ODE_U:
            this->EvalHelpNyNu2.scale_by(-this->h / 2);
            // this->Array_S[index]->PushMatrix(0, this->P->Ny, this->EvalHelpNyNu2);
            this->Array_S[index].load_block(this->EvalHelpNyNu2, 0, this->P->Ny);

            // Parameter:
            if (this->P->Np > 0)
            {
                this->EvalHelpNyNp2.zero_fill();
                // Auswertung fuer t_{k+1}:
                this->P->D_ODE_P(t + this->h, yKp, uKp, p, this->EvalHelpNyNp2);
                // *this->EvalHelpNyNy1 = (-h/2) * (*this->EvalHelpNyNy1 + *this->EvalHelpNyNy2):
                this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
                this->EvalHelpNyNp1.scale_by(-this->h / 2);
                this->Array_R[index].load(this->EvalHelpNyNp1);

                // Auswertung fuer t_k in naechsten Schritt:
                this->EvalHelpNyNp1.load(this->EvalHelpNyNp2);
            }
        }
    }

    template <typename real>
    void Trapez<real>::JacobiSetBoundConstraints()
    {
        if (this->P->N_BOUND > 0)
        {
            real * p  = this->getP(); // p
            real * yF = &(this->x[(this->N) * (this->P->Ny + this->P->Nu)]);
            real * y0 = &(this->x[0]);
            // *this->b_Lower:
            this->P->D_BOUND_X0(this->t0, this->tF, yF, y0, p, this->b_Lower);
            // *this->b_Upper:
            this->P->D_BOUND_XF(this->t0, this->tF, yF, y0, p, this->b_Upper);

            // Parameter:
            if (this->P->Np > 0)
                this->P->D_BOUND_P(this->t0, this->tF, yF, y0, p, this->R);
        }
    }

    template <typename real>
    void Trapez<real>::JacobiSetNLConstraints()
    {
        if (this->P->N_NLCSTR > 0)
        {
            real   t = this->t0;     // t
            real * p = this->getP(); // p
            real * yK;               // y_{k}
            real * uK;               // u_{k}
            for (int_type index = 0; index < this->N + 1; ++index)
            {
                this->EvalHelp_N_NLCTR_Ny.zero_fill();
                this->EvalHelp_N_NLCTR_Nu.zero_fill();
                yK = this->getYk(index);
                uK = this->getUk(index);

                this->P->D_NLCSTR_X(t, yK, uK, p, this->EvalHelp_N_NLCTR_Ny);
                // this->Array_G[index].PushMatrix(0, 0, (this->EvalHelp_N_NLCTR_Ny));
                this->Array_G[index].load_block(this->EvalHelp_N_NLCTR_Ny, 0, 0);

                this->P->D_NLCSTR_U(t, yK, uK, p, this->EvalHelp_N_NLCTR_Nu);
                // this->Array_G[index].PushMatrix(0, this->P->Ny, (this->EvalHelp_N_NLCTR_Nu));
                this->Array_G[index].load_block(this->EvalHelp_N_NLCTR_Nu, 0, this->P->Ny);

                // Parameter:
                if (this->P->Np > 0)
                    this->P->D_NLCSTR_P(t, yK, uK, p, this->Array_q[index]);
            }
        }
    }

    template <typename real>
    void Trapez<real>::HesseSetInitialFinalState()
    {
        real * p  = this->getP();
        real * y0 = getYk(0);
        real * u0 = getUk(0);
        real * yF = getYk(this->N);
        real * uF = getUk(this->N);
        // Determine Lagrangemultiplikatoren:
        real * sigma    = this->getSIGMA();
        real * lambda0  = this->getLAMBDAk(0);
        real * mu0      = this->getMUk(0);
        real * lambdaF_ = this->getLAMBDAk(this->N - 1);
        real * muF      = this->getMUk(this->N);

        // Zielfunktion:
        this->EvalHelpNyNy1.zero_fill();
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_VARPHI_XFX0(this->t0, this->tF, yF, y0, p, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Randbedingungen:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_BOUND_XFX0(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Transponierte:
        this->EvalHelpNyNy1.get_transposed(this->EvalHelpNyNy2);
        this->PHI.load_block(this->EvalHelpNyNy2, 0, 0);

        // W00:
        // y0y0:
        this->EvalHelpNyNy1.zero_fill();
        // Zielfunktion:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_VARPHI_X0X0(this->t0, this->tF, yF, y0, p, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Rechte Seite:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_ODE_XX(this->t0, y0, u0, p, lambda0, this->EvalHelpNyNy2);
        this->EvalHelpNyNy2.scale_by(-this->h / 2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_NLCSTR_XX(this->t0, y0, u0, p, mu0, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Randbedingungen:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_BOUND_X0X0(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // this->Array_W[0]->PushMatrix(0, 0, (this->EvalHelpNyNy1));
        this->Array_W[0].load_block(this->EvalHelpNyNy1, 0, 0);
        // u0u0:
        // Rechte Seite:
        this->EvalHelpNuNu1.zero_fill();
        this->EvalHelpNuNu2.zero_fill();
        this->P->L_DD_ODE_UU(this->t0, y0, u0, p, lambda0, this->EvalHelpNuNu2);
        this->EvalHelpNuNu2.scale_by(-this->h / 2);
        this->EvalHelpNuNu1.add(this->EvalHelpNuNu2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNuNu2.zero_fill();
        this->P->L_DD_NLCSTR_UU(this->t0, y0, u0, p, mu0, this->EvalHelpNuNu2);
        this->EvalHelpNuNu1.add(this->EvalHelpNuNu2);
        this->Array_W[0].load_block(this->EvalHelpNuNu1, this->P->Ny, this->P->Ny);

        // x0u0:
        // Rechte Seite:
        this->EvalHelpNyNu1.zero_fill();
        this->EvalHelpNyNu2.zero_fill();
        this->P->L_DD_ODE_XU(this->t0, y0, u0, p, lambda0, this->EvalHelpNyNu2);
        this->EvalHelpNyNu2.scale_by(-this->h / 2);
        this->EvalHelpNyNu1.add(this->EvalHelpNyNu2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNyNu2.zero_fill();
        this->P->L_DD_NLCSTR_XU(this->t0, y0, u0, p, mu0, this->EvalHelpNyNu2);
        this->EvalHelpNyNu1.add(this->EvalHelpNyNu2);
        this->Array_W[0].load_block(this->EvalHelpNyNu1, 0, this->P->Ny);
        // Transponierte:
        this->EvalHelpNyNu1.get_transposed(this->EvalHelpNuNy1);
        this->Array_W[0].load_block(this->EvalHelpNuNy1, this->P->Ny, 0);

        // xFxF:
        this->EvalHelpNyNy1.zero_fill();
        // Zielfunktion:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_VARPHI_XFXF(this->t0, this->tF, yF, y0, p, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Rechte Seite:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_ODE_XX(this->tF, yF, uF, p, lambdaF_, this->EvalHelpNyNy2);
        this->EvalHelpNyNy2.scale_by(-this->h / 2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_NLCSTR_XX(this->tF, yF, uF, p, muF, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
        // Randbedingungen:
        this->EvalHelpNyNy2.zero_fill();
        this->P->L_DD_BOUND_XFXF(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNyNy2);
        this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);

        // this->Array_W[this->N]->PushMatrix(0, 0, (this->EvalHelpNyNy1));
        this->Array_W[this->N].load_block(this->EvalHelpNyNy1, 0, 0);

        // xFuF:
        this->EvalHelpNyNu1.zero_fill();
        // Rechte Seite:
        this->EvalHelpNyNu2.zero_fill();
        this->P->L_DD_ODE_XU(this->tF, yF, uF, p, lambdaF_, this->EvalHelpNyNu2);
        this->EvalHelpNyNu2.scale_by(-this->h / 2);
        this->EvalHelpNyNu1.add(this->EvalHelpNyNu2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNyNu2.zero_fill();
        this->P->L_DD_NLCSTR_XU(this->tF, yF, uF, p, muF, this->EvalHelpNyNu2);
        this->EvalHelpNyNu1.add(this->EvalHelpNyNu2);

        this->Array_W[this->N].load_block(this->EvalHelpNyNu1, 0, this->P->Ny);
        // Transponierte:
        this->EvalHelpNyNu1.get_transposed(this->EvalHelpNuNy1);
        this->Array_W[this->N].load_block(this->EvalHelpNuNy1, this->P->Ny, 0);

        // uFuF:
        this->EvalHelpNuNu1.zero_fill();
        // Rechte Seite:
        this->EvalHelpNuNu2.zero_fill();
        this->P->L_DD_ODE_UU(this->tF, yF, uF, p, lambdaF_, this->EvalHelpNuNu2);
        this->EvalHelpNuNu2.scale_by(-this->h / 2);
        this->EvalHelpNuNu1.add(this->EvalHelpNuNu2);
        // Nichtlineare Nebenbedingungen:
        this->EvalHelpNuNu2.zero_fill();
        this->P->L_DD_NLCSTR_UU(this->tF, yF, uF, p, muF, this->EvalHelpNuNu2);
        this->EvalHelpNuNu1.add(this->EvalHelpNuNu2);

        this->Array_W[this->N].load_block(this->EvalHelpNuNu1, this->P->Ny, this->P->Ny);

        if (this->P->Np > 0)
        {
            // y0p:
            this->EvalHelpNyNp1.zero_fill();
            // Zielfunktion:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_VARPHI_X0P(this->t0, this->tF, yF, y0, p, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Rechte Seite:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_ODE_XP(this->t0, y0, u0, p, lambda0, this->EvalHelpNyNp2);
            this->EvalHelpNyNp2.scale_by(-this->h / 2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Nichtlineare Nebenbedingungen:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_NLCSTR_XP(this->t0, y0, u0, p, mu0, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Randbedingungen:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_BOUND_X0P(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);

            this->Array_P[0].load_block(this->EvalHelpNyNp1, 0, 0);

            // u0p:
            this->EvalHelpNuNp1.zero_fill();
            // Rechte Seite:
            this->EvalHelpNuNp2.zero_fill();
            this->P->L_DD_ODE_UP(this->t0, y0, u0, p, lambda0, this->EvalHelpNuNp2);
            this->EvalHelpNuNp2.scale_by(-this->h / 2);
            this->EvalHelpNuNp1.add(this->EvalHelpNuNp2);
            // Nichtlineare Nebenbedingungen:
            this->EvalHelpNuNp2.zero_fill();
            this->P->L_DD_NLCSTR_UP(this->t0, y0, u0, p, mu0, this->EvalHelpNuNp2);
            this->EvalHelpNuNp1.add(this->EvalHelpNuNp2);
            // Transponierte:
            this->Array_P[0].load_block(this->EvalHelpNuNp1, this->P->Ny, 0);

            // yFp:
            this->EvalHelpNyNp1.zero_fill();
            // Zielfunktion:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_VARPHI_XFP(this->t0, this->tF, yF, y0, p, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Rechte Seite:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_ODE_XP(this->tF, yF, uF, p, lambdaF_, this->EvalHelpNyNp2);
            this->EvalHelpNyNp2.scale_by(-this->h / 2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Nichtlineare Nebenbedingungen:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_NLCSTR_XP(this->tF, yF, uF, p, muF, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Randbedingungen:
            this->EvalHelpNyNp2.zero_fill();
            this->P->L_DD_BOUND_XFP(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNyNp2);
            this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);
            // Transponierte:
            this->Array_P[this->N].load_block(this->EvalHelpNyNp1, 0, 0);

            // uFp:
            this->EvalHelpNuNp1.zero_fill();
            // Rechte Seite:
            this->EvalHelpNuNp2.zero_fill();
            this->P->L_DD_ODE_UP(this->tF, yF, uF, p, lambdaF_, this->EvalHelpNuNp2);
            this->EvalHelpNuNp2.scale_by(-this->h / 2);
            this->EvalHelpNuNp1.add(this->EvalHelpNuNp2);
            // Nichtlineare Nebenbedingungen:
            this->EvalHelpNuNp2.zero_fill();
            this->P->L_DD_NLCSTR_UP(this->tF, yF, uF, p, muF, this->EvalHelpNuNp2);
            this->EvalHelpNuNp1.add(this->EvalHelpNuNp2);
            // Transponierte:
            this->Array_P[this->N].load_block(this->EvalHelpNuNp1, this->P->Ny, 0);
        }
    }

    template <typename real>
    void Trapez<real>::HesseSetGridPointSubmatrix()
    {
        real * p = this->getP();
        real * lambda_k;
        real * mu_k;
        real * lambda_k_;
        // real* mu_k_;
        real * y_k;
        real * u_k;
        real * y_k_;
        real * u_k_;
        real   t = this->t0;
        for (int_type Index = 1; Index < this->N; ++Index)
        {
            t += this->h;
            lambda_k  = this->getLAMBDAk(Index);
            mu_k      = this->getMUk(Index);
            lambda_k_ = this->getLAMBDAk(Index - 1);
            // mu_k_     = this->getMUk(Index - 1);
            // Determine y_k und u_k:
            // y_k:
            y_k = this->getYk(Index);
            // u_k:
            u_k = this->getUk(Index);
            // Determine y_{k-1} und u_{k-1}:
            // y_{k-1}:
            y_k_ = this->getYk(Index - 1);
            // u_{k-1}:
            u_k_ = this->getUk(Index - 1);
            // Rechte Seite:
            // SMatrix l_dd_ODEYY;
            this->EvalHelpNyNy1.zero_fill();
            this->P->L_DD_ODE_XX(t, y_k, u_k, p, lambda_k, this->EvalHelpNyNy1);
            this->EvalHelpNyNy1.scale_by(-this->h / 2);
            // SMatrix l_dd_ODEYY_;
            this->EvalHelpNyNy2.zero_fill();
            this->P->L_DD_ODE_XX(t - this->h, y_k_, u_k_, p, lambda_k, this->EvalHelpNyNy2);
            this->EvalHelpNyNy2.scale_by(-this->h / 2);
            // Nichtlineare Nebenbedingungen:
            // SMatrix m_dd_NLCSTRYY;
            this->EvalHelpNyNy3.zero_fill();
            this->P->L_DD_NLCSTR_XX(t, y_k, u_k, p, mu_k, this->EvalHelpNyNy3);
            // this->Array_L[Index].clear();
            this->EvalHelpNyNy1.add(this->EvalHelpNyNy2);
            this->EvalHelpNyNy1.add(this->EvalHelpNyNy3);
            this->Array_W[Index].load_block(this->EvalHelpNyNy1, 0, 0);
            // Rechte Seite:
            // SMatrix l_dd_ODEuu;
            this->EvalHelpNuNu1.zero_fill();
            this->P->L_DD_ODE_UU(t, y_k, u_k, p, lambda_k, this->EvalHelpNuNu1);
            this->EvalHelpNuNu1.scale_by(-this->h / 2);
            // SMatrix l_dd_ODEuu_;
            this->EvalHelpNuNu2.zero_fill();
            this->P->L_DD_ODE_UU(t - this->h, y_k_, u_k_, p, lambda_k_, this->EvalHelpNuNu2);
            this->EvalHelpNuNu2.scale_by(-this->h / 2);
            // Nichtlineare Nebenbedingungen:
            // SMatrix m_dd_NLCSTRuu;
            this->EvalHelpNuNu3.zero_fill();
            this->P->L_DD_NLCSTR_UU(t, y_k, u_k, p, mu_k, this->EvalHelpNuNu3);
            // this->Array_U[Index].clear();
            this->EvalHelpNuNu1.add(this->EvalHelpNuNu2);
            this->EvalHelpNuNu1.add(this->EvalHelpNuNu3);
            this->Array_W[Index].load_block(this->EvalHelpNuNu1, this->P->Ny, this->P->Ny);

            // yiui:
            // Rechte Seite:
            // SMatrix l_dd_ODEYU;
            this->EvalHelpNyNu1.zero_fill();
            this->P->L_DD_ODE_XU(t, y_k, u_k, p, lambda_k, this->EvalHelpNyNu1);
            this->EvalHelpNyNu1.scale_by(-this->h / 2);
            // SMatrix l_dd_ODEYU_;
            this->EvalHelpNyNu2.zero_fill();
            this->P->L_DD_ODE_XU(t - this->h, y_k_, u_k_, p, lambda_k_, this->EvalHelpNyNu2);
            this->EvalHelpNyNu2.scale_by(-this->h / 2);
            // Nichtlineare Nebenbedingungen:
            // SMatrix m_dd_NLCSTRYU;
            this->EvalHelpNyNu3.zero_fill();
            this->P->L_DD_NLCSTR_XU(t, y_k, u_k, p, mu_k, this->EvalHelpNyNu3);
            // this->Array_l[Index].clear();
            this->EvalHelpNyNu1.add(this->EvalHelpNyNu2);
            this->EvalHelpNyNu1.add(this->EvalHelpNyNu3);
            this->Array_W[Index].load_block(this->EvalHelpNyNu1, 0, this->P->Ny);
            // Transponierte:
            this->EvalHelpNyNu1.get_transposed(this->EvalHelpNuNy1);
            this->Array_W[Index].load_block(this->EvalHelpNuNy1, this->P->Ny, 0);
            if (this->P->Np > 0)
            {
                real lam_sum[this->P->Ny];
                for (int_type s = 0; s < this->P->Ny; ++s)
                    lam_sum[s] = lambda_k[s] + lambda_k_[s];
                // Rechte Seite:
                // SMatrix l_dd_ODEYP ;
                this->EvalHelpNyNp1.zero_fill();
                this->P->L_DD_ODE_XP(t, y_k, u_k, p, lam_sum, this->EvalHelpNyNp1);
                this->EvalHelpNyNp1.scale_by(-this->h / 2);
                // Nichtlineare Nebenbedingungen:
                // SMatrix m_dd_NLCSTRYP;
                this->EvalHelpNyNp2.zero_fill();
                this->P->L_DD_NLCSTR_XP(t, y_k, u_k, p, mu_k, this->EvalHelpNyNp2);
                this->EvalHelpNyNp1.add(this->EvalHelpNyNp2);

                this->Array_P[Index].load_block(this->EvalHelpNyNp1, 0, 0);

                this->EvalHelpNuNp1.zero_fill();
                this->P->L_DD_ODE_UP(t, y_k, u_k, p, lam_sum, this->EvalHelpNuNp1);
                this->EvalHelpNuNp1.scale_by(-this->h / 2);
                // Nichtlineare Nebenbedingungen:
                // SMatrix m_dd_NLCSTRUP;
                this->EvalHelpNuNp2.zero_fill();
                this->P->L_DD_NLCSTR_UP(t, y_k, u_k, p, mu_k, this->EvalHelpNuNp2);
                this->EvalHelpNuNp1.add(this->EvalHelpNuNp2);
                this->Array_P[Index].load_block(this->EvalHelpNuNp1, this->P->Ny, 0);
            }
        }
    }

    template <typename real>
    void Trapez<real>::HesseSetSumParameterMatrix()
    {
        // Zustände:
        real * y_k;
        real * u_k;
        real   t = this->t0;
        // Determine Parameter:
        real * p  = this->getP();
        real * y0 = getYk(0);
        real * yF = getYk(this->N);
        // Determine Lagrangemultiplikatoren:
        real * lambda_k = 0;
        real * mu_k     = 0;
        real * sigma    = this->getSIGMA();

        for (int_type Index = 0; Index < this->N + 1; ++Index)
        {
            // Determine Lagrangemultiplikatoren:
            lambda_k = this->getLAMBDAk(Index);
            mu_k     = this->getMUk(Index);
            // Determine y_k und u_k:
            // y_k:
            y_k = getYk(Index);
            // u_k:
            u_k = getUk(Index);

            if (Index < this->N)
            {
                // Rechte Seite:
                // SMatrix l_dd_ODEpp; l_dd_ODEpp = this->SMatrHLP1
                this->EvalHelpNpNp1.zero_fill();
                this->P->L_DD_ODE_PP(t, y_k, u_k, p, lambda_k, this->EvalHelpNpNp1);
                this->EvalHelpNpNp1.scale_by(-this->h / 2);
                this->MP.add(this->EvalHelpNpNp1);
                // Nichtlineare Nebenbedingungen:
                // SMatrix m_dd_NLCSTRpp; m_dd_NLCSTRpp = this->SMatrHLP2
                this->EvalHelpNpNp1.zero_fill();
                this->P->L_DD_NLCSTR_PP(t, y_k, u_k, p, mu_k, this->EvalHelpNpNp1);
                this->MP.add(this->EvalHelpNpNp1);
            }
            else
            {
                // Nichtlineare Nebenbedingungen:
                // SMatrix m_dd_NLCSTRpp; m_dd_NLCSTRpp = this->SMatrHLP1
                this->EvalHelpNpNp1.zero_fill();
                this->P->L_DD_NLCSTR_PP(t, y_k, u_k, p, mu_k, this->EvalHelpNpNp1);
                this->MP.add(this->EvalHelpNpNp1);
            }

            t += this->h;
        }

        // Zielfunktion:
        // SMatrix dd_PSIpp; dd_PSIpp = this->SMatrHLP1
        this->EvalHelpNpNp1.zero_fill();
        this->P->L_DD_VARPHI_PP(this->t0, this->tF, yF, y0, p, this->EvalHelpNpNp1);
        this->MP.add(this->EvalHelpNpNp1);
        // Randbedingungen:
        // SMatrix s_dd_BOUNDpp; s_dd_BOUNDpp = this->SMatrHLP2
        this->EvalHelpNpNp1.zero_fill();
        this->P->L_DD_BOUND_PP(this->t0, this->tF, yF, y0, p, sigma, this->EvalHelpNpNp1);
        this->MP.add(this->EvalHelpNpNp1);
    }

    template class Trapez<double>;
    template class Trapez<float>;

} // namespace OCPbasic
