/* (C) 2018, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file InteriorPointMethod.cpp
 * Implementation of the class InteriorPointMethod.
 *
 * @author A. Huber
 * @since 07.11.2018
 */

#include "InteriorPointMethod.h"
#include "../Discretization/trapez.h"
#include "InteriorPointMethod/ip_nlp.h"
#include "InteriorPointMethod/ip_solver.h"
#include <iostream>

#include "../linalg/LapackKKTSolver.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#include <IPbasic/LinSolvers/IPbasic_LinearSolver_MA57.h>

namespace OCPbasic
{

    using IPbasic::Errors;

    template <typename real>
    InteriorPointMethod<real>::InteriorPointMethod(Problem<real> * P, int_type NGit, real t0, real tF, OCPbasic::LinearSolverFlag MyLinearSolverFlag, IPbasic::LineSearchMethod UsedLineSearchMethod)
        : Optimizer<real>(P, t0, tF)
    {
        this->LocalInitTimer.tic();
        this->discretization = new Trapez<real>(NGit, t0, tF, P);
        int_type DIM_W       = this->P->Ny + this->P->Nu;
        int_type DIM_C =
            this->discretization->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->discretization->MyDims->getBoundDimSTATE().N_Dim_ROW + this->discretization->MyDims->getBoundDimCONTROL().N_Dim_ROW;
        int_type DIM_M = this->P->Ny;
        // int_type NDimRowBand = (NGit)*(DIM_W + DIM_C + DIM_M);
        // int_type sizeBand    = (DIM_W + DIM_C + DIM_M);
        int_type              NBlocks = 2 * NGit;
        std::vector<int_type> DimBlocks(NBlocks + 1);
        int_type              ipos = 0;
        DimBlocks[0]               = ipos;
        for (int_type i = 0; i < NGit; ++i)
        {
            ipos += DIM_W + DIM_C;
            DimBlocks[2 * i + 1] = ipos;
            ipos += DIM_M;
            DimBlocks[2 * i + 2] = ipos;
            // ipos += DIM_W + DIM_C + DIM_M; DimBlocks[i+1] = ipos;
        }
        // std::cout << ipos << "  " << NDimRowBand << std::endl;
        TicToc T;
        T.tic();
        /*
        switch (this->Param->getFlagLinGLSol())
        {
        case LinearSolverFlags::DEFAULT_LR:
            this->MyLinearSolver  =  new DefaultLR<real>(N_Dim);
            break;
        case LinearSolverFlags::HSL:
            this->MyLinearSolver  =  new HSLLR<real>(this->Diskr->IPMatrix);
            break;
        case LinearSolverFlags::KKT_SOLVER_LAPACK:
            NDimRowBand = (NGit)*(DIM_W + DIM_C + DIM_M);
            this->MyLinearSolver  =  new LapackKKTSolver<real>(NGit,
                                                               NDimRowBand, sizeBand);
            break;
        case LinearSolverFlags::KKT_SOLVER_BLOCKTRIDIAG:
          this->MyLinearSolver  =  new BandtridiagKKTSolver<real>(NGit,
                                                                    NDimRowBand, NBlocks,
                                                                    DimBlocks);
            break;
        default:
            Errors::set_Error("Unzulässiger Flag für linearen Gleichungssystemlöser gesetzt!");
            this->MyLinearSolver  =  new DefaultLR<real>(N_Dim);
            break;
        }
        */
        // Set up NLP:
        this->MyNLP = new IP_NLP<real>(this->discretization);

        int_type DimRHS    = this->discretization->getDimStateIP() + this->discretization->getDimLambdaIP();
        this->SortIndArray = new int_type[DimRHS];
        this->setIndexMap();
        this->TimeLinearSolver = 0.0;

        // int_type DIM_E = this->discretization->MyDims->getBoundDimPARAM().N_Dim_ROW;
        int_type NDimRowBand = (NGit) * (DIM_W + DIM_C + DIM_M); // + DIM_W + DIM_C + DIM_E;
        int_type SizeBand    = (DIM_W + DIM_C + DIM_M);
        if (MyLinearSolverFlag == OCPbasic::LINSOL_HSL_MA57)
        {
            this->MyLinSolver = new IPbasic::LinearSolver_MA57<real>(false, false);
            this->MyNLP->SetHasToPushZeros(true);
        }
        else if (MyLinearSolverFlag == OCPbasic::LINSOL_LAPACK_BANDED_KKT)
        {
            this->MyLinSolver = new LapackKKTSolver<real>(DimRHS, NDimRowBand, SizeBand, this->SortIndArray);
            this->MyNLP->SetHasToPushZeros(false);
        }
        else
        {
            this->MyLinSolver = new LapackKKTSolver<real>(DimRHS, NDimRowBand, SizeBand, this->SortIndArray);
            this->MyNLP->SetHasToPushZeros(false);
            Errors::set_Error("No linear solver specified!");
        }

        this->MyNLP_Solver = new IP_Solver<real>(this->MyNLP, this->MyLinSolver, this->discretization->MyOCP_Params, this->discretization->MyOCP_Timings, UsedLineSearchMethod);
        this->LocalInitTimer.toc();
        this->discretization->MyOCP_Timings->TimeInitializeOCP += this->LocalInitTimer.elapsed_s();
    }

    template <typename real>
    InteriorPointMethod<real>::~InteriorPointMethod()
    {
        delete this->discretization;
        delete this->MyLinSolver;
        delete this->MyNLP;
        delete this->MyNLP_Solver;
        delete[] this->SortIndArray;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   public Memberfunktionen                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    bool InteriorPointMethod<real>::optimize(Solution<real> & Sol)
    {
        if (!Sol.isSet)
        {
            IPbasic::Errors::set_Error("Start solution unset!");
            return false;
        }
        if (!this->initStartSolution(Sol))
        {
            IPbasic::Errors::set_Error("Could not initialize start solution!");
            return false;
        }
        bool ret = this->MyNLP_Solver->optimize();
        if (true)
        {
            this->discretization->writeToSolution(Sol);
            Sol.isMaxIter = this->MyNLP_Solver->checkMaxIter();
        }
        this->PrintTiming();
        return ret;
    }

    template <typename real>
    IPbasic::IterationDataPoint<real> InteriorPointMethod<real>::getLastIterationDataPoint() const
    {
        return this->MyNLP_Solver->getLastIterationDataPoint();
    }

    template <typename real>
    IPbasic::IPparameter<real> * InteriorPointMethod<real>::getIPparameter()
    {
        return this->MyNLP_Solver->getIPparameter();
    }

    template <typename real>
    OCPbasic::OCP_Params<real> * InteriorPointMethod<real>::getOCPParameter()
    {
        return this->discretization->MyOCP_Params;
    }

    template <typename real>
    IPbasic::IPtimings<real> * InteriorPointMethod<real>::getIPtimings()
    {
        return this->MyNLP_Solver->getIPtimings();
    }

    template <typename real>
    OCPbasic::OCP_Timings<real> * InteriorPointMethod<real>::getOCPTimings()
    {
        return this->discretization->MyOCP_Timings;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                  private Memberfunktionen                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    bool InteriorPointMethod<real>::initStartSolution(Solution<real> & Sol)
    {
        if (!this->discretization->readFromSolution(Sol))
            return false;
        this->MyNLP->set_initial_guess(this->discretization->x);
        return true;
    }

    template <typename real>
    void InteriorPointMethod<real>::setIndexMap()
    {
        int_type DIM_W = this->P->Ny + this->P->Nu;
        int_type DIM_M = this->P->Ny;
        int_type DIM_C = this->getDiscretization()->MyDims->getBoundDimNLCSTR().N_Dim_ROW + this->getDiscretization()->MyDims->getBoundDimSTATE().N_Dim_ROW +
                         this->getDiscretization()->MyDims->getBoundDimCONTROL().N_Dim_ROW;

        int_type NGrid = this->getDiscretization()->N;
        int_type NRHS  = this->discretization->getDimStateIP() + this->discretization->getDimLambdaIP();
        // Hesse: W
        int_type Index = 0;
        int_type Pos   = 0;
        for (int_type i = 0; i < NGrid + 1; i++)
        {
            Pos = i * (DIM_W + DIM_C + DIM_M);

            for (int_type j = 0; j < DIM_W; j++)
            {
                this->SortIndArray[Index + j] = Pos + j;
            }
            Index += DIM_W;
        }

        // Parameter Hesse:
        Pos = NRHS - this->getProblem()->Np;
        for (int_type j = 0; j < this->P->Np; j++)
        {
            this->SortIndArray[Index + j] = Pos + j;
        }
        Index += this->P->Np;

        // ODE Gradient M, S:
        for (int_type i = 0; i < NGrid; i++)
        {
            Pos = i * (DIM_W + DIM_C + DIM_M) + DIM_W + DIM_C;
            for (int_type j = 0; j < DIM_M; j++)
            {
                this->SortIndArray[Index + j] = Pos + j;
            }
            Index += DIM_M;
        }

        // Randbedingungen psi:
        Pos = (NGrid) * (DIM_W + DIM_C + DIM_M) + DIM_W + DIM_C + this->discretization->MyDims->getBoundDimPARAM().N_Dim_ROW;
        for (int_type j = 0; j < this->discretization->MyDims->getBoundDimBOUND().N_Dim_ROW; j++)
        {
            this->SortIndArray[Index + j] = Pos + j;
        }
        Index += this->discretization->MyDims->getBoundDimBOUND().N_Dim_ROW;

        // Nebenbedingungen:
        for (int_type i = 0; i < NGrid + 1; i++)
        {
            Pos = i * (DIM_W + DIM_C + DIM_M) + DIM_W;
            for (int_type j = 0; j < DIM_C; j++)
            {
                this->SortIndArray[Index + j] = Pos + j;
            }
            Index += DIM_C;
        }

        // Parameter Nebenbedingungen:
        Pos = (NGrid) * (DIM_W + DIM_C + DIM_M) + DIM_W + DIM_C;
        for (int_type j = 0; j < this->discretization->MyDims->getBoundDimPARAM().N_Dim_ROW; j++)
        {
            this->SortIndArray[Index + j] = Pos + j;
        }
        Index += this->discretization->MyDims->getBoundDimPARAM().N_Dim_ROW;
    }

    template <typename real>
    void InteriorPointMethod<real>::PrintTiming()
    {
        IPbasic::IPparameter<real> * Params = this->getIPparameter();
        if (Params->PrintLev > IPbasic::PRINT_NONE)
        {
#ifdef DEBUG
            this->MyNLP_Solver->getIPtimings()->printTimeings(std::cout);
            real TimeFactorizeKKT;
            real TimeSolveKKT;
            real TimeFactorizeBanded;
            real TimeSolveBanded;
            real TimeFactorizeDense;
            real TimeSolveDense;

            LapackKKTSolver<real> * LinSol = dynamic_cast<LapackKKTSolver<real> *>(this->MyLinSolver);

            if (LinSol == nullptr)
                return;

            LinSol->getTimes(TimeFactorizeKKT, TimeSolveKKT, TimeFactorizeBanded, TimeSolveBanded, TimeFactorizeDense, TimeSolveDense);
            std::cout << "Timings KKT(fact, solve):    " << TimeFactorizeKKT << "s" << '\t' << TimeSolveKKT << "s" << std::endl;
            std::cout << "Timings banded(fact, solve): " << TimeFactorizeBanded << "s" << '\t' << TimeSolveBanded << "s" << std::endl;
            std::cout << "Timings dense(fact, solve):  " << TimeFactorizeDense << "s" << '\t' << TimeSolveDense << "s" << std::endl;
#endif
        }
    }

    template class InteriorPointMethod<double>;
    template class InteriorPointMethod<float>;

} // namespace OCPbasic
