/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file ip_nlp.cpp
 * Implementation of the class IP_NLP.
 *
 * @author A. Huber
 * @since 20.02.2019
 */

#include "ip_nlp.h"
#include "../../linalg/sparsematrix.h"
#include <memory.h>

namespace OCPbasic
{

    template <typename real>
    IP_NLP<real>::IP_NLP(Discretization<real> * Diskr)
        : IPbasic::NLP<real>()
        , Diskr(Diskr)
    {
        this->DimX      = this->Diskr->getDimStateIP();
        this->DimConstr = this->Diskr->getDimLambdaIP();
        this->xStart    = new real[this->DimX];

        for (int_type i = 0; i < this->DimX; ++i)
            this->xStart[i] = 1.0;

        // Allokiere Speicher des NLPs:
        this->Jacobi              = new SparseMatrix<real>(true, NOT_SYM);
        this->constraint_jacobian = this->Jacobi;
        this->Hesse               = new SparseMatrix<real>(true, SYM);
        this->lagrangian_hessian  = this->Hesse;
        this->is_equality_vec     = new bool[this->DimConstr];
        this->internalIPMatrix    = new SparseMatrix<real>(true, SYM);
        this->IPMatrix            = this->internalIPMatrix;

        // Initialisiere Speicher des NLPs:
        this->Diskr->JacobiMatrix->getSparseMatrix(*this->Jacobi);
        this->Diskr->HesseMatrix->getSparseMatrix(*this->Hesse);
        int_type NIP = this->state_dimension() + this->constraint_dimension();
        int_type nnz = this->Hesse->getN_Aktuell() + this->Jacobi->getN_Aktuell() + this->constraint_dimension();
        this->IPMatrix->init(NIP, NIP, nnz, false);
        this->setUpBools();
    }

    template <typename real>
    IP_NLP<real>::~IP_NLP()
    {
        delete[] this->xStart;
        // Deallokiere Speicher des NLPs:
        delete this->Jacobi;
        delete this->Hesse;
        delete this->IPMatrix;
        delete[] this->is_equality_vec;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   public Memberfunktionen                                                           //
    //                                                          Overloaded                                                                 //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void IP_NLP<real>::initial_guess(real x[]) const
    {
        memcpy(x, this->xStart, this->DimX * sizeof(real));
    }

    template <typename real>
    int_type IP_NLP<real>::state_dimension() const
    {
        return this->DimX;
    }

    template <typename real>
    int_type IP_NLP<real>::constraint_dimension() const
    {
        return this->DimConstr;
    }

    template <typename real>
    int_type IP_NLP<real>::number_of_equality_constraints() const
    {
        int_type counter = 0;
        for (int_type i = 0; i < this->constraint_dimension(); ++i)
            if (this->is_equality(i))
                ++counter;
        return counter;
    }

    template <typename real>
    int_type IP_NLP<real>::number_of_unilateral_constraints() const
    {
        int_type counter = 0;
        for (int_type i = 0; i < this->constraint_dimension(); ++i)
            if (!this->is_equality(i))
                ++counter;
        return counter;
    }

    template <typename real>
    real IP_NLP<real>::objective(real const states[])
    {
        return this->Diskr->getVARPHI(states);
    }

    template <typename real>
    void IP_NLP<real>::objective_gradient(real const states[], real grad[])
    {
        real varphi = 0.0;
        this->Diskr->Grad_und_VARPHI(states, &varphi, grad);
    }

    template <typename real>
    void IP_NLP<real>::constraint(real const states[], real constr[])
    {
        this->Diskr->CON_Diskr(states, constr);
    }

    template <typename real>
    void IP_NLP<real>::constraint_jacobian_update(real const x[])
    {
        this->Diskr->update_Jacobian(x);
        this->LocalTimerNLP.tic();
        this->Diskr->JacobiMatrix->getSparseMatrix(*this->Jacobi);
        this->LocalTimerNLP.toc();
        this->Diskr->MyOCP_Timings->TimeMatrixConversion += this->LocalTimerNLP.elapsed_s();
    }

    template <typename real>
    void IP_NLP<real>::lagrangian_hessian_update(real, real const x[], real const lambda[])
    {
        this->Diskr->update_Hessian(x, lambda);
        this->Diskr->regularizeHessian();
        this->LocalTimerNLP.tic();
        this->Diskr->HesseMatrix->getSparseMatrix(*this->Hesse);
        if (this->Diskr->MyOCP_Params->MyRegularization == REGULARIZE_GERSCHGORIN)
            this->Hesse->GerschgorinRegularization();
        this->LocalTimerNLP.toc();
        this->Diskr->MyOCP_Timings->TimeMatrixConversion += this->LocalTimerNLP.elapsed_s();
    }

    template <typename real>
    void IP_NLP<real>::ip_hessian_jacobian_update(real, real const x[], real const lambda[])
    {
        this->Diskr->update_IPData(x, lambda);
        this->Diskr->regularizeHessian();
        this->LocalTimerNLP.tic();
        this->Diskr->JacobiMatrix->getSparseMatrix(*this->Jacobi);
        this->Diskr->HesseMatrix->getSparseMatrix(*this->Hesse);
        if (this->Diskr->MyOCP_Params->MyRegularization == REGULARIZE_GERSCHGORIN)
            this->Hesse->GerschgorinRegularization();
        this->LocalTimerNLP.toc();
        this->Diskr->MyOCP_Timings->TimeMatrixConversion += this->LocalTimerNLP.elapsed_s();
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   public Memberfunktionen                                                           //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void IP_NLP<real>::set_initial_guess(real const StartX[])
    {
        memcpy(this->xStart, StartX, this->DimX * sizeof(real));
    }

    template <typename real>
    Discretization<real> * IP_NLP<real>::getOCPDiskr()
    {
        return this->Diskr;
    }

    template <typename real>
    void IP_NLP<real>::SetHasToPushZeros(bool HasToPushZeros)
    {
        this->Jacobi->SetHasToPushZeros(HasToPushZeros);
        this->Hesse->SetHasToPushZeros(HasToPushZeros);
        // this->internalIPMatrix->SetHasToPushZeros(HasToPushZeros);
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                   protected Memberfunktionen                                                        //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void IP_NLP<real>::setUpBools()
    {
        this->Diskr->MyDims->get_isEquVec(this->is_equality_vec);
    }

    template class IP_NLP<double>;
    template class IP_NLP<float>;

} // namespace OCPbasic
