/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file ip_nlp.h
 * Header definition of the class IP_NLP.
 *
 * @author A. Huber
 * @since 20.02.2019
 */

#ifndef IP_NLP_H
#define IP_NLP_H

#include "../../Discretization/Discretization.h"
#include <IPbasic/IPbasic_NLP.h>

namespace OCPbasic
{

    template <typename real>
    class IP_NLP : public IPbasic::NLP<real>
    {
      public:
        /**
         * \brief IP_NLP:
         *        Contructor of the class IP_NLP.
         *
         * \param[in] Diskr   Discretization of the optimal control problem.
         *
         */
        IP_NLP(Discretization<real> * Diskr);

        /**
         * \brief ~IP_NLP:
         *        Virtual destructor of the class IP_NLP.
         */
        virtual ~IP_NLP();

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   public Memberfunktionen                                                           //
        //                                                          Overloaded                                                                 //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief initial_guess:
         *        Returns the initial guess.
         *
         * \param[out] x    Initial guess.
         *
         */
        virtual void initial_guess(real x[]) const;

        /**
         * \brief state_dimension:
         *        Returns the statedimension.
         *
         * @return Returns the statedimension.
         *
         */
        virtual int_type state_dimension() const;

        /**
         * \brief constraint_dimension:
         *        Returns the dimension of the constraints.
         *
         * @return Returns the dimension of the constraints.
         *
         */
        virtual int_type constraint_dimension() const;

        /**
         * \brief number_of_equality_constraints:
         *        Returns the number of equality constraints.
         *
         * @return Returns the number of equality constraints.
         *
         */
        virtual int_type number_of_equality_constraints() const;

        /**
         * \brief number_of_unilateral_constraints:
         *        Returns the number of inequality constraints.
         *
         * @return Returns the number of inequality constraints.
         *
         */
        virtual int_type number_of_unilateral_constraints() const;

        /*\
         |   ___  _     _        _   _
         |  / _ \| |__ (_)___ __| |_(_)_ _____
         | | (_) | '_ \| / -_) _|  _| \ V / -_)
         |  \___/|_.__// \___\__|\__|_|\_/\___|
         |           |__/
        \*/

        /**
         * \brief objective:
         *        Returns the value of the target function.
         *
         * \param[in] states    Statevector.
         *
         * @return Returns the value of the target function.
         *
         */
        virtual real objective(const real states[]);

        /**
         * \brief objective_gradient:
         *        Returns the Gradient of the target function.
         *
         * \param[in]  states    Statevector.
         * \param[out] grad      Gradient of the target function.
         *
         */
        virtual void objective_gradient(const real states[], real grad[]);

        /**
         * \brief constraint:
         *        Returns the constraints of the NLP.
         *
         * \param[in]  states   Statevector.
         * \param[out] constr   Vector with the constraints of the NLP.
         *
         */
        virtual void constraint(const real states[], real constr[]);

        /**
         * \brief constraint_jacobian_update:
         *        Update of the jacobian.
         *
         * \param[in] states     Statevector.
         *
         */
        virtual void constraint_jacobian_update(real const states[]);

        /**
         * \brief lagrangian_hessian_update:
         *        Update of the hessian.
         *
         * \param[in] sigma     Scaling of the target function.
         * \param[in] states    Statevector.
         * \param[in] lambda    Vector of the lagrangian multipliers.
         *
         */
        virtual void lagrangian_hessian_update(real sigma, real const states[], real const lambda[]);

        /**
         * \brief ip_hessian_jacobian_update:
         *        Update of the jacobian and hessian.
         *
         * \param[in] sigma     Scaling of the target function.
         * \param[in] x         Statevector.
         * \param[in] lambda    Vector of the lagrangian multipliers.
         *
         */
        virtual void ip_hessian_jacobian_update(real sigma, real const x[], real const lambda[]);

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   public Memberfunktionen                                                           //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief initial_guess:
         *        Sets the initial guess.
         *
         * \param[out] StartX    Initial guess of the statevector.
         *
         */
        virtual void set_initial_guess(const real StartX[]);

        /**
         * \brief getOCPDiskr:
         *        Returns the discretization of the optimal control problem.
         *
         * @return Returns the discretization of the optimal control problem.
         *
         */
        Discretization<real> * getOCPDiskr();

        /**
         * \brief SetHasToPushZeros:
         *        Sets the HasToPushZeros flag in the sparse matrix.
         *
         * \param[in] HasToPushZeros    Flag to tell wether zeroes should be pushed.
         *
         */
        void SetHasToPushZeros(bool HasToPushZeros);

      protected:
        /// Ddiscretization of the optimal control problem.
        Discretization<real> * Diskr;
        /// Dimension of the states.
        int_type DimX;
        /// Dimension of the constraints.
        int_type DimConstr;
        /// Initial guess of the statevector.
        real * xStart;
        /// Jacobian of the constraints.
        SparseMatrix<real> * Jacobi;
        /// Hessian of the lagrangian.
        SparseMatrix<real> * Hesse;
        /// Member for the internal IPMatrix.
        SparseMatrix<real> * internalIPMatrix;
        /// Local Timer of the nlp.
        TicToc LocalTimerNLP;

        //-------------------------------------------------------------------------------------------------------------------------------------//
        //                                                                                                                                     //
        //                                                   protected Memberfunktionen                                                        //
        //                                                                                                                                     //
        //-------------------------------------------------------------------------------------------------------------------------------------//

        /**
         * \brief setUpBools:
         *        Setup of the bool vector, which tells wether there is an equality or an inequality.
         */
        void setUpBools();
    };

} // namespace OCPbasic

#endif // IP_NLP_H
