/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "BandedLU.h"

#if defined(OCPBASIC_USE_ALGLIN)

#else

#include <IPbasic/IPbasic.h>

#if defined(OCPBASIC_USE_PARALLEL_OPENMP)
#include <omp.h>
#endif

using IPbasic::int_type;

template <typename real>
BandedLU<real>::BandedLU(int_type Dim_RHS, int_type SizeBand)
    : mem_real("BandedLUReal")
    , mem_int("BandedLUIntType")
    , isFactorized(false)
    , Dim_RHS(Dim_RHS)
    , SizeBand(SizeBand)
{
    this->KL   = SizeBand - 1;
    this->KU   = SizeBand - 1;
    this->LDAB = 2 * this->KL + this->KU + 1;
    this->mem_real.allocate(this->LDAB * this->Dim_RHS);
    this->mem_int.allocate(Dim_RHS);
    this->AB   = this->mem_real(size_t(this->LDAB * this->Dim_RHS));
    this->IPIV = this->mem_int(size_t(Dim_RHS));
    this->clear();
}

template <typename real>
BandedLU<real>::~BandedLU()
{
    this->mem_real.free();
    this->mem_int.free();
}

template <typename real>
bool BandedLU<real>::factorize()
{
    if (this->isFactorized)
    {
        Errors::set_Warning("Banded matrix already factorized!");
        return true;
    }

    // Factorize matrix:
    int_type info = this->gbtrf(this->Dim_RHS, this->Dim_RHS, this->KL, this->KU, this->AB, this->LDAB, this->IPIV);

    if (info == 0)
    {
        this->isFactorized = true;
        return true;
    }
    else
    {
        this->isFactorized = false;
        return false;
    }
}

template <typename real>
bool BandedLU<real>::solve(real RHS[])
{
    if (!this->isFactorized)
        return false;
    // Solve matrix:
    int_type info = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, 1, this->AB, this->LDAB, this->IPIV, RHS, this->Dim_RHS);
    if (info == 0)
        return true;
    else
        return false;
}

template <typename real>
void BandedLU<real>::solver(real RHS[], bool & ret)
{
    if (!this->isFactorized)
        ret = false;
    // Solve matrix:
    int_type info = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, 1, this->AB, this->LDAB, this->IPIV, RHS, this->Dim_RHS);
    if (info == 0)
        ret = true;
    else
        ret = false;
}

template <typename real>
bool BandedLU<real>::solve(int_type NRHS, real RHS[])
{

    if (!this->isFactorized)
        return false;
#if defined(OCPBASIC_USE_PARALLEL_OPENMP)
    // If openmp is defined:
    bool ret = true;
    if (omp_get_proc_bind() == omp_proc_bind_false)
    {
        // Solve matrix:
        int_type info = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, NRHS, this->AB, this->LDAB, this->IPIV, RHS, this->Dim_RHS);
        if (info == 0)
            return true;
        else
            return false;
    }
#pragma omp parallel shared(ret, NRHS, RHS, std::cout) default(none)
    {
#pragma omp for
        for (int_type i = 0; i < NRHS; i++)
        {
            real *   dataRHS = &RHS[i * this->Dim_RHS];
            int_type info    = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, 1, this->AB, this->LDAB, this->IPIV, dataRHS, this->Dim_RHS);
            if (info != 0)
            {
#pragma omp critical
                ret = false;
            }
        }
        /*
        #pragma omp single
        for(int_type i = 0; i < NRHS; i++)
        {
            #pragma omp task  shared(ret, NRHS, RHS, std::cout) firstprivate(i) default(none)
            {
            real * dataRHS = &RHS[i*this->Dim_RHS];
            int_type info = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, 1, this->AB, this->LDAB, this->IPIV, dataRHS, this->Dim_RHS);
            if (info != 0)
            {
                #pragma omp critical
                ret = false;
            }
            }
        }
        */
    }
    return ret;
#else
    // Solve matrix:
    int_type info = this->gbtrs('N', this->Dim_RHS, this->KL, this->KU, NRHS, this->AB, this->LDAB, this->IPIV, RHS, this->Dim_RHS);
    if (info == 0)
        return true;
    else
        return false;
#endif
}

template <typename real>
void BandedLU<real>::clear()
{
    std::fill(this->AB, this->AB + size_t(this->LDAB * this->Dim_RHS), real(0.0));
    this->isFactorized = false;
}

template <typename real>
int_type BandedLU<real>::gbtrf(int_type M, int_type N, int_type KL, int_type KU, real AB[], int_type LDAB, int_type IPIV[])
{
    int_type INFO = lapack_wrapper::gbtrf(M, N, KL, KU, AB, LDAB, IPIV);
    return INFO;
}

template <typename real>
int_type BandedLU<real>::gbtrs(char TRANS, int_type N, int_type KL, int_type KU, int_type NRHS, real AB[], int_type LDAB, int_type IPIV[], real B[], int_type LDB)
{
    int_type INFO = 0;
    if (TRANS == 'T')
        INFO = lapack_wrapper::gbtrs(lapack_wrapper::TRANSPOSE, N, KL, KU, NRHS, AB, LDAB, IPIV, B, LDB);
    else
        INFO = lapack_wrapper::gbtrs(lapack_wrapper::NO_TRANSPOSE, N, KL, KU, NRHS, AB, LDAB, IPIV, B, LDB);
    return INFO;
}

template class BandedLU<double>;
template class BandedLU<float>;

#endif
