/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "LapackKKTSolver.h"

#include <IPbasic/IPbasic.h>

#if defined(OCPBASIC_USE_ALGLIN)

#include <IPbasic/Errors/IPbasic_Errors.h>
#include <memory.h>

using IPbasic::Errors;

template <typename real>
LapackKKTSolver<real>::LapackKKTSolver(int_type DimRHS, int_type NDimRowBand, int_type SizeBand, int_type * SortIndexArray)
    : IPbasic::LinearSolver_interface<real>()
    , mem_int("LapackKKTSolverIntType")
    , DimRHS(DimRHS)
    , NDimRowBand(NDimRowBand)
    , SizeBand(SizeBand)
{
    this->AlglinSolver   = new alglin::KKT<real>();
    this->NRest          = this->DimRHS - this->NDimRowBand;
    this->SortIndexArray = new int_type[this->DimRHS];
    this->copyarray      = new real[this->DimRHS];
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        this->SortIndexArray[i] = SortIndexArray[i];
        this->copyarray[i]      = real(0.0);
    }

    // Set Timing to zero.
    this->TimeFactorizeKKT    = real(0.0);
    this->TimeSolveKKT        = real(0.0);
    this->TimeFactorizeBanded = real(0.0);
    this->TimeSolveBanded     = real(0.0);
    this->TimeFactorizeDense  = real(0.0);
    this->TimeSolveDense      = real(0.0);
}

template <typename real>
LapackKKTSolver<real>::~LapackKKTSolver()
{
    delete this->AlglinSolver;
    delete[] this->SortIndexArray;
    delete[] this->copyarray;
}

template <typename real>
bool LapackKKTSolver<real>::factorize(const IPbasic::SparseMatrixBase<real> & Sparse, bool)
{
    const int_type * rows;
    const int_type * cols;
    const real *     vals;
    int_type         nnz = Sparse.get_nnz();
    Sparse.get_data(rows, cols, vals);
    this->mem_int.allocate(2 * nnz);
    int_type * rows_tmp = this->mem_int(size_t(nnz));
    int_type * cols_tmp = this->mem_int(size_t(nnz));
    // Sort Matrix:
    if (Sparse.FORTRAN_indexing())
    {
        for (int_type i = 0; i < nnz; i++)
        {
            rows_tmp[i] = this->SortIndexArray[rows[i] - 1] + 1;
            cols_tmp[i] = this->SortIndexArray[cols[i] - 1] + 1;
        }
    }
    else
    {
        for (int_type i = 0; i < nnz; i++)
        {
            rows_tmp[i] = this->SortIndexArray[rows[i]] + 1;
            cols_tmp[i] = this->SortIndexArray[cols[i]] + 1;
        }
    }

    try
    {
        this->AlglinSolver->load_banded(this->NDimRowBand, this->NRest, this->SizeBand - 1, this->SizeBand - 1, vals, rows_tmp, -1, cols_tmp, -1, nnz, true);
        this->AlglinSolver->factorize();
    }
    catch (std::exception & e)
    {
        std::string err = "Alglin: " + std::string(e.what());
        Errors::set_Error(err);
        this->mem_int.free();
        return false;
    }
    catch (...)
    {
        Errors::set_Error("Solving with AlglinSolver failed!");
        this->mem_int.free();
        return false;
    }
    this->mem_int.free();
    return true;
}

template <typename real>
bool LapackKKTSolver<real>::solve(real RHS[])
{
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        this->copyarray[this->SortIndexArray[i]] = RHS[i];
    }
    try
    {
        this->AlglinSolver->solve(this->copyarray);
    }
    catch (std::exception & e)
    {
        std::string err = "Alglin: " + std::string(e.what());
        Errors::set_Error(err);
        return false;
    }
    catch (...)
    {
        Errors::set_Error("Solving with AlglinSolver failed!");
        return false;
    }

    // Copy Array to RHS:
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        RHS[i] = this->copyarray[this->SortIndexArray[i]];
    }

    return true;
}

#else

#include "BandedLU.h"
#include "DenseLU.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#include <memory.h>

using IPbasic::Errors;

template <typename real>
LapackKKTSolver<real>::LapackKKTSolver(int_type DimRHS, int_type NDimRowBand, int_type SizeBand, int_type * SortIndexArray)
    : IPbasic::LinearSolver_interface<real>()
    , mem_real("LapackKKTSolverReal")
    , mem_int("LapackKKTSolverIntType")
    , DimRHS(DimRHS)
    , NDimRowBand(NDimRowBand)
    , SizeBand(SizeBand)
{
    this->NRest          = this->DimRHS - this->NDimRowBand;
    this->SortIndexArray = new int_type[this->DimRHS];
    this->copyarray      = new real[this->DimRHS];
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        this->SortIndexArray[i] = SortIndexArray[i];
        this->copyarray[i]      = real(0.0);
    }
    // Allocate LU solvers:
    this->MyBandedLUSolver = new BandedLU<real>(this->NDimRowBand, this->SizeBand);
    this->MyDenseLUSolver  = new DenseLU<real>(this->NRest);

    // Allocate submatrices:
    this->mem_real.allocate(size_t(2 * this->NDimRowBand * this->NRest + this->NRest * this->NRest + this->NDimRowBand));
    this->Mem_A_top  = this->mem_real(size_t(this->NDimRowBand * this->NRest));
    this->Mem_A_top2 = this->mem_real(size_t(this->NDimRowBand * this->NRest));
    this->Mem_B      = this->mem_real(size_t(this->NRest * this->NRest));
    this->real_hlp   = this->mem_real(size_t(this->NDimRowBand));
    this->A_top.setup(this->Mem_A_top, this->NDimRowBand, this->NRest, this->NDimRowBand);
    this->A_top2.setup(this->Mem_A_top2, this->NDimRowBand, this->NRest, this->NDimRowBand);
    this->B.setup(this->Mem_B, this->NRest, this->NRest, this->NRest);

    // Set Timing to zero.
    this->TimeFactorizeKKT    = real(0.0);
    this->TimeSolveKKT        = real(0.0);
    this->TimeFactorizeBanded = real(0.0);
    this->TimeSolveBanded     = real(0.0);
    this->TimeFactorizeDense  = real(0.0);
    this->TimeSolveDense      = real(0.0);
}

template <typename real>
LapackKKTSolver<real>::~LapackKKTSolver()
{
    delete[] this->SortIndexArray;
    delete[] this->copyarray;
    // Deallocate LU solvers:
    delete this->MyBandedLUSolver;
    delete this->MyDenseLUSolver;
    this->mem_real.free();
}

template <typename real>
bool LapackKKTSolver<real>::factorize(const IPbasic::SparseMatrixBase<real> & Sparse, bool)
{
    this->KKT_Timer.tic();
    const int_type * rows;
    const int_type * cols;
    const real *     vals;
    int_type         nnz = Sparse.get_nnz();
    Sparse.get_data(rows, cols, vals);
    this->mem_int.allocate(2 * nnz);
    int_type * rows_tmp = this->mem_int(size_t(nnz));
    int_type * cols_tmp = this->mem_int(size_t(nnz));
    // Sort Matrix:
    if (Sparse.FORTRAN_indexing())
    {
        for (int_type i = 0; i < nnz; i++)
        {
            rows_tmp[i] = this->SortIndexArray[rows[i] - 1];
            cols_tmp[i] = this->SortIndexArray[cols[i] - 1];
        }
    }
    else
    {
        for (int_type i = 0; i < nnz; i++)
        {
            rows_tmp[i] = this->SortIndexArray[rows[i]];
            cols_tmp[i] = this->SortIndexArray[cols[i]];
        }
    }
    // Load matrix:
    this->loadSymSparse_C(nnz, rows_tmp, cols_tmp, vals);
    this->MyBandedLUSolver->regularize(real(1e-11));

    // Factorize:
    // Factorize banded matrix:
    this->Internal_Timer.tic();
    if (!this->MyBandedLUSolver->factorize())
    {
        Errors::set_Warning("Regularize banded matrix!");
        this->loadSymSparse_C(nnz, rows_tmp, cols_tmp, vals);
        this->MyBandedLUSolver->regularize(real(1e-8));
        if (!this->MyBandedLUSolver->factorize())
        {
            this->Internal_Timer.toc();
            this->TimeFactorizeBanded += this->Internal_Timer.elapsed_s();
            Errors::set_Error("Could not factorize banded matrix!");
            this->mem_int.free();
            this->KKT_Timer.toc();
            this->TimeFactorizeKKT += this->KKT_Timer.elapsed_s();
            return false;
        }
    }
    this->Internal_Timer.toc();
    this->TimeFactorizeBanded += this->Internal_Timer.elapsed_s();

    this->mem_int.free();
    this->KKT_Timer.toc();
    this->TimeFactorizeKKT += this->KKT_Timer.elapsed_s();
    return true;
}

template <typename real>
bool LapackKKTSolver<real>::solve(real RHS[])
{
    this->KKT_Timer.tic();
    // Copy A_top to A_top2:
    std::copy_n(this->Mem_A_top, this->NDimRowBand * this->NRest, this->Mem_A_top2);
    // Get data:
    real * x         = &this->copyarray[0];
    real * y         = &this->copyarray[this->NDimRowBand];
    real * Mem_Dense = this->MyDenseLUSolver->getValueArray();
    // Copy B to Mem_Dense:
    std::copy_n(this->Mem_B, this->NRest * this->NRest, Mem_Dense);
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        this->copyarray[this->SortIndexArray[i]] = RHS[i];
    }

    // Solve:
    std::copy_n(x, this->NDimRowBand, this->real_hlp);
    // RHS:
    this->Internal_Timer.tic();
    if (!this->MyBandedLUSolver->solve(this->real_hlp))
    {
        Errors::set_Error("Could not solve banded matrix with right-hand-side!");
        this->Internal_Timer.toc();
        this->TimeSolveBanded += this->Internal_Timer.elapsed_s();
        this->KKT_Timer.toc();
        this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
        return false;
    }
    this->Internal_Timer.toc();
    this->TimeSolveBanded += this->Internal_Timer.elapsed_s();
    this->gemv(true, this->NDimRowBand, this->NRest, real(-1.0), this->Mem_A_top, this->NDimRowBand, this->real_hlp, 1, real(1.0), y, 1);
    // Matrix:
    // Get W^-1 A_top into A_top2:
    this->Internal_Timer.tic();
    if (!this->MyBandedLUSolver->solve(this->NRest, this->Mem_A_top2))
    {
        Errors::set_Error("Could not solve banded matrix with right-hand-side!");
        this->Internal_Timer.toc();
        this->TimeSolveBanded += this->Internal_Timer.elapsed_s();
        this->KKT_Timer.toc();
        this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
        return false;
    }
    this->Internal_Timer.toc();
    this->TimeSolveBanded += this->Internal_Timer.elapsed_s();
    // Get B-A W^-1 A_top int B:
    this->gemm(true, false, this->NRest, this->NRest, this->NDimRowBand, real(-1.0), this->Mem_A_top, this->NDimRowBand, this->Mem_A_top2, this->NDimRowBand, real(1.0), Mem_Dense, this->NRest);
    // Now B-A W^-1 A_top is part of MyDenseLUSolver:
    this->Internal_Timer.tic();
    if (!this->MyDenseLUSolver->factorize())
    {
        Errors::set_Error("Could not factorize dense matrix!");
        this->Internal_Timer.toc();
        this->TimeFactorizeDense += this->Internal_Timer.elapsed_s();
        this->KKT_Timer.toc();
        this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
        return false;
    }
    this->Internal_Timer.toc();
    this->TimeFactorizeDense += this->Internal_Timer.elapsed_s();
    this->Internal_Timer.tic();
    if (!this->MyDenseLUSolver->solve(y))
    {
        Errors::set_Error("Could not solve dense matrix with right-hand-side!");
        this->Internal_Timer.toc();
        this->TimeSolveDense += this->Internal_Timer.elapsed_s();
        this->KKT_Timer.toc();
        this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
        return false;
    }
    this->Internal_Timer.toc();
    this->TimeSolveDense += this->Internal_Timer.elapsed_s();
    // Solve for X:
    // RHS:
    this->gemv(false, this->NDimRowBand, this->NRest, real(-1.0), this->Mem_A_top, this->NDimRowBand, y, 1, real(1.0), x, 1);
    // Solve with Banded Matrix:
    this->Internal_Timer.tic();
    if (!this->MyBandedLUSolver->solve(x))
    {
        Errors::set_Error("Could not solve banded matrix with right-hand-side!");
        this->Internal_Timer.toc();
        this->TimeSolveBanded += this->Internal_Timer.elapsed_s();
        this->KKT_Timer.toc();
        this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
        return false;
    }
    this->Internal_Timer.toc();
    this->TimeSolveBanded += this->Internal_Timer.elapsed_s();

    // Copy Array to RHS:
    for (int_type i = 0; i < this->DimRHS; i++)
    {
        RHS[i] = this->copyarray[this->SortIndexArray[i]];
    }

    this->KKT_Timer.toc();
    this->TimeSolveKKT += this->KKT_Timer.elapsed_s();
    return true;
}

template <typename real>
void LapackKKTSolver<real>::loadSymSparse_C(int_type nnz, const int_type rows[], const int_type cols[], const real vals[])
{
    this->clear();
    int_type i, j;
    for (int_type index = 0; index < nnz; index++)
    {
        i = rows[index];
        j = cols[index];
        if (i < this->NDimRowBand && j < this->NDimRowBand)
        {
            // Load W: (banded matrix)
            this->MyBandedLUSolver->push_Val_C(i, j, vals[index]);
            if (i != j)
                this->MyBandedLUSolver->push_Val_C(j, i, vals[index]);
        }
        else if (i >= this->NDimRowBand && j < this->NDimRowBand)
        {
            // Load A':
            this->A_top(j, i - this->NDimRowBand) = vals[index];
        }
        else if (i < this->NDimRowBand && j >= this->NDimRowBand)
        {
            // Load A':
            this->A_top(i, j - this->NDimRowBand) = vals[index];
        }
        else
        {
            // Load B:
            this->B(i - this->NDimRowBand, j - this->NDimRowBand) = vals[index];
            if (i != j)
                this->B(j - this->NDimRowBand, i - this->NDimRowBand) = vals[index];
        }
    }
}

template <typename real>
void LapackKKTSolver<real>::loadSymSparse_Fortran(int_type nnz, const int_type rows[], const int_type cols[], const real vals[])
{
    this->clear();
    int_type i, j;
    for (int_type index = 0; index < nnz; index++)
    {
        i = rows[index] - 1;
        j = cols[index] - 1;
        if (i < this->NDimRowBand && j < this->NDimRowBand)
        {
            // Load W: (banded matrix)
            this->MyBandedLUSolver->push_Val_C(i, j, vals[index]);
        }
        else if (i >= this->NDimRowBand && j < this->NDimRowBand)
        {
            // Load A':
            this->A_top(j, i - this->NDimRowBand) = vals[index];
        }
        else if (i < this->NDimRowBand && j >= this->NDimRowBand)
        {
            // Load A':
            this->A_top(i, j - this->NDimRowBand) = vals[index];
        }
        else
        {
            // Load B:
            this->B(i - this->NDimRowBand, j - this->NDimRowBand) = vals[index];
            if (i != j)
                this->B(j - this->NDimRowBand, i - this->NDimRowBand) = vals[index];
        }
    }
}

template <typename real>
void LapackKKTSolver<real>::clear()
{
    this->MyBandedLUSolver->clear();
    this->A_top.zero_fill();
    this->B.zero_fill();
}

template <typename real>
void LapackKKTSolver<real>::gemv(bool transA, int_type M, int_type N, real alpha, real A[], int_type LDA, real X[], int_type incX, real beta, real Y[], int_type incY)
{
    lapack_wrapper::Transposition tra = lapack_wrapper::NO_TRANSPOSE;
    if (transA)
    {
        tra = lapack_wrapper::TRANSPOSE;
    }
    lapack_wrapper::gemv(tra, M, N, alpha, A, LDA, X, incX, beta, Y, incY);
}

template <typename real>
void LapackKKTSolver<real>::gemm(bool transA, bool transB, int_type M, int_type N, int_type K, real alpha, real A[], int_type LDA, real B[], int_type LDB, real beta, real C[], int_type LDC)
{
    lapack_wrapper::Transposition traA = lapack_wrapper::NO_TRANSPOSE;
    lapack_wrapper::Transposition traB = lapack_wrapper::NO_TRANSPOSE;
    if (transA)
    {
        traA = lapack_wrapper::TRANSPOSE;
    }
    if (transB)
    {
        traB = lapack_wrapper::TRANSPOSE;
    }
    lapack_wrapper::gemm(traA, traB, M, N, K, alpha, A, LDA, B, LDB, beta, C, LDC);
}

#endif

// Functions for all:

template <typename real>
void LapackKKTSolver<real>::getTimes(real & TimeFactorizeKKT, real & TimeSolveKKT, real & TimeFactorizeBanded, real & TimeSolveBanded, real & TimeFactorizeDense, real & TimeSolveDense)
{
    TimeFactorizeKKT    = this->TimeFactorizeKKT;
    TimeSolveKKT        = this->TimeSolveKKT;
    TimeFactorizeBanded = this->TimeFactorizeBanded;
    TimeSolveBanded     = this->TimeSolveBanded;
    TimeFactorizeDense  = this->TimeFactorizeDense;
    TimeSolveDense      = this->TimeSolveDense;
}

template class LapackKKTSolver<double>;
template class LapackKKTSolver<float>;
