/* (C) 2019, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "OCPbasic_SymEigenValues.h"
#include "../OCPbasic_config.h"
#include "OCPbasic_HouseholderQR.h"
#include <lapack_wrapper/lapack_wrapper++.hh>

// OCPBASIC_USE_EIGENVAL_LAPACK_WRAPPER, OCPBASIC_USE_EIGENVAL_SYM, OCPBASIC_USE_EIGENVAL_OWN
#ifdef OCPBASIC_USE_EIGENVAL_SYM
#include "OCPbasic_lapack_blas.h"
#endif

namespace OCPbasic
{

    template <typename real>
    SymEigenValues<real>::SymEigenValues()
        : mem_real("SymEigenValuesReal")
        , mem_int("SymEigenValuesInt")
    {
        this->tolQR   = real(1e-8);
        this->tolDiag = real(1e-10);
        this->Maxiter = 100;
    }

    template <typename real>
    SymEigenValues<real>::~SymEigenValues()
    {
    }

#ifdef OCPBASIC_USE_EIGENVAL_SYM

    template <>
    int_type SymEigenValues<double>::get_Symmetric_Eigenvalues(MatrixWrapper<double> const & ConstA, double eigenVals[])
    {
        int_type nr    = ConstA.numRows();
        int_type nc    = ConstA.numCols();
        int_type lda   = ConstA.lDim();
        int_type nnz   = nr * nc;
        int_type lwork = 3 * nr;
        int_type info  = 0;
        char     jobz  = 'N';
        char     uplo  = 'L';
        if (nr < 0 || nc < 0)
            return -1;
        if (lda < nr)
            return -1;
        if (nr != nc)
            return -1;
        // A symmetric matrix should also be a square matrix.

        // Reserve temporary memory.
        this->mem_real.allocate(size_t(nnz + lwork));
        double * mem_matirx = this->mem_real(size_t(nnz));
        double * work       = this->mem_real(size_t(lwork));
        std::copy_n(ConstA.get_data(), nnz, mem_matirx);
        dsyev_(&jobz, &uplo, &nr, mem_matirx, &lda, eigenVals, work, &lwork, &info);
        this->mem_real.free();
        return info;
    }

    template <>
    int_type SymEigenValues<float>::get_Symmetric_Eigenvalues(MatrixWrapper<float> const & ConstA, float eigenVals[])
    {
        int_type nr    = ConstA.numRows();
        int_type nc    = ConstA.numCols();
        int_type lda   = ConstA.lDim();
        int_type nnz   = nr * nc;
        int_type lwork = 3 * nr;
        int_type info  = 0;
        char     jobz  = 'N';
        char     uplo  = 'L';
        if (nr < 0 || nc < 0)
            return -1;
        if (lda < nr)
            return -1;
        if (nr != nc)
            return -1;
        // A symmetric matrix should also be a square matrix.

        // Reserve temporary memory.
        this->mem_real.allocate(size_t(nnz + lwork));
        float * mem_matirx = this->mem_real(size_t(nnz));
        float * work       = this->mem_real(size_t(lwork));
        std::copy_n(ConstA.get_data(), nnz, mem_matirx);
        ssyev_(&jobz, &uplo, &nr, mem_matirx, &lda, eigenVals, work, &lwork, &info);
        this->mem_real.free();
        return info;
    }
#elif defined OCPBASIC_USE_EIGENVAL_LAPACK_WRAPPER

    template <typename real>
    int_type SymEigenValues<real>::get_Symmetric_Eigenvalues(MatrixWrapper<real> const & ConstA, real eigenVals[])
    {
        real                              tol = 1e-8;
        lapack_wrapper::Eigenvalues<real> eig(ConstA);
        std::vector<real>                 re;
        std::vector<real>                 im;
        eig.getEigenvalues(re, im);

        if (re.size() != im.size())
            return -1;

        for (size_t i = 0; i < re.size(); i++)
        {
            if (fabs(im.at(i)) > tol)
                return -2;
            else
                eigenVals[i] = re.at(i);
        }

        return 0;
    }

#elif defined OCPBASIC_USE_EIGENVAL_OWN
    template <typename real>
    int_type SymEigenValues<real>::get_Symmetric_Eigenvalues(MatrixWrapper<real> const & ConstA, real eigenVals[])
    {
        int_type nr  = ConstA.numRows();
        int_type nc  = ConstA.numCols();
        int_type lda = ConstA.lDim();
        int_type nnz = nr * nc;
        if (nr < 0 || nc < 0)
            return -1;
        if (lda < nr)
            return -1;
        if (nr != nc)
            return -1;
        // A symmetric matrix should also be a square matrix.

        // Reserve temporary memory.

        this->mem_int.allocate(size_t(nc));
        this->mem_real.allocate(size_t(2 * nnz + nr));

        real *                mem_First = this->mem_real(size_t(nnz));
        real *                mem_Sec   = this->mem_real(size_t(nnz));
        MatrixWrapper<real>   First(mem_First, nr, nc, lda);
        MatrixWrapper<real>   Sec(mem_Sec, nr, nc, lda);
        MatrixWrapper<real> * pQR = &First;
        MatrixWrapper<real> * pR  = &Sec;
        First.load(ConstA);
        int_type * ipvt   = this->mem_int(size_t(nc));
        real *     qr_aux = this->mem_real(size_t(nr));

        int_type rank     = 0;
        int_type Info     = 0;
        real     mu_shift = real(0.0);

        for (int_type i = 0; i < this->Maxiter; ++i)
        {
            // Make shift:
            for (int_type k = 0; k < nr; ++k)
                (*pQR)(k, k) -= mu_shift;
            // Make QR factorisation:
            Info = QR(*pQR, ipvt, qr_aux, this->tolQR, false, rank);
            if (0 != Info)
            {
                this->mem_int.free();
                this->mem_real.free();
                return -2;
            }

            // Get Q and R:
            Info = get_R(*pQR, ipvt, *pR);
            if (0 != Info)
            {
                this->mem_int.free();
                this->mem_real.free();
                return -3;
            }

            // Update iteration matrix:
            Info = QMult_Right(*pQR, ipvt, qr_aux, *pR);
            if (0 != Info)
            {
                this->mem_int.free();
                this->mem_real.free();
                return -3;
            }

            std::swap(pQR, pR);
            // Undo shift:
            for (int_type k = 0; k < nr; ++k)
                (*pQR)(k, k) += mu_shift;
            // Change shift:
            mu_shift = (*pQR)(nr - 1, nr - 1);
            if (this->checkMatrix(*pQR))
            {
                // std::cout << "Number of iterations: " << i << std::endl;
                break;
            }
        }

        this->get_Diag(*pQR, eigenVals);
        this->mem_int.free();
        this->mem_real.free();
        return 0;
    }

#else
#error "Set a define for Eigenvalues! (OCPBASIC_USE_EIGENVAL_LAPACK_WRAPPER, OCPBASIC_USE_EIGENVAL_SYM, OCPBASIC_USE_EIGENVAL_OWN)"
#endif

    template <typename real>
    void SymEigenValues<real>::get_Diag(MatrixWrapper<real> const & A, real Diag[])
    {
        int_type nr = A.numRows();
        for (int_type i = 0; i < nr; ++i)
            Diag[i] = A(i, i);
    }

    template <typename real>
    bool SymEigenValues<real>::checkMatrix(MatrixWrapper<real> const & A)
    {
        for (int_type i = 0; i < A.numRows(); ++i)
        {
            for (int_type j = 0; j < A.numCols(); ++j)
            {
                if (i != j)
                {
                    if (std::abs(A(i, j)) > this->tolDiag)
                        return false;
                }
            }
        }
        return true;
    }

    template class SymEigenValues<double>;
    template class SymEigenValues<float>;

} // namespace OCPbasic
