/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file collectormatrix.cpp
 * Implementierung der Klasse CollectorMatrix.
 *
 * @author A. Huber
 * @since 08.12.2016
 */

#include "collectormatrix.h"
#include <IPbasic/Errors/IPbasic_Errors.h>

using IPbasic::Errors;

template <typename real>
CollectorMatrix<real>::CollectorMatrix(int_type N_Coll_Row, int_type N_Coll_Col, int_type const dimRow[], int_type const dimCol[], isSym Flag_Sym)
    : nRows(0)
    , nCols(0)
    , NNz_Estimate(0)
    , Flag_Sym(Flag_Sym)
{
    this->N_Coll_Row = N_Coll_Row;
    this->N_Coll_Col = N_Coll_Col;

    this->dimRow.resize(this->N_Coll_Row);
    this->dimCol.resize(this->N_Coll_Col);

    this->ResultingDimRow.resize(this->N_Coll_Row + 1);
    this->ResultingDimCol.resize(this->N_Coll_Col + 1);

    if ((SYM == Flag_Sym) && (this->N_Coll_Row == this->N_Coll_Col))
    {
        // Kopiere rown und Splaten:
        for (int_type i = 0; i < this->N_Coll_Row; i++)
        {
            this->dimRow[i] = dimRow[i];
            this->dimCol[i] = dimCol[i];
            if (this->dimRow[i] != this->dimCol[i])
                Errors::set_Error("Symmetrische Matrix kann nicht unsymmetrisch belegt werden!");
        }
    }
    else if (NOT_SYM == Flag_Sym)
    {
        // Kopiere Rows:
        for (int_type i = 0; i < this->N_Coll_Row; i++)
            this->dimRow[i] = dimRow[i];
        // Kopiere Column:
        for (int_type i = 0; i < this->N_Coll_Col; i++)
            this->dimCol[i] = dimCol[i];
    }
    else
    {
        Errors::set_Error("Symmetrische Matrix kann nicht unsymmetrisch belegt werden!");
    }
    // Belege die Arrays der resultierenden Dimensionen.
    this->internalSetResultingDims();
}

template <typename real>
CollectorMatrix<real>::~CollectorMatrix()
{
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                     public Memberfunktionen                                                         //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

template <typename real>
isSym CollectorMatrix<real>::getFlag_Sym()
{
    return this->Flag_Sym;
}

template <typename real>
int_type CollectorMatrix<real>::getNMaxSparse()
{
    return this->NNz_Estimate;
}

template <typename real>
void CollectorMatrix<real>::getSparseMatrix(SparseMatrix<real> & Sparse)
{
    if (((this->Flag_Sym == SYM) && Sparse.isSymmetric()) || ((this->Flag_Sym == NOT_SYM) && !Sparse.isSymmetric()))
    {
        Sparse.setZero();
        int_type dim_i_Sparse = this->getRowDimSparse();
        int_type dim_j_Sparse = this->getColDimSparse();
        Sparse.setDimension(dim_i_Sparse, dim_j_Sparse);
        // Untransponierte Matrix Anteile:
        for (size_t i = 0; i < this->AlglinMatrixContainer.size(); i++)
        {
            int_type i_Sparse = this->ResultingDimRow[this->Rows[i]];
            int_type j_Sparse = this->ResultingDimCol[this->Cols[i]];
            Sparse.PushMatrix(i_Sparse, j_Sparse, *this->AlglinMatrixContainer[i]);
        }
        // Transponierte Matrix Anteile:
        for (size_t i = 0; i < this->AlglinMatrixContainer_Transposed.size(); i++)
        {
            int_type i_Sparse = this->ResultingDimRow[this->Row_Transposed[i]];
            int_type j_Sparse = this->ResultingDimCol[this->Col_Transposed[i]];
            Sparse.PushMatrixTransposed(i_Sparse, j_Sparse, *this->AlglinMatrixContainer_Transposed[i]);
        }
        // Diagonalmatrix Anteile:
        for (size_t i = 0; i < this->AlglinMatrixContainer_Diag.size(); i++)
        {
            int_type i_Sparse = this->ResultingDimRow[this->Row_Diag[i]];
            int_type j_Sparse = this->ResultingDimCol[this->Col_Diag[i]];
            for (int_type j = 0; j < this->AlglinMatrixContainer_Diag[i]->getDim(); j++)
                Sparse.PushVal(i_Sparse + j, j_Sparse + j, (*this->AlglinMatrixContainer_Diag[i])[j]);
        }
    }
    else
    {
        Errors::set_Error("Symmetrieeigenschaften von CollectorMatrix und SparseMatrix stimmen nicht überein!");
    }
}

template <typename real>
void CollectorMatrix<real>::getMatrix(MatrixWrapper<real> & M)
{
    M.zero_fill();
    // Untransponierte Matrix Anteile:
    for (size_t i = 0; i < this->AlglinMatrixContainer.size(); i++)
    {
        int_type                    i_Matrix = this->ResultingDimRow[this->Rows[i]];
        int_type                    j_Matrix = this->ResultingDimCol[this->Cols[i]];
        MatrixWrapper<real> const & MI       = *this->AlglinMatrixContainer[i];
        M.load_block(MI, i_Matrix, j_Matrix);
        if (this->Flag_Sym == SYM && i_Matrix != j_Matrix)
            M.load_block_transposed(MI, j_Matrix, i_Matrix);
    }
    // Transponierte Matrix Anteile:
    for (size_t i = 0; i < this->AlglinMatrixContainer_Transposed.size(); i++)
    {
        int_type                    i_Matrix = this->ResultingDimRow[this->Row_Transposed[i]];
        int_type                    j_Matrix = this->ResultingDimCol[this->Col_Transposed[i]];
        MatrixWrapper<real> const * MIJ      = this->AlglinMatrixContainer_Transposed[i];
        M.load_block_transposed(*MIJ, i_Matrix, j_Matrix);
        if (this->Flag_Sym == SYM && i_Matrix != j_Matrix)
            M.load_block(*MIJ, j_Matrix, i_Matrix);
    }
    // Diagonalmatrix Anteile:
    for (size_t i = 0; i < this->AlglinMatrixContainer_Diag.size(); i++)
    {
        int_type                        i_Matrix = this->ResultingDimRow[this->Row_Diag[i]];
        int_type                        j_Matrix = this->ResultingDimCol[this->Col_Diag[i]];
        DiagMatrixWrapper<real> const * DM       = this->AlglinMatrixContainer_Diag[i];
        M.load_block(*DM, i_Matrix, j_Matrix);
        if (this->Flag_Sym == SYM && i_Matrix != j_Matrix)
            M.load_block(*DM, j_Matrix, i_Matrix);
    }
}

template <typename real>
int_type CollectorMatrix<real>::getRowDimSparse()
{
    return this->ResultingDimRow[this->N_Coll_Row];
}

template <typename real>
int_type CollectorMatrix<real>::getColDimSparse()
{
    return this->ResultingDimCol[this->N_Coll_Col];
}

template <typename real>
void CollectorMatrix<real>::clear()
{
    this->AlglinMatrixContainer.clear();
    this->Rows.clear();
    this->Cols.clear();
    this->AlglinMatrixContainer_Transposed.clear();
    this->Row_Transposed.clear();
    this->Col_Transposed.clear();
    this->AlglinMatrixContainer_Diag.clear();
    this->Row_Diag.clear();
    this->Col_Diag.clear();
    this->NNz_Estimate = 0;
}

template <typename real>
void CollectorMatrix<real>::PushMatrix(int_type row, int_type col, MatrixWrapper<real> * M)
{
    if (this->Flag_Sym == SYM)
    {
        if (col <= row)
        {
            this->internalMatrixPush(row, col, M);
        }
        else
        {
            Errors::set_Warning("Unsymmetrischer Push in symmetrischer CollectorMatrix!");
        }
    }
    else
    {
        this->internalMatrixPush(row, col, M);
    }
}

template <typename real>
void CollectorMatrix<real>::PushMatrixTransposed(int_type row, int_type col, MatrixWrapper<real> * M)
{
    if (this->Flag_Sym == SYM)
    {
        if (col <= row)
            this->internalMatrixPushTransposed(row, col, M);
        else
            Errors::set_Warning("Unsymmetrischer Push in symmetrischer CollectorMatrix!");
    }
    else
    {
        this->internalMatrixPushTransposed(row, col, M);
    }
}

template <typename real>
void CollectorMatrix<real>::PushDiagMatrix(int_type row, int_type col, DiagMatrixWrapper<real> * D)
{
    if (this->Flag_Sym == SYM)
    {
        if (col <= row)
            this->internalMatrixPushDiag(row, col, D);
        else
            Errors::set_Warning("Unsymmetrischer Push in symmetrischer CollectorMatrix!");
    }
    else
    {
        this->internalMatrixPushDiag(row, col, D);
    }
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                    private Memberfunktionen                                                         //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

template <typename real>
void CollectorMatrix<real>::internalMatrixPush(int_type row, int_type col, MatrixWrapper<real> * M)
{
    if (M->numRows() == this->dimRow[row] && M->numCols() == this->dimCol[col])
    {
        this->AlglinMatrixContainer.push_back(M);
        this->Rows.push_back(row);
        this->Cols.push_back(col);
        this->NNz_Estimate += M->numRows() * M->numCols();
    }
    else
    {
        Errors::set_Error("Matrix Push an unzulässiger Stelle!");
    }
}

template <typename real>
void CollectorMatrix<real>::internalMatrixPushTransposed(int_type row, int_type col, MatrixWrapper<real> * M)
{
    int_type RowDimM = M->numCols();
    int_type ColDimM = M->numRows();
    if (RowDimM == this->dimRow[row] && ColDimM == this->dimCol[col])
    {
        this->AlglinMatrixContainer_Transposed.push_back(M);
        this->Row_Transposed.push_back(row);
        this->Col_Transposed.push_back(col);
        this->NNz_Estimate += M->numRows() * M->numCols();
    }
    else
    {
        Errors::set_Error("Matrix Push an unzulässiger Stelle!");
    }
}

template <typename real>
void CollectorMatrix<real>::internalMatrixPushDiag(int_type row, int_type col, DiagMatrixWrapper<real> * D)
{
    int_type Dim = D->getDim();
    if (Dim == this->dimRow[row] && Dim == this->dimCol[col])
    {
        this->AlglinMatrixContainer_Diag.push_back(D);
        this->Row_Diag.push_back(row);
        this->Col_Diag.push_back(col);
        this->NNz_Estimate += Dim;
    }
    else
    {
        Errors::set_Error("Matrix Push an unzulässiger Stelle!");
    }
}

template <typename real>
void CollectorMatrix<real>::internalSetResultingDims()
{
    // Belege rows:
    this->ResultingDimRow[0] = 0;
    for (int_type i = 1; i <= this->N_Coll_Row; ++i)
        this->ResultingDimRow[i] = this->ResultingDimRow[i - 1] + this->dimRow[i - 1];

    // Belege Column:
    this->ResultingDimCol[0] = 0;
    for (int_type i = 1; i <= this->N_Coll_Col; ++i)
        this->ResultingDimCol[i] = this->ResultingDimCol[i - 1] + this->dimCol[i - 1];

    this->nRows = this->ResultingDimRow[this->N_Coll_Row];
    this->nCols = this->ResultingDimCol[this->N_Coll_Col];
}

template class CollectorMatrix<double>;
template class CollectorMatrix<float>;
