/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file sparsematrix.cpp
 * Implementierung der Klasse SparseMatrix.
 *
 * @author A. Huber
 * @since 08.11.2016
 */

#include "sparsematrix.h"
#include <IPbasic/Errors/IPbasic_Errors.h>
#include <fstream>
#include <iostream>
#include <math.h>

using IPbasic::Errors;

// Belegen der statischen Variablen:
template <typename real>
int_type SparseMatrix<real>::NNz_Estimate = 0;

template <typename real>
real SparseMatrix<real>::tol_zero = 1e-12;

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                     public Memberfunktionen                                                         //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

template <typename real>
void SparseMatrix<real>::PushVal(int_type row, int_type col, real val)
{
    if (this->has_to_push_zeroes == false && std::abs(val) < SparseMatrix<real>::tol_zero)
        return;
    if (this->Flag_Sym == SYM)
    {
        // Überprüfung der Dimensionen:
        if (row >= this->nRows)
        {
            this->nRows = row + 1;
            this->nCols = this->nRows;
        }
        if (col >= this->nCols)
        {
            this->nCols = col + 1;
            this->nRows = this->nCols;
        }
        if (col <= row)
            this->PushBackNote(row, col, val);
    }
    else if (this->Flag_Sym == NOT_SYM)
    {
        // Überprüfung der Dimensionen:
        if (row >= this->nRows)
            this->nRows = row + 1;
        if (col >= this->nCols)
            this->nCols = col + 1;
        this->PushBackNote(row, col, val);
    }
    else
    {
        Errors::set_Error("Falscher Flag im Bezug auf die Symmetrie bei einer Sparsematrix!");
    }
}

template <typename real>
void SparseMatrix<real>::PushMatrix(int_type row, int_type col, MatrixWrapper<real> const & M)
{
    if (this->nRows < row + M.numRows())
    {
        this->nRows = row + M.numRows();
        if (this->isSymmetric())
            this->nCols = this->nRows;
    }
    if (this->nCols < col + M.numCols())
    {
        this->nCols = col + M.numCols();
        if (this->isSymmetric())
            this->nRows = this->nCols;
    }
    for (int_type i = 0; i < M.numRows(); ++i)
    {
        for (int_type j = 0; j < M.numCols(); ++j)
        {
            this->PushVal(row + i, col + j, M(i, j));
        }
    }
}

template <typename real>
void SparseMatrix<real>::PushMatrixTransposed(int_type row, int_type col, MatrixWrapper<real> const & M)
{
    int_type rowDim = M.numCols();
    int_type colDim = M.numRows();
    if (this->nRows < row + rowDim)
    {
        this->nRows = row + rowDim;
        if (this->isSymmetric())
            this->nCols = this->nRows;
    }
    if (this->nCols < col + colDim)
    {
        this->nCols = col + colDim;
        if (this->isSymmetric())
            this->nRows = this->nCols;
    }
    for (int_type i = 0; i < M.numRows(); ++i)
    {
        for (int_type j = 0; j < M.numCols(); ++j)
        {
            this->PushVal(row + j, col + i, M(i, j));
        }
    }
}

template <typename real>
int_type SparseMatrix<real>::getN_Aktuell()
{
    if (this->nnz == (int_type)this->rows.size() && this->nnz == (int_type)this->cols.size() && this->nnz == (int_type)this->vals.size())
    {
        return this->nnz;
    }
    else if (this->rows.size() == this->cols.size() && this->rows.size() == this->vals.size())
    {
        Errors::set_Error("N_Aktuell wurde nicht aktualisiert! Versuche autofix!");
        this->nnz = (int)this->vals.size();
        return this->nnz;
    }
    else
    {
        Errors::set_Error("Indexarrays und Wertearray einer Sparsematrix haben unterschiedliche Größen!");
        return this->nnz;
    }
}

template <typename real>
void SparseMatrix<real>::export_data(int_type N_Aktuell, int_type i[], int_type j[], real val[])
{
    if (N_Aktuell == (int)this->rows.size() && N_Aktuell == (int)this->cols.size() && N_Aktuell == (int)this->vals.size())
    {
        std::copy_n(this->rows.begin(), N_Aktuell, i);
        std::copy_n(this->cols.begin(), N_Aktuell, j);
        std::copy_n(this->vals.begin(), N_Aktuell, val);
    }
    else
    {
        Errors::set_Error("Fehler in der Funktion copyArrays: Angegbene Dimension stimmt nicht mit der Dimension des Sparsematrix überein!");
    }
}

template <typename real>
void SparseMatrix<real>::FillDiag()
{
    bool has_to_push_zeroes  = this->has_to_push_zeroes;
    this->has_to_push_zeroes = true;
    if (this->Flag_Sym == SYM && this->nRows == this->nCols)
    {
        std::vector<bool> IDiag(this->nRows);
        std::fill(IDiag.begin(), IDiag.end(), true);
        for (int_type Index = 0; Index < this->getN_Aktuell(); ++Index)
            if (this->rows[Index] == this->cols[Index])
                IDiag[this->rows[Index] - 1] = false;
        for (int_type Index = 0; Index < this->nRows; ++Index)
            if (IDiag[Index])
                this->PushVal(Index, Index, 0.0);
    }
    else if (this->nRows == this->nCols)
    {
        Errors::set_Error("Symetrische Matrix hat nicht die gleiche rown- und Column anzahl!");
    }
    else
    {
        Errors::set_Error("FillDiag bei unsymetrischer Matrix!");
    }
    this->has_to_push_zeroes = has_to_push_zeroes;
}

template <typename real>
void SparseMatrix<real>::GerschgorinRegularization()
{
    // Allokiere Speicher:
    std::vector<real> reg(this->nRows);
    std::vector<real> vals(this->nRows);
    real              hlp_val = 0.0;
    real              reg_min = 0.0;

    // Initialisierung mit dem Wert 0.0:
    for (int_type i = 0; i < this->nRows; i++)
    {
        reg[i]  = hlp_val;
        vals[i] = hlp_val;
    }

    // Aufsummieren der Column:
    for (int_type Index = 0; Index < this->getN_Aktuell(); Index++)
    {
        int_type i = this->rows[Index];
        int_type j = this->cols[Index];
        real     v = this->vals[Index];
        if (i != j)
        {
            if (this->fortran_indexing)
                --i;
            reg[i] += std::abs(v);
        }
        else
        {
            if (this->fortran_indexing)
                --i;
            vals[i] = v;
        }
    }

    // Ermitteln einer Abschätzung des kleinsten Eigenwerts:
    for (int_type i = 0; i < this->nRows; ++i)
    {
        hlp_val = vals[i] - 2 * reg[i];
        if (hlp_val < reg_min)
            reg_min = hlp_val;
    }

    // Abzug auf der Hauptdiagonalen:
    for (int_type Index = 0; Index < this->getN_Aktuell(); Index++)
    {
        int_type i = this->rows[Index];
        int_type j = this->cols[Index];
        if (i == j)
            this->vals[Index] -= reg_min - 1E-8;
    }
}

template <typename real>
bool SparseMatrix<real>::compareDim(const SparseMatrix & B)
{
    if (this->nnz == (int)this->rows.size() && this->nnz == (int)this->cols.size() && this->nnz == (int)this->vals.size())
    {
        if (B.nnz == (int)B.rows.size() && B.nnz == (int)B.cols.size() && B.nnz == (int)B.vals.size())
        {
            return this->nnz == B.nnz && this->nRows == B.nRows && this->nCols == B.nCols;
        }
        else
        {
            Errors::set_Error("Fehler in der Funktion compareDim: Angegbene Dimension stimmt nicht mit der Dimension des Sparsematrix überein!");
            return false;
        }
    }
    else
    {
        Errors::set_Error("Fehler in der Funktion compareDim: Angegbene Dimension stimmt nicht mit der Dimension des Sparsematrix überein!");
        return false;
    }
}

template <typename real>
bool SparseMatrix<real>::isSymmetric() const
{
    switch (this->Flag_Sym)
    {
    case SYM:
        return true;
        break;
    case NOT_SYM:
        return false;
        break;
    default:
        return false;
        break;
    }
}

template <typename real>
void SparseMatrix<real>::SymmetricalFilling()
{
    if (this->Flag_Sym == SYM)
    {
        this->Flag_Sym = NOT_SYM;
        int_type N     = this->getN_Aktuell();
        for (int_type Index = 0; Index < N; ++Index)
        {
            int_type i = this->rows[Index];
            int_type j = this->cols[Index];
            if (i != j)
            {
                if (this->fortran_indexing)
                {
                    --i;
                    --j;
                }
                this->PushVal(i, j, this->vals[Index]);
            }
        }
    }
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                   private Memberfunktionen                                                          //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

template <typename real>
void SparseMatrix<real>::PushBackNote(int_type row, int_type col, real val)
{
    if (this->fortran_indexing)
    {
        ++row;
        ++col;
    }
    this->rows.push_back(row);
    this->cols.push_back(col);
    this->vals.push_back(val);
    this->nnz = int_type(this->vals.size());
}

template <typename real>
void SparseMatrix<real>::sortSparseMatrix(int_type DimROW, int_type SortROW[], int_type DimCOL, int_type SortCOL[])
{
    for (int_type i = 0; i < this->getN_Aktuell(); ++i)
    {
        int_type row = SortROW[this->rows[i] - 1] + 1;
        int_type col = SortCOL[this->cols[i] - 1] + 1;
        if (row <= 0 || row > DimROW)
            Errors::set_Error("Dimensionen des rownsotierungsarrays passt nicht zur Matrix!");
        if (col <= 0 || col > DimCOL)
            Errors::set_Error("Dimensionen des Column sotierungsarrays passt nicht zur Matrix!");
        if (this->isSymmetric() && row < col)
            std::swap(row, col);
        this->rows[i] = row;
        this->cols[i] = col;
    }
}

template <typename real>
void SparseMatrix<real>::SetHasToPushZeros(bool HasToPushZeros)
{
    this->has_to_push_zeroes = HasToPushZeros;
}

//-------------------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                                     //
//                                                  Externe Memberoperatoren                                                           //
//                                                                                                                                     //
//-------------------------------------------------------------------------------------------------------------------------------------//

template <typename real>
SparseMatrix<real> & SparseMatrix<real>::operator=(SparseMatrix<real> const & B)
{
    // Kopieren der Dimensionen und Flags:
    this->has_to_push_zeroes = B.has_to_push_zeroes;
    this->nnz                = B.nnz;
    this->nRows              = B.nRows;
    this->nCols              = B.nCols;
    this->fortran_indexing   = B.fortran_indexing;
    this->Flag_Sym           = B.Flag_Sym;
    // Kopieren der dynamischen Vektoren:
    this->rows = B.rows;
    this->cols = B.cols;
    this->vals = B.vals;
    return *this;
}

template <typename real>
void SparseMatrix<real>::toOStream(std::ostream & out)
{
    if (this->rows.size() == this->cols.size() && this->rows.size() == this->vals.size())
    {
        for (int_type i = 0; i < (int)this->vals.size(); i++)
            out << this->rows[i] << "   " << this->cols[i] << "   " << this->vals[i] << std::endl;
    }
    else
    {
        Errors::set_Error("Indexarrays und Wertearray einer Sparsematrix haben unterschiedliche Größen!");
    }
}

template class SparseMatrix<double>;
template class SparseMatrix<float>;
