/* (C) 2016, Andreas Huber, UniBw München LRT 1, Germany. All rights reserved. */

/*
 * Commercial license:
 *
 * To get more information about the commercial licence and the commercial use of OCPbasic and IPbasic please contact
 *      info@ocpbasic.com
 *
 * Open source license: (LGPLv3)
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 3.0 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file problem.cpp
 * Implementation of the class Problem.
 *
 * @author A. Huber
 * @since 21.04.2016
 */

#include "problem.h"
#include <lapack_wrapper/lapack_wrapper.hh>

namespace OCPbasic
{

    template <>
    const double Problem<double>::Inf = 1E+30;

    template <>
    const float Problem<float>::Inf = 1E+15;

    template <typename real>
    Problem<real>::Problem(int_type Ny, int_type Nu, int_type Np, int_type N_NLCSTR, int_type N_BOUND)
        : Ny(Ny)
        , Nu(Nu)
        , Np(Np)
        , N_NLCSTR(N_NLCSTR)
        , N_BOUND(N_BOUND)
        , N_VARPHI(1)
    {
        real eps             = 1e-16;
        this->tolCentralDiff = sqrt(eps);
        // Set bounds:
        this->g_max   = new real[this->N_NLCSTR];
        this->g_min   = new real[this->N_NLCSTR];
        this->psi_max = new real[this->N_BOUND];
        this->psi_min = new real[this->N_BOUND];
        this->y_max   = new real[this->Ny];
        this->y_min   = new real[this->Ny];
        this->u_max   = new real[this->Nu];
        this->u_min   = new real[this->Nu];
        this->p_max   = new real[this->Np];
        this->p_min   = new real[this->Np];

        // Finit diff workspace:
        int Nmax = Ny;
        if (Nu > Nmax)
            Nmax = Nu;
        if (Np > Nmax)
            Nmax = Np;
        if (N_NLCSTR > Nmax)
            Nmax = N_NLCSTR;
        if (N_BOUND > Nmax)
            Nmax = N_BOUND;
        this->workspace1       = new real[Nmax * Nmax];
        this->workspaceVector1 = new real[Nmax];
        this->workspaceVector2 = new real[Nmax];
        this->workspaceVector3 = new real[Nmax];

        this->Set_Default_Bounds();
    }

    template <typename real>
    Problem<real>::~Problem()
    {
        delete[] this->g_max;
        delete[] this->g_min;
        delete[] this->psi_max;
        delete[] this->psi_min;
        delete[] this->y_max;
        delete[] this->y_min;
        delete[] this->u_max;
        delete[] this->u_min;
        delete[] this->p_max;
        delete[] this->p_min;
        delete[] this->workspace1;
        delete[] this->workspaceVector1;
        delete[] this->workspaceVector2;
        delete[] this->workspaceVector3;
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Erste Ableitungen                                                                //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::D_ODE_X(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y, this->Ny, y_);
            y_[i] = y[i] + h;
            this->ODE(t, y_, u, p, res1);
            y_[i] = y[i] - h;
            this->ODE(t, y_, u, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_ODE_U(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->ODE(t, y, u_, p, res1);
            u_[i] = u[i] - h;
            this->ODE(t, y, u_, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_ODE_P(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->ODE(t, y, u, p_, res1);
            p_[i] = p[i] - h;
            this->ODE(t, y, u, p_, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::D_NLCSTR_X(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y, this->Ny, y_);
            y_[i] = y[i] + h;
            this->NLCSTR(t, y_, u, p, res1);
            y_[i] = y[i] - h;
            this->NLCSTR(t, y_, u, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_NLCSTR_U(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->NLCSTR(t, y, u_, p, res1);
            u_[i] = u[i] - h;
            this->NLCSTR(t, y, u_, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_NLCSTR_P(real t, real const y[], real const u[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->NLCSTR(t, y, u, p_, res1);
            p_[i] = p[i] - h;
            this->NLCSTR(t, y, u, p_, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::D_BOUND_XF(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * yF_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(yF, this->Ny, yF_);
            yF_[i] = yF[i] + h;
            this->BOUND(t0, tF, yF_, y0, p, res1);
            yF_[i] = yF[i] - h;
            this->BOUND(t0, tF, yF_, y0, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_BOUND_X0(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->BOUND(t0, tF, yF, y0_, p, res1);
            y0_[i] = y0[i] - h;
            this->BOUND(t0, tF, yF, y0_, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_BOUND_P(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->BOUND(t0, tF, yF, y0, p_, res1);
            p_[i] = p[i] - h;
            this->BOUND(t0, tF, yF, y0, p_, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::D_VARPHI_XF(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * yF_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(yF, this->Ny, yF_);
            yF_[i] = yF[i] + h;
            this->VARPHI(t0, tF, yF_, y0, p, res1);
            yF_[i] = yF[i] - h;
            this->VARPHI(t0, tF, yF_, y0, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_VARPHI_X0(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->VARPHI(t0, tF, yF, y0_, p, res1);
            y0_[i] = y0[i] - h;
            this->VARPHI(t0, tF, yF, y0_, p, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::D_VARPHI_P(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real h  = this->tolCentralDiff;
        real h2 = this->tolCentralDiff * 2.0;
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->VARPHI(t0, tF, yF, y0, p_, res1);
            p_[i] = p[i] - h;
            this->VARPHI(t0, tF, yF, y0, p_, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //-------------------------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                                     //
    //                                                    Zweite Ableitungen                                                               //
    //                                                                                                                                     //
    //-------------------------------------------------------------------------------------------------------------------------------------//

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                             Rechte Seite der DGL                                                    //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::L_DD_ODE_XX(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y, this->Ny, y_);
            y_[i] = y[i] + h;
            this->D_ODE_X(t, y_, u, p, m);
            this->multMatrixVector(m, lambda, res1);
            y_[i] = y[i] - h;
            this->D_ODE_X(t, y_, u, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_ODE_UU(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Nu, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->D_ODE_U(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res1);
            u_[i] = u[i] - h;
            this->D_ODE_U(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_ODE_PP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Np, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_ODE_P(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_ODE_P(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_ODE_XU(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->D_ODE_X(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res1);
            u_[i] = u[i] - h;
            this->D_ODE_X(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_ODE_XP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_ODE_X(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_ODE_X(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_ODE_UP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->Ny)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->Ny;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Nu, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_ODE_U(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_ODE_U(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                        Nichtlineare Nebenbedingungen                                                //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_XX(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y, this->Ny, y_);
            y_[i] = y[i] + h;
            this->D_NLCSTR_X(t, y_, u, p, m);
            this->multMatrixVector(m, lambda, res1);
            y_[i] = y[i] - h;
            this->D_NLCSTR_X(t, y_, u, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_UU(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Nu, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->D_NLCSTR_U(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res1);
            u_[i] = u[i] - h;
            this->D_NLCSTR_U(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_PP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Np, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_NLCSTR_P(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_NLCSTR_P(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_XU(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * u_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(u, this->Nu, u_);
            u_[i] = u[i] + h;
            this->D_NLCSTR_X(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res1);
            u_[i] = u[i] - h;
            this->D_NLCSTR_X(t, y, u_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_XP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_NLCSTR_X(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_NLCSTR_X(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_NLCSTR_UP(real t, real const y[], real const u[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_NLCSTR)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_NLCSTR;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Nu, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_NLCSTR_U(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_NLCSTR_U(t, y, u, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                               Randbedingungen                                                       //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::L_DD_BOUND_XFXF(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * yF_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(yF, this->Ny, yF_);
            yF_[i] = yF[i] + h;
            this->D_BOUND_XF(t0, tF, yF_, y0, p, m);
            this->multMatrixVector(m, lambda, res1);
            yF_[i] = yF[i] - h;
            this->D_BOUND_XF(t0, tF, yF_, y0, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_BOUND_X0X0(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->D_BOUND_X0(t0, tF, yF, y0_, p, m);
            this->multMatrixVector(m, lambda, res1);
            y0_[i] = y0[i] - h;
            this->D_BOUND_X0(t0, tF, yF, y0_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_BOUND_PP(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Np, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_BOUND_P(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_BOUND_P(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_BOUND_XFX0(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->D_BOUND_XF(t0, tF, yF, y0_, p, m);
            this->multMatrixVector(m, lambda, res1);
            y0_[i] = y0[i] - h;
            this->D_BOUND_XF(t0, tF, yF, y0_, p, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_BOUND_XFP(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_BOUND_XF(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_BOUND_XF(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_BOUND_X0P(real t0, real tF, real const yF[], real const y0[], real const p[], real const lambda[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols() || 0 == this->N_BOUND)
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = this->N_BOUND;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_BOUND_X0(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res1);
            p_[i] = p[i] - h;
            this->D_BOUND_X0(t0, tF, yF, y0, p_, m);
            this->multMatrixVector(m, lambda, res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    //---------------------------------------------------------------------------------------------------------------------//
    //                                                                                                                     //
    //                                                 Zielfunktion                                                        //
    //                                                                                                                     //
    //---------------------------------------------------------------------------------------------------------------------//

    template <typename real>
    void Problem<real>::L_DD_VARPHI_XFXF(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * yF_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(yF, this->Ny, yF_);
            yF_[i] = yF[i] + h;
            this->D_VARPHI_XF(t0, tF, yF_, y0, p, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            yF_[i] = yF[i] - h;
            this->D_VARPHI_XF(t0, tF, yF_, y0, p, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_VARPHI_X0X0(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->D_VARPHI_X0(t0, tF, yF, y0_, p, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            y0_[i] = y0[i] - h;
            this->D_VARPHI_X0(t0, tF, yF, y0_, p, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_VARPHI_PP(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Np, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_VARPHI_P(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            p_[i] = p[i] - h;
            this->D_VARPHI_P(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_VARPHI_XFX0(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * y0_  = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(y0, this->Ny, y0_);
            y0_[i] = y0[i] + h;
            this->D_VARPHI_XF(t0, tF, yF, y0_, p, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            y0_[i] = y0[i] - h;
            this->D_VARPHI_XF(t0, tF, yF, y0_, p, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_VARPHI_XFP(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_VARPHI_XF(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            p_[i] = p[i] - h;
            this->D_VARPHI_XF(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    template <typename real>
    void Problem<real>::L_DD_VARPHI_X0P(real t0, real tF, real const yF[], real const y0[], real const p[], OCPbasic::MatrixWrapper<real> & S_OUT)
    {
        S_OUT.zero_fill();
        if (0 == S_OUT.numRows() || 0 == S_OUT.numCols())
        {
            return;
        }
        real                          h       = this->tolCentralDiff;
        real                          h2      = this->tolCentralDiff * 2.0;
        int_type                      NLambda = 1;
        OCPbasic::MatrixWrapper<real> m(this->workspace1, NLambda, this->Ny, NLambda);
        for (int_type i = 0; i < S_OUT.numCols(); i++)
        {
            real * p_   = this->workspaceVector1;
            real * res1 = this->workspaceVector2;
            real * res2 = this->workspaceVector3;
            std::copy_n(p, this->Np, p_);
            p_[i] = p[i] + h;
            this->D_VARPHI_X0(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res1);
            p_[i] = p[i] - h;
            this->D_VARPHI_X0(t0, tF, yF, y0, p_, m);
            std::copy_n(m.get_data(), m.numRows(), res2);
            for (int_type j = 0; j < S_OUT.numRows(); j++)
            {
                real val = res1[j];
                res1[j]  = (val - res2[j]) / (h2);
            }
            S_OUT.load_column(res1, i);
        }
    }

    // Default bounds:
    template <typename real>
    void Problem<real>::Set_Default_Bounds()
    {
        // states:
        for (int_type i = 0; i < this->Ny; ++i)
        {
            this->y_min[i] = (-1.0) * Problem<real>::Inf;
            this->y_max[i] = Problem<real>::Inf;
        }
        // controlls:
        for (int_type i = 0; i < this->Nu; ++i)
        {
            this->u_min[i] = (-1.0) * Problem<real>::Inf;
            this->u_max[i] = Problem<real>::Inf;
        }
        // parameters:
        for (int_type i = 0; i < this->Np; ++i)
        {
            this->p_min[i] = (-1.0) * Problem<real>::Inf;
            this->p_max[i] = Problem<real>::Inf;
        }
        // nonlinear constraints:
        for (int_type i = 0; i < this->N_NLCSTR; ++i)
        {
            this->g_max[i] = Problem<real>::Inf;
            this->g_min[i] = (-1.0) * Problem<real>::Inf;
        }
        // boundary constraints:
        for (int_type i = 0; i < this->N_BOUND; ++i)
        {
            this->psi_max[i] = Problem<real>::Inf;
            this->psi_min[i] = (-1.0) * Problem<real>::Inf;
        }
    }

    template <typename real>
    void Problem<real>::multMatrixVector(const OCPbasic::MatrixWrapper<real> & matrix, const real * vector, real * result)
    {
        OCPbasic::int_type M     = matrix.numRows();
        OCPbasic::int_type N     = matrix.numCols();
        OCPbasic::int_type LDA   = matrix.lDim();
        const real         alpha = 1.0;
        const real         beta  = 0.0;
        const real *       A     = matrix.get_data();
        std::fill_n(result, M, 0.0);
        lapack_wrapper::Transposition tra = lapack_wrapper::TRANSPOSE;
        /*
        if (transA)
        {
            tra = lapack_wrapper::TRANSPOSE;
        }
        */
        lapack_wrapper::gemv(tra, M, N, alpha, A, LDA, vector, 1, beta, result, 1);
    }

    template class Problem<double>;
    template class Problem<float>;

} // namespace OCPbasic
